Package: SpaceTrooper
Type: Package
Title: SpaceTrooper performs Quality Control analysis of Image-Based
        spatial
Version: 1.0.0
Authors@R: 
        c(person("Dario", "Righelli",
            email="dario.righelli@gmail.com",
            role=c("aut", "cre"),
            comment=c(ORCID="0000-0003-1504-3583")),
        person(given="Benedetta", family="Banzi", role="aut"),
        person(given="Oriana", family="Romano", role="ctb"),
        person(given="Matteo", family="Merchionni", role="ctb"),
        person(given="Mattia", family="Forcato", role="ctb"),
        person(given="Silvio", family="Bicciato", role="aut"),
        person(given="Davide", family="Risso", role="ctb"))
Description: SpaceTrooper performs Quality Control analysis using data driven 
    GLM models of Image-Based spatial data, providing exploration plots, 
    QC metrics computation, outlier detection.
    It implements a GLM strategy for the detection of low quality cells in 
    imaging-based spatial data (Transcriptomics and Proteomics). 
    It additionally implements several plots for the visualization of 
    imaging based polygons through the ggplot2 package.
License: MIT + file LICENSE
Encoding: UTF-8
Depends: R (>= 4.4.0), SpatialExperiment
Imports: DropletUtils, S4Vectors, SummarizedExperiment, arrow,
        data.table, dplyr, e1071, ggplot2, ggpubr, robustbase, scater,
        scuttle, sf, sfheaders, cowplot, glmnet, rhdf5, methods, rlang,
        SpatialExperimentIO
Suggests: knitr, rmarkdown, BiocStyle, testthat (>= 3.0.0), withr,
        viridis
biocViews: Software, Transcriptomics, GeneExpression, QualityControl,
        Spatial, SingleCell, DataImport, ImmunoOncology
RoxygenNote: 7.3.3
VignetteBuilder: knitr
BugReports: https://github.com/drighelli/SpaceTrooper/issues
URL: https://github.com/drighelli/SpaceTrooper
Config/testthat/edition: 3
git_url: https://git.bioconductor.org/packages/SpaceTrooper
git_branch: RELEASE_3_22
git_last_commit: 88c0696
git_last_commit_date: 2025-10-29
Repository: Bioconductor 3.22
Date/Publication: 2025-10-29
NeedsCompilation: no
Packaged: 2025-10-30 06:38:00 UTC; biocbuild
Author: Dario Righelli [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-1504-3583>),
  Benedetta Banzi [aut],
  Oriana Romano [ctb],
  Matteo Merchionni [ctb],
  Mattia Forcato [ctb],
  Silvio Bicciato [aut],
  Davide Risso [ctb]
Maintainer: Dario Righelli <dario.righelli@gmail.com>

# Generated by roxygen2: do not edit by hand

export(.getActiveGeometryName)
export(.renameGeometry)
export(.setActiveGeometry)
export(addPolygonsToSPE)
export(checkOutliers)
export(computeAreaFromPolygons)
export(computeAspectRatioFromPolygons)
export(computeCenterFromPolygons)
export(computeLambda)
export(computeMissingMetricsMerfish)
export(computeMissingMetricsXenium)
export(computeOutliersQCScore)
export(computeQCScore)
export(computeQCScoreFlags)
export(computeSpatialOutlier)
export(computeThresholdFlags)
export(computeTrainDF)
export(getFencesOutlier)
export(getModelFormula)
export(plotCellsFovs)
export(plotCentroids)
export(plotMetricHist)
export(plotPolygons)
export(plotQScoreTerms)
export(plotZoomFovsMap)
export(qcFlagPlots)
export(readAndAddPolygonsToSPE)
export(readCosmxProteinSPE)
export(readCosmxSPE)
export(readMerfishSPE)
export(readPolygons)
export(readPolygonsCosmx)
export(readPolygonsMerfish)
export(readPolygonsXenium)
export(readXeniumSPE)
export(spatialPerCellQC)
export(trainModel)
export(updateCosmxProteinSPE)
export(updateCosmxSPE)
export(updateXeniumSPE)
import(SpatialExperiment)
importFrom(DropletUtils,read10xCounts)
importFrom(S4Vectors,'metadata<-')
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,cbind.DataFrame)
importFrom(S4Vectors,metadata)
importFrom(SpatialExperiment,SpatialExperiment)
importFrom(SpatialExperimentIO,readCosmxSXE)
importFrom(SpatialExperimentIO,readXeniumSXE)
importFrom(SummarizedExperiment,assays)
importFrom(SummarizedExperiment,colData)
importFrom(SummarizedExperiment,rowData)
importFrom(arrow,read_parquet)
importFrom(cowplot,get_legend)
importFrom(cowplot,plot_grid)
importFrom(data.table,fread)
importFrom(data.table,merge.data.table)
importFrom(dplyr,case_when)
importFrom(dplyr,distinct)
importFrom(dplyr,filter)
importFrom(dplyr,group_by)
importFrom(dplyr,left_join)
importFrom(dplyr,mutate)
importFrom(dplyr,n)
importFrom(dplyr,pull)
importFrom(dplyr,select)
importFrom(dplyr,ungroup)
importFrom(e1071,skewness)
importFrom(ggplot2,aes)
importFrom(ggplot2,aes_string)
importFrom(ggplot2,annotate)
importFrom(ggplot2,coord_fixed)
importFrom(ggplot2,element_blank)
importFrom(ggplot2,element_line)
importFrom(ggplot2,element_rect)
importFrom(ggplot2,element_text)
importFrom(ggplot2,geom_histogram)
importFrom(ggplot2,geom_point)
importFrom(ggplot2,geom_sf)
importFrom(ggplot2,geom_text)
importFrom(ggplot2,geom_vline)
importFrom(ggplot2,ggplot)
importFrom(ggplot2,ggtitle)
importFrom(ggplot2,guide_legend)
importFrom(ggplot2,guides)
importFrom(ggplot2,labs)
importFrom(ggplot2,margin)
importFrom(ggplot2,scale_color_gradient)
importFrom(ggplot2,scale_color_gradientn)
importFrom(ggplot2,scale_color_manual)
importFrom(ggplot2,scale_colour_manual)
importFrom(ggplot2,scale_fill_identity)
importFrom(ggplot2,scale_fill_manual)
importFrom(ggplot2,scale_fill_viridis_c)
importFrom(ggplot2,theme)
importFrom(ggplot2,theme_bw)
importFrom(ggplot2,theme_minimal)
importFrom(ggplot2,theme_void)
importFrom(ggpubr,annotate_figure)
importFrom(ggpubr,ggarrange)
importFrom(glmnet,cv.glmnet)
importFrom(glmnet,glmnet)
importFrom(grDevices,colors)
importFrom(methods,is)
importFrom(rhdf5,h5dump)
importFrom(rlang,.data)
importFrom(robustbase,adjbox)
importFrom(robustbase,mc)
importFrom(scater,addPerCellQC)
importFrom(scater,plotColData)
importFrom(scuttle,isOutlier)
importFrom(scuttle,outlier.filter)
importFrom(sf,st_as_sf)
importFrom(sf,st_buffer)
importFrom(sf,st_cast)
importFrom(sf,st_geometry)
importFrom(sf,st_geometry_type)
importFrom(sf,st_is_valid)
importFrom(sf,st_polygon)
importFrom(sf,st_sf)
importFrom(sf,st_sfc)
importFrom(sfheaders,sf_polygon)
importFrom(stats,as.formula)
importFrom(stats,model.matrix)
importFrom(stats,predict)
importFrom(stats,quantile)

# SpaceTrooper
[![R](https://img.shields.io/badge/R-%3E%3D4.4.0-blue)](https://www.r-project.org/)
[![License: MIT](https://img.shields.io/badge/License-MIT-blue.svg)](https://opensource.org/license/mit)

`SpaceTrooper` is an `R/Bioconductor` package for Quality Control
of imaging-based spatial transcriptomics and proteomics data. It provides 
multi-platform data harmonization, cell-level QC, and visualization utilities 
for **CosMx**, **Xenium**, and **MERFISH** technologies.

<p align="left">
	<img src="inst/SpaceTrooper_logo.png" alt="SpaceTrooper Logo" width="150"/>
</p>



`SpaceTrooper` main highlights:
- Read and harmonize imaging-based spatial data into SpatialExperiment objects.
- Compute polygon-derived metrics (area, aspect ratio, borders).
- Calculate QC metrics, detect spatial outliers and calculate Quality Score.
- Visualize computed metrics on centroids or polygons.
- Support CosMx RNA/Protein Assay, Xenium, and MERFISH data.

<p align="left">
	<img src="inst/GraphAbstract.jpg" alt="SpaceTrooper Logo" height="500"/>
</p>


<br/><br/>


## System requirements and installation

### Requirements
`SpaceTrooper` depends on **R version >= 4.4.0**, **Bioconductor >= 3.22** and **SpatialExperiment (1.20.0)**.
The package has been tested on Windows, macOS, and Linux.

### Bioconductor (Release 3.22)
The stable version of the package is in `Bioconductor 3.22`, and it can be installed with the following code.
If you already have Bioconductor 3.22, it usually takes less than a minute.

```r
if (!requireNamespace("BiocManager", quietly = TRUE)) {
	install.packages("BiocManager")
}
BiocManager::install("SpaceTrooper")
```

We suggest to use the a version of `SpaceTrooper` higher or equal to 1.1.3, which can be downloaded as 
explained in the "devel" section.

### Bioconductor (devel)
The devel version of `SpaceTrooper` can be downloaded as follows. 
Please remember that it is stable only if using an R version >= 4.4.0.

```r
if (!requireNamespace("BiocManager", quietly = TRUE)) {
	install.packages("BiocManager")
}

BiocManager::install("SpaceTrooper", ref="devel")
```

### Anaconda (for conda/mamba environments)

`SpaceTrooper` package can be installed in a Conda-compatible environment on Linux (x86_64), using the provided installation file [SpaceTrooper.yml](inst/SpaceTrooper.yml). 
Before proceeding with the provided installation, you have to install a Conda-compatible distribution, such as [Anaconda](https://www.anaconda.com/download).

To create the environment run the following code:

```bash
conda env create -f SpaceTrooper.yml
conda activate SpaceTrooper
```

Otherwise, if you installed Mamba/Micromamba distribution you have to run 'mamba' instead of 'conda'

```bash
mamba env create -f SpaceTrooper.yml
mamba activate SpaceTrooper
```

Once the environment is activated, open `R` and install the latest developmental `SpaceTrooper` version as described above.

```r
if (!requireNamespace("BiocManager", quietly = TRUE)) {
	install.packages("BiocManager")
}
BiocManager::install("SpaceTrooper", ref="devel") 
```
<br/><br/>


## Quick start

The typical `SpaceTrooper` workflow require the following steps:

1. read a dataset into a SpatialExperiment object
2. load existing polygon data (optionally)
3. calculate per cell-QC metrics
4. compute QS
5. visualize polygons, centroids, and QC summaries.

Minimal example pipeline:

```r
library(SpaceTrooper)

# 1. Read a dataset (choose the reader that matches your platform).
spe <- readCosmxSPE("path/to/cosmx")
# spe <- readXeniumSPE("path/to/xenium")
# spe <- readMerfishSPE("path/to/merfish")

# 2. Add polygons from existing polygon data.
spe <- readAndAddPolygonsToSPE(spe, boundariesType="csv")

# 3. Calculate and add QC metrics to cell metadata.
spe <- spatialPerCellQC(spe)

# 4. Compute QS and, optionally, flag cells with a score higher than 'qsThreshold'.
spe <- computeQScore(spe)
spe <- computeQScoreFlags(spe, qsThreshold=0.5)

# 5. Visualization
## Visualize cells as dots in their centroid coordinates, colored by a column in `colData(spe)` (e.g., QS computed above).
plotCentroids(spe, colourBy="QScore")

## Visualize cells using their polygon shapes, colored by a column in `colData(spe)` (e.g., QS computed above).
## To visualize polygons step 2 is mandatory.
## Polygons can be cumbersome to plot for large datasets (e.g., entire slides with more than 100,000 cells), hence centroids may be preferred. 
plotPolygons(spe, colourBy="QScore")
```

Custom `modelFormula` values may select any subset of the supported predictors
`log2SignalDensity`, `Area_um`, `log2AspectRatio`, and
`log2Ctrl_total_ratio`. Additive formulas and selected interactions are
preserved exactly; unsupported predictors and transformations are rejected.
The border-effect expression is available only for CosMx data:
`I(abs(log2AspectRatio) * as.numeric(dist_border < 50))`.

The former `computeQCScore()`, `computeQCScoreFlags()`,
`computeOutliersQCScore()`, and `applyQCScoreModel()` APIs remain available as
deprecated compatibility functions and retain their historical output names.
<br/><br/>


## Data support
`SpaceTrooper` provides readers and QC utilities tailored to:

- **CosMx** RNA and protein data
- **Xenium**
- **MERFISH**

<br/><br/>


## Vignettes
The package includes three vignettes that cover:

- [Imaging-based spatial transcriptomics workflows](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/RNA_tutorial_vignette.html)
- [CosMx protein workflow](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/Protein_tutorial_vignette.html)
- [SpaceTrooper utilities](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/SpaceTrooper_utilities.html)
- [SpaceTrooper interoperability](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/interoperability.html)


Refer to the first vignette if working with CosMx, Xenium or MERFISH spatial trancriptomics platforms; refer to the second vignette if working with CosMx Protein Assay data. Third vignette illustrates how SPE objects obtained with SpatialExperimentIO can be imported into our pipeline, along with some other package utility functions and plots. 

Function-level help is available through standard R documentation, for example:

```r
?computeQScore
?SpaceTrooper-deprecated
?plotPolygons
```

<br/><br/>


## Issues
Please report bugs or feature requests at:
[https://github.com/drighelli/SpaceTrooper/issues](https://github.com/drighelli/SpaceTrooper/issues).

## License
`SpaceTrooper` is released under a [MIT license](LICENSE) file.

## Citations

Benedetta Banzi, Dario Righelli, Matteo Marchionni, Oriana Romano, Mattia Forcato, Davide Risso, Silvio Bicciato. *SpaceTrooper: a quality control framework for imaging-based spatial omics data*. bioRxiv 2025.12.24.696336; doi: [https://doi.org/10.64898/2025.12.24.696336](https://doi.org/10.64898/2025.12.24.696336)

Sys.setenv(LIBARROW_MINIMAL = "false")
Sys.setenv(ARROW_WITH_ZSTD = "ON")
# Sys.setenv(ARROW_S3="ON")
Sys.setenv(NOT_CRAN="true")
# install.packages("arrow", repos = "https://arrow-r-nightly.s3.amazonaws.com")
system("echo $ARROW_WITH_ZSTD")
system("echo $LIBARROW_MINIMAL")
install.packages("arrow")
library(arrow)
arrow::install_arrow(minimal=FALSE,nightly=TRUE,verbose=TRUE)
arrow::arrow_info()
pol_file <- "~/Downloads/Xenium_data/pancreas/outs/cell_boundaries.parquet"
devtools::load_all()
polygons <- readPolygonsXenium(pol_file, keepMultiPol=TRUE)

## to read xenium parquet it's needed the support for ZSTD in arrow package
## just install arrow from R not with external libraries

#' readCosmxSPE
#' @name readCosmxSPE
#' @rdname readCosmxSPE
#' @aliases readCosmxSPE readCosmxProteinSPE
#' @description
#' Read and Construct a SpatialExperiment Object from CosMx Data
#'
#' This function reads in data from Nanostring CosMx files and constructs a
#' `SpatialExperiment` object, optionally including polygon data.
#'
#' @param dirName A character string specifying the directory containing the
#' CosMx data files.
#' @param sampleName A character string specifying the sample name. Default is
#' `"sample01"`.
#' @param coordNames A character vector specifying the names of the spatial
#' coordinate columns in the data. Default is `c("CenterX_global_px",
#' "CenterY_global_px")`.
#' @param countMatFPattern A character string specifying the pattern to match
#' the count matrix file. Default is `"exprMat_file.csv"`.
#' @param metadataFPattern A character string specifying the pattern to match
#' the metadata file. Default is `"metadata_file.csv"`.
#' @param polygonsFPattern A character string specifying the pattern to match
#' the polygons file. Default is `"polygons.csv"`.
#' @param fovPosFPattern A character string specifying the pattern to match the
#' FOV positions file. Default is `"fov_positions_file.csv"`.
#' @param fovdims A named numeric vector specifying the dimensions of the FOV
#' in pixels. Default is `c(xdim=4256, ydim=4256)`.
#'
#' @return A `SpatialExperiment` object containing the read CosMx data,
#' including count matrices, metadata, and optionally polygons.
#'
#' @details The function reads in the specified files for count matrices,
#' metadata, and FOV positions, and constructs a `SpatialExperiment` object.
#' Optionally, polygon data can be read and added to the object.
#'
#' readCosmxProteinSPE is a wrapper of readCosmxSPE, it only changes the
#' technology metadata in Nanostring_CosMx_Protein.
#'
#' @author Dario Righelli, Benedetta Banzi
#'
#' @importFrom data.table fread merge.data.table
#' @importFrom SpatialExperiment SpatialExperiment
#' @importFrom S4Vectors DataFrame
#' @importFrom dplyr mutate
#' @importFrom SpatialExperimentIO readCosmxSXE
#' @export
#'
#' @examples
#' cospath <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"),
#'    package="SpaceTrooper")
#' spe <- readCosmxSPE(cospath, sampleName="DBKero_Tiny")
readCosmxSPE <- function(dirName, sampleName="sample01",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    countMatFPattern="exprMat_file.csv", metadataFPattern="metadata_file.csv",
    polygonsFPattern="polygons.csv", fovPosFPattern="fov_positions_file.csv",
    fovdims=c(xdim=4256, ydim=4256)) {

    stopifnot(all(names(fovdims) == c("xdim", "ydim"), file.exists(dirName)))

    spe <- SpatialExperimentIO::readCosmxSXE(dirName=dirName, returnType="SPE",
        countMatPattern=countMatFPattern, metaDataPattern=metadataFPattern,
        coordNames=coordNames, addFovPos=TRUE, fovPosPattern=fovPosFPattern,
        altExps=NULL,addParquetPaths=FALSE)

    spe <- .setupCosmxSPE(spe, dirName, sampleName, polygonsFPattern, fovdims)
    return(spe)
}


#' @export
readCosmxProteinSPE <- function(dirName, sampleName="sample01",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    countMatFPattern="exprMat_file.csv", metadataFPattern="metadata_file.csv",
    polygonsFPattern="polygons.csv", fovPosFPattern="fov_positions_file.csv",
    fovdims=c(xdim=4256, ydim=4256)) {

    spe <- readCosmxSPE(dirName, sampleName, coordNames, countMatFPattern,
        metadataFPattern, polygonsFPattern, fovPosFPattern, fovdims)

    metadata(spe)$technology <- "Nanostring_CosMx_Protein"
    return(spe)
}

#' .checkFovPositionVersion
#' @rdname dot-checkFovPositionVersion
#' @description
#'
#' Check and Standardize FOV Position Column Names
#'
#' This internal utility function standardizes column names of a data frame
#' containing Field of View (FOV) positional information.
#' It modifies column names to ensure compatibility with expected naming
#' conventions, including support for older formats.
#'
#' Specifically, it:
#' - Renames any column containing "FOV" to "fov"
#' - Converts columns with coordinates matching "X", "Y", or "Z" to lowercase
#' - Replaces suffix "_px" with "_global_px" for coordinate pixel columns
#' - If the input contains `x_mm` and `y_mm` columns, the function computes
#' corresponding `x_global_px` and `y_global_px` values by converting from
#' millimeters to pixels using a fixed resolution factor (0.12028 mm/pixel).
#' @param spe A `SpatialExperiment` containing FOV position information
#' in the metadata to be standardized.
#'
#' @return A `SpatialExperiment` with updated and standardized column names
#' for the metadata `fov_position` `data.frame`.
#'
#' @keywords internal
.checkFovPositionVersion <- function(spe)
{
    fovpos <- metadata(spe)$fov_positions
    fovcidx <- grep("FOV", colnames(fovpos)) # works also with older vers
    if(length(fovcidx)!=0) colnames(fovpos)[fovcidx] <- "fov"
    fovcrdx <- grep("[X|Y|Z]", colnames(fovpos))
    if(length(fovcrdx)!=0) colnames(fovpos)[fovcrdx] <-
        tolower(colnames(fovpos)[fovcrdx])
    fovccdx <- grep("[x|y]_px", colnames(fovpos))
    if(length(fovccdx)!=0) colnames(fovpos)[fovccdx] <-
        gsub("_px", "_global_px", colnames(fovpos)[fovccdx])

    if(length(grep("x_mm", colnames(fovpos))!=0)) {
        fovpos <- fovpos |>
            dplyr::mutate(x_global_px = x_mm/0.12028*10^3,
                        y_global_px = (y_mm/0.12028*10^3) - 4256)
    }
    idx <- fovpos$fov %in% unique(spe$fov)
    fovpos <- fovpos[idx, ]

    fovpos <- fovpos[order(fovpos$fov), ]
    metadata(spe)$fov_positions <- fovpos
    return(spe)
}


#' updateCosmxSPE
#'
#' @description
#' Update a SpatialExperiment object derived from CosMx data by adding
#' polygons, FOV dimensions, standardized column names, and metadata.
#'
#' @param spe SpatialExperiment object.
#' @param dirName Directory containing CosMx output files (e.g., polygon CSVs).
#' @param sampleName Character scalar, sample identifier stored in
#'   \code{colData(spe)$sample_id}. Default \code{"sample01"}.
#' @param polygonsFPattern Character, pattern used by \code{list.files()} to
#'   locate polygon files. Default \code{"polygons.csv"}.
#' @param fovdims Named numeric vector with entries \code{xdim} and \code{ydim}
#'   representing the FOV dimensions in pixels.
#'
#' @details
#' The function standardizes CosMx SPE structure by:
#' - creating unique cell names of the form \code{f<fov>_c<cell_ID>};
#' - ensuring consistent cell identifiers and sample metadata;
#' - recording FOV dimensions, polygon paths, and technology type in
#'   \code{metadata(spe)}.
#'
#' @return A SpatialExperiment object with updated metadata and column names.
#'
#' @seealso readCosmxSPE, readCosmxProteinSPE
#'
#' @export
#' @examples
#' cospath <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"),
#'     package="SpaceTrooper")
#' spe <- SpatialExperimentIO::readCosmxSXE(dirName=cospath,
#'     addParquetPaths=FALSE)
#' spe <- updateCosmxSPE(spe, dirName=cospath, sampleName="DBKero_Tiny")
updateCosmxSPE <- function(spe, dirName, sampleName="sample01",
                        polygonsFPattern="polygons.csv",
                        fovdims=c(xdim=4256, ydim=4256)) {
    stopifnot("spe is not a SpatialExperiment"=is(spe, "SpatialExperiment"))
    stopifnot("fovdims not x|y dim"=all(names(fovdims) == c("xdim", "ydim")))
    stopifnot("dirName not exists"=file.exists(dirName))
    spe <- .setupCosmxSPE(spe, dirName, sampleName, polygonsFPattern, fovdims)
    return(spe)
}

#' .setupCosmxSPE
#'
#' @description
#' Internal helper to configure a SpatialExperiment from CosMx data:
#' standardizes column names, sets cell IDs, and updates metadata.
#'
#' @param spe SpatialExperiment object.
#' @param dirName Directory containing CosMx output files.
#' @param sampleName Character sample identifier. Default \code{"sample01"}.
#' @param polygonsFPattern Character, pattern for polygon files.
#' Default \code{"polygons.csv"}.
#' @param fovdims Named numeric vector \code{c(xdim=, ydim=)} giving
#' field-of-view size.
#'
#' @return Updated SpatialExperiment object.
#'
#' @keywords internal
#' @noRd
.setupCosmxSPE <- function(spe, dirName, sampleName="sample01",
                            polygonsFPattern="polygons.csv",
                            fovdims=c(xdim=4256, ydim=4256))
{
    pol_file <- list.files(dirName, polygonsFPattern, full.names=TRUE)
    cn <- paste0("f", spe$fov, "_c", spe$cell_ID)
    colnames(spe) <- cn
    rownames(colData(spe)) <- cn
    if(length(grep("cell_id", colnames(colData)))!=0)
        warning("Overwriting existing cell_id column in colData")
    spe$cell_id <- cn
    spe <- .checkFovPositionVersion(spe)
    metadata(spe) <- list(fov_positions=metadata(spe)$fov_positions,
            fov_dim=fovdims, polygons=pol_file, technology="Nanostring_CosMx")

    names(colData(spe))[names(colData(spe)) == "cell_ID"] <- "cellID"
    spe$sample_id <- sampleName
    return(spe)
}

#' updateCosmxProteinSPE
#'
#' @description
#' Update a SpatialExperiment object corresponding to Nanostring CosMx
#' Protein data by adding metadata identifying the technology and
#' optionally passing through file-location parameters.
#'
#' @param spe SpatialExperiment object.
#' @param dirName Directory containing CosMx Protein data files.
#' @param sampleName Character sample ID. Default \code{"sample01"}.
#' @param coordNames Character vector of length two indicating coordinate
#'   column names in the per-cell metadata. Default
#'   \code{c("CenterX_global_px","CenterY_global_px")}.
#' @param countMatFPattern Character pattern for the counts matrix file.
#' @param metadataFPattern Character pattern for the single-cell metadata file.
#' @param polygonsFPattern Character pattern for the polygon file(s).
#' @param fovPosFPattern Character pattern for the FOV positions file.
#' @param fovdims Named numeric vector with FOV size in pixels.
#'
#' @details
#' This function sets \code{metadata(spe)$technology <- "Nanostring_CosMx_Protein"}.
#' It does not modify other assay or metadata components.
#'
#' @return SpatialExperiment object with updated technology metadata.
#'
#' @seealso readCosmxProteinSPE, readCosmxSPE
#'
#' @export
#' @examples
#' protfolder <- system.file( "extdata", "S01_prot", package="SpaceTrooper")
#' spe <- SpatialExperimentIO::readCosmxSXE(dirName=protfolder,
#'     addParquetPaths=FALSE)
#' spe <- updateCosmxProteinSPE(spe, protfolder, sampleName="cosmx_prots")
updateCosmxProteinSPE <- function(spe, dirName, sampleName="sample01",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    countMatFPattern="exprMat_file.csv", metadataFPattern="metadata_file.csv",
    polygonsFPattern="polygons.csv", fovPosFPattern="fov_positions_file.csv",
    fovdims=c(xdim=4256, ydim=4256)) {

    stopifnot(is(spe, "SpatialExperiment"))
    spe <- updateCosmxSPE(spe, dirName, sampleName, polygonsFPattern, fovdims)
    metadata(spe)$technology <- "Nanostring_CosMx_Protein"
    return(spe)
}

#' plotCellsFovs
#' @name plotCellsFovs
#' @rdname plotCellsFovs
#'
#' @description
#' Plot cell centroids in FoVs
#' Creates a scatter plot with cell centroids arranged in their FoVs as an
#' overlapping grid.
#'
#' @param spe A `SpatialExperiment` object with `fov` in `colData`.
#' @param sampleId Character string identifying which sample to plot.
#'   Default: `unique(spe$sample_id)`.
#' @param pointCol Color for the cell centroids. Default: `"firebrick"`.
#' @param numbersCol Color for the FoV labels. Default: `"black"`.
#' @param alphaNumbers Numeric transparency for FoV labels. Default: `0.8`.
#' @param fovDim numeric with two named dimensions xdim, ydim. (Default is
#' metadata(spe)$fov_dim)
#'
#' @return A `ggplot` object showing cell centroids and FoV boundaries.
#'
#' @importFrom ggplot2 ggplot aes annotate geom_point geom_text ggtitle
#' coord_fixed
#' @importFrom rlang .data
#' @export
#'
#' @examples
#' example(readCosmxSPE)
#' g <- plotCellsFovs(spe)
#' print(g)
plotCellsFovs <- function(spe, sampleId=unique(spe$sample_id),
                        pointCol="firebrick", numbersCol="black",
                        alphaNumbers=0.8, fovDim=metadata(spe)$fov_dim)
{
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot("fov" %in% names(colData(spe)))
    stopifnot( all(names(fovDim) %in% c("xdim","ydim")) )

    spd <- as.data.frame(spatialCoords(spe))
    x_coord <- spatialCoordsNames(spe)[1]
    y_coord <- spatialCoordsNames(spe)[2]
    ggp <- ggplot() +
        geom_point(data=spd, mapping=aes(x=.data[[x_coord]],
                                        y=.data[[y_coord]]),
                    colour=pointCol,
                    fill=pointCol,
                    size=0.05, alpha=0.8) +
        annotate("rect",
            xmin=metadata(spe)$fov_positions["x_global_px"][ , , drop=TRUE],
            xmax=metadata(spe)$fov_positions["x_global_px"][ , , drop=TRUE] +
                fovDim[["xdim"]],
            ymin=metadata(spe)$fov_positions["y_global_px"][ , , drop=TRUE],
            ymax=metadata(spe)$fov_positions["y_global_px"][ , , drop=TRUE] +
                fovDim[["ydim"]],
            alpha=.2, color="black", linewidth=0.2) +
        geom_text(aes(x=metadata(spe)$fov_positions["x_global_px"][,,drop=TRUE]+
                        fovDim[["xdim"]]/2,
                    y=metadata(spe)$fov_positions["y_global_px"][,,drop=TRUE]+
                        fovDim[["ydim"]]/2,
                    label=metadata(spe)$fov_positions["fov"][,,drop=TRUE]),
                    color=numbersCol, fontface="bold", alpha=alphaNumbers) +
        ggtitle(sampleId) +
        .fov_image_theme(backColor="white", backBorder="white",
                        titleCol="black") + ggplot2::coord_fixed()
    return(ggp)
}

#' plotCentroids
#' @name plotCentroids
#' @rdname plotCentroids
#' @description
#' Plot Spatial Coordinates for a SpatialExperiment Object
#' This function generates a ggplot of spatial coordinates from a
#' `SpatialExperiment` object, optionally coloring the points by a specified
#' column in `colData`.
#'
#' @param spe A `SpatialExperiment` object containing spatial
#' transcriptomics data.
#' @param colourBy An optional character string specifying the column in
#' `colData(spe)` to use for coloring the points. If `NULL`, all points will be
#' colored the same.
#' @param colourLog Logical to log-transform the data to enhance visualization
#' (Default is FALSE).
#' @param sampleId A character string specifying the sample identifier to be
#' used as the plot title. (Default is the unique sample ID from `spe`)
#' @param isNegativeProbe A logical value indicating whether to apply a custom
#' color gradient for negative probe data. (Default is `FALSE`)
#' @param palette A vector of colors to be used as a custom palette. For
#' categorical data, this should be a vector of colors with the same length as
#' the number of levels in `colourBy`. For continuous data, this should be a
#' vector of colors used to create a gradient.
#' @param pointCol A character string specifying the color of the points when
#' `colourBy` is `NULL`. (Default is `"darkmagenta"`)
#' @param size A numeric value specifying the size of the points. (Default is
#' `0.05`)
#' @param alpha A numeric value specifying the transparency level of the points.
#' (Default is `0.2`)
#' @param aspectRatio A numeric value specifying the aspect ratio of the plot.
#' (Default is `1`)
#' @return A `ggplot` object representing the spatial coordinates plot of
#' polygon centroids.
#'
#' @import SpatialExperiment
#' @importFrom ggplot2 geom_point aes_string theme_bw theme ggtitle guides
#' guide_legend scale_color_gradient scale_color_manual scale_color_gradientn
#' @importFrom scater plotColData
#' @importFrom SummarizedExperiment colData
#' @importFrom ggplot2 ggplot geom_point aes ggtitle theme_bw coord_fixed
#' @importFrom ggplot2 scale_color_manual scale_color_gradientn
#' @export
#'
#' @examples
#' example(readCosmxSPE)
#' g <- plotCentroids(spe, colourBy="Mean.DAPI")
#' print(g)
plotCentroids <- function(spe, colourBy=NULL, colourLog=FALSE,
                        sampleId=unique(spe$sample_id),
                        isNegativeProbe=FALSE, palette=NULL,
                        pointCol="darkmagenta", size=0.05, alpha=0.8,
                        aspectRatio=1) {
    stopifnot(is(spe, "SpatialExperiment"))
    if(is.null(colourBy)) {
        ggp <- ggplot(data.frame(spatialCoords(spe)),
                    aes(x=.data[[spatialCoordsNames(spe)[1]]],
                        y=.data[[spatialCoordsNames(spe)[2]]])) +
            geom_point(colour=pointCol, fill=pointCol,
                        size=size, alpha=alpha) + ggplot2::ggtitle(sampleId) +
            ggplot2::theme_bw() + ggplot2::coord_fixed()
    } else {
        if(colourLog) {
            stopifnot(colourBy %in% names(colData(spe)))
            colourByo <- colourBy
            colourBy <- paste0("log(", colourByo, ")")
            colData(spe)[[colourBy]] <- log1p(colData(spe)[[colourByo]])
        }
        ggp <- ggplot(data.frame(colData(spe), spatialCoords(spe)),
                    aes(x=.data[[spatialCoordsNames(spe)[1]]],
                        y=.data[[spatialCoordsNames(spe)[2]]],
                        colour=.data[[colourBy]],
                        fill=.data[[colourBy]])) +
            geom_point(size=size, alpha=alpha) + coord_fixed()
        if( all(!is(spe[[colourBy]], "factor"),
                !is(spe[[colourBy]], "logical"))) {
            ggp <- ggp + ggplot2::scale_colour_viridis_c() +
                ggplot2::scale_fill_viridis_c() + coord_fixed()
        }
        if(isNegativeProbe) {
            ggp <- ggp + scale_color_gradient(low="white", high="red",
                name=colourBy) + .negative_image_theme()
        } else if(all(!is.null(palette), (palette %in% names(colData(spe))))) {
            palette <- createPaletteFromColData(spe, paletteNames=colourBy,
                                                    paletteColors=palette)
            if(is.factor(colData(spe)[[colourBy]])) {
                ggp <- ggp + scale_color_manual(values=palette)
            } else if(is.numeric(colData(spe)[[colourBy]])) {
                ggp <- ggp + scale_color_gradientn(colors=palette)
            }
        }
    }
    ggp <- ggp + ggtitle(sampleId) +
        theme(aspect.ratio=aspectRatio, plot.title=element_text(hjust=0.5))
    if(!isNegativeProbe) ggp <- ggp + theme_bw()
    return(ggp)
}


#' plotMetricHist
#' @name plotMetricHist
#' @rdname plotMetricHist
#' @description Plot a Histogram for a Given Metric in a SpatialExperiment
#' Object
#'
#' This function generates a histogram for a specified metric in a
#' `SpatialExperiment` object.
#'
#' @param spe A `SpatialExperiment` object containing spatial transcriptomics
#' data.
#' @param metric A character string specifying the name of the metric
#' (column in `colData(spe)`) to plot.
#' @param fillColor A character string specifying the fill color of the
#' histogram bars. (Default is `"#69b3a2"`)
#' @param useFences A character string specifying the name of the column in
#' `colData(spe)` that contains the fence thresholds (typically from an outlier
#' filter). If `NULL`, no fences will be plotted. (Default is `NULL`)
#' @param fencesColors A named character vector specifying the colors to use
#' for the lower and higher fences. The names should be `"lower"` and `"higher"`
#'. (Default is `c("lower"="purple4", "higher"="tomato")`)
#' @param bins An integer specifying the number of bins to use in the histogram.
#' (Default is `30`)
#' @param binWidth A numeric value specifying the width of the bins. If `NULL`,
#' the bin width will be automatically determined based on the `bins` parameter.
#' (Default is `NULL`)
#'
#' @return A `ggplot` object representing the histogram of the specified metric.
#'
#' @importFrom ggplot2 ggplot geom_histogram aes ggtitle theme_bw geom_vline
#' labs scale_colour_manual
#' @importFrom SummarizedExperiment colData
#' @importFrom methods is
#' @export
#' @examples
#' example(readCosmxSPE)
#' g <- plotMetricHist(spe, metric="Mean.DAPI")
#' print(g)
plotMetricHist <- function(spe, metric, fillColor="#c0c8cf",
        useFences=NULL, fencesColors=c("lower"="purple4", "higher"="tomato"),
        bins=30, binWidth=NULL)
{
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot(metric %in% names(colData(spe)))

    ggp <- ggplot(data=as.data.frame(colData(spe))) +
            geom_histogram(aes(x=.data[[metric]]), fill=fillColor,
                    bins=bins, binwidth=binWidth)
    if (!is.null(useFences))
    {
        stopifnot(useFences %in% names(colData(spe)))
        fences <- getFencesOutlier(spe, useFences, "both", 2)
        fences_labs <- paste0(names(fences), ": ", fences)
        names(fencesColors) <- fences_labs
        ggp <- ggp +
            geom_vline(aes(xintercept=fences[1], color=fences_labs[1])) +
            geom_vline(aes(xintercept=fences[2], color=fences_labs[2])) +
            labs(color=paste0("Fences ", useFences)) +
            scale_colour_manual(values=fencesColors)
    }
    ggp <- ggp + ggtitle(metric) + theme_bw()

    return(ggp)
}


#' plotPolygons
#' @name plotPolygons
#' @rdname plotPolygons
#'
#' @description Plot polygons from a `SpatialExperiment` object using ggplot2.
#'
#' @param spe A `SpatialExperiment` object with polygon data as an `sf` object.
#' @param colourBy A column in `colData(spe)` for coloring the polygons or a
#' string color in colors(). (Default is "darkgrey")
#' @param colourLog Logical to log-transform the data to enhance visualization
#' (Default is FALSE).
#' @param sampleId Sample ID for plot title. Default is the unique sample ID.
#' @param fillAlpha Transparency level for polygon fill. Default is `1`.
#' @param palette Colors to use if `colourBy` is a factor. Default is `NULL`.
#' @param borderCol Color of polygon borders. Default is `"black"`.
#' @param borderAlpha Transparency level for borders. Default is `1`.
#' @param borderLineWidth Width of polygon borders. Default is `0.1`.
#' @param drawBorders Logical; whether to draw borders. Default is `TRUE`.
#' @param polyColumn character for the name of the column where the polygons sf
#' are stored (default is "polygons.global")
#' @param bgColor character indicating color for the background
#' (default is "white")
#'
#' @return A `ggplot` object representing the polygon plot of the spatial data.
#' @export
#'
#' @importFrom ggplot2 ggplot geom_sf aes scale_fill_manual scale_fill_viridis_c
#' scale_fill_identity theme_minimal theme element_text margin labs
#' @importFrom sf st_as_sf
#' @importFrom SummarizedExperiment colData
#' @importFrom grDevices colors
#'
#' @examples
#' example(readAndAddPolygonsToSPE)
#' plotPolygons(spe, colourBy="Mean.DAPI")
plotPolygons <- function(spe, colourBy="darkgrey", colourLog=FALSE,
    polyColumn="polygons.global", sampleId=unique(spe$sample_id),
    bgColor="white", fillAlpha=1, palette=NULL, borderCol=NA,
    borderAlpha=1, borderLineWidth=0.1, drawBorders=TRUE) {
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot("polygons" %in% names(colData(spe)))
    df <- data.frame(colData(spe))
    polflag <- FALSE
    if(!is.null(colourBy)) {
        if(colourBy %in% names(colData(spe))) {
            if(colourLog) {
                colourByo <- colourBy
                colourBy <- paste0("log(", colourByo, ")")
                df[[colourBy]] <- log1p(df[[colourByo]])
            }
            polflag <- TRUE
        } else {
            if(!(colourBy %in% grDevices::colors())) {
                warning(colourBy, " not in known colors nor in colData")
                colourBy <- "darkgrey"
            }}}
    border_params <- if(drawBorders) {
        list(color=borderCol, size=borderLineWidth)
    } else { list(color=NA, size=0) }
    if(polflag) {
        p <- ggplot(df, aes(geometry=.data[[polyColumn]],
            fill=.data[[colourBy]])) + geom_sf(alpha=fillAlpha,
            color=border_params$color, size=border_params$size)
    } else {
        p <- ggplot(df, aes(geometry = .data[[polyColumn]])) +
            geom_sf(fill=colourBy, alpha=fillAlpha,
                color=border_params$color, size=border_params$size)
    }
    if(!is.null(colourBy) && (is.factor(df[[colourBy]]) ||
                                is.logical(df[[colourBy]]))) {
        if(!is.null(palette)) { p <- p + scale_fill_manual(values=palette) }
    } else if(!is.null(colourBy) && is.numeric(df[[colourBy]])) {
        p <- p + scale_fill_viridis_c(option="D")
    } else { p <- p + scale_fill_identity() }
    p <- p + theme_minimal() + theme(legend.position="right",
        plot.title.position="plot",plot.title=element_text(face="bold",size=14),
            plot.margin=margin(0, 0, 0, 0),
            panel.background=element_rect(fill=bgColor, color=bgColor, linewidth=1),
            panel.border = element_rect(color="black", fill=NA, linewidth=0.1),
            panel.grid.minor=element_blank()
        ) + labs(title=sampleId, fill=colourBy)
    return(p)
}


#' plotZoomFovsMap
#' @name plotZoomFovsMap
#' @rdname plotZoomFovsMap
#' @description
#'
#' Plot Zoomed-in FOVs with Map and Polygons
#'
#' This function generates a plot that shows a map of all fields of view (FOVs)
#' within a `SpatialExperiment` object, alongside a zoomed-in view of the
#' specified FOVs with an overlay of polygons and optional coloring.
#'
#' @param spe A `SpatialExperiment` object containing spatial transcriptomics
#' data.
#' @param fovs A character vector specifying the FOVs to be zoomed in and
#' plotted. Must match values in the `fov` column of `colData(spe)`.
#' @param mapPointCol A character string specifying the color of the points
#' in the map. Default is `"darkmagenta"`.
#' @param mapNumbersCol A character string specifying the color of the
#' numbers on the map. Default is `"black"`.
#' @param mapAlphaNumbers A numeric value specifying the transparency of the
#' numbers on the map. Default is `0.8`.
#' @param title An optional character string specifying the title of the final
#' plot. If `NULL`, no title is added. Default is `NULL`.
#' @param ... Additional arguments passed to `plotPolygons`.
#'
#' @return A combined plot showing a map of all FOVs with zoomed-in views of
#' the specified FOVs and their associated polygons.
#'
#' @details The function first filters the `SpatialExperiment` object to the
#' specified FOVs, generates a plot of the cells for the entire map, then
#' creates a detailed polygon plot of the selected FOVs, and finally combines
#' these into a single side-by-side visualization. If `title` is not `NULL`, it
#' adds a title to the combined plot.
#'
#' @importFrom ggpubr ggarrange annotate_figure
#' @export
#'
#' @examples
#' example(readAndAddPolygonsToSPE)
#' plotZoomFovsMap(spe, fovs=16, title="FOV 16")
plotZoomFovsMap <- function(spe, fovs=NULL,
                            mapPointCol="darkmagenta",
                            mapNumbersCol="black",
                            mapAlphaNumbers=0.8,
                            title=NULL, ...) {
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot("fov" %in% names(colData(spe)))
    stopifnot(all(fovs %in% spe$fov))
    spefovs <- spe[, spe$fov %in% fovs]
    map <- plotCellsFovs(spefovs, pointCol=mapPointCol,
        numbersCol=mapNumbersCol, alphaNumbers=mapAlphaNumbers,
        sampleId=NULL)
    g2 <- plotPolygons(spefovs, sampleId=NULL, ...)
    final_plot <- ggpubr::ggarrange(map, g2, ncol=2)
    if (!is.null(title)) {
        final_plot <- ggpubr::annotate_figure(final_plot,
            top=ggpubr::text_grob(title, face="bold", size=14))
    }
    return(final_plot)
}

#' plotQScoreTerms
#' @name plotQScoreTerms
#' @rdname plotQScoreTerms
#'
#' @description
#' Plots the individual terms that combine into the quality score formula,
#' allowing assessment of each term’s impact on the final score.
#'
#' @param spe A `SpatialExperiment` object with `quality_score` and term
#'   columns in `colData`.
#' @param sampleId Character string for plot title. Must match values in the
#'   `fov` column of `colData(spe)`. Default: `unique(spe$sample_id)`.
#' @param size Numeric point size for the scatter plots. Default: `0.05`.
#' @param alpha Numeric transparency for the scatter plots. Default: `0.2`.
#' @param aspectRatio Numeric aspect ratio of the plots. Default: `1`.
#' @param custom Logical; if `TRUE`, use custom polygon‐derived metrics.
#'
#' @return A combined plot (via `cowplot::plot_grid`) showing spatial maps
#'   of each QS term.
#'
#' @importFrom scater plotColData
#' @importFrom ggplot2 ggtitle coord_fixed
#' @importFrom cowplot plot_grid
#' @export
#'
#' @examples
#' example(readAndAddPolygonsToSPE)
#' example(spatialPerCellQC)
#' p <- plotQScoreTerms(spe)
#' print(p)
plotQScoreTerms <- function(spe, sampleId=unique(spe$sample_id), size=0.05,
                            alpha=0.8, aspectRatio=1, custom = FALSE) {
    if(metadata(spe)$technology=="Nanostring_CosMx") {
        if(custom==TRUE) {
            ggp <- scater::plotColData(spe, x=spatialCoordsNames(spe)[1],
                                        y=spatialCoordsNames(spe)[2],
                                        colour_by="cust_log2CountArea",
                                        point_size=size, point_alpha=alpha)+
                ggplot2::ggtitle(sampleId)+ .centroid_image_theme() +
                ggplot2::coord_fixed()
            ggp2 <- scater::plotColData(spe, x=spatialCoordsNames(spe)[1],
                                        y=spatialCoordsNames(spe)[2],
                                        colour_by="cust_log2AspectRatio",
                                        point_size=size, point_alpha=alpha)+
                ggplot2::ggtitle(sampleId) + .centroid_image_theme() +
                ggplot2::coord_fixed()
        } else {
            ggp <- scater::plotColData(spe, x=spatialCoordsNames(spe)[1],
                                        y=spatialCoordsNames(spe)[2],
                                        colour_by="log2CountArea",
                                        point_size=size, point_alpha=alpha)+
                ggtitle(sampleId)+ .centroid_image_theme() + coord_fixed()
            ggp2 <- scater::plotColData(spe, x=spatialCoordsNames(spe)[1],
                                        y=spatialCoordsNames(spe)[2],
                                        colour_by="log2AspectRatio",
                                        point_size=size, point_alpha=alpha)+
                ggplot2::ggtitle(sampleId) + .centroid_image_theme() +
                ggplot2::coord_fixed()
        }
        ggp3 <- scater::plotColData(spe, x=spatialCoordsNames(spe)[1],
                                    y=spatialCoordsNames(spe)[2],
                                    colour_by="dist_border",
                                    point_size=size, point_alpha=alpha)+
            ggplot2::ggtitle(sampleId) + .centroid_image_theme() +
            ggplot2::coord_fixed()
        ggp <- cowplot::plot_grid(ggp, ggp2, ggp3, ncol = 2)
    } else {
        ## check if column variable is logical to impose our colors
        ggp <- ggp1 <- scater::plotColData(spe, x=spatialCoordsNames(spe)[1],
                                            y=spatialCoordsNames(spe)[2],
                                            colour_by="log2CountArea",
                                            point_size=size, point_alpha=alpha)+
            ggplot2::ggtitle(sampleId)+ .centroid_image_theme() +
            ggplot2::coord_fixed()
    }
    return(ggp)
}



.assign_outlier_color <- function(vals, outlier_mc, high_label, low_label) {
    thr <- round(attr(outlier_mc, "thresholds"), 2)
    dplyr::case_when(
        vals > thr[2] ~ high_label,
        vals < thr[1] ~ low_label,
        TRUE ~ "unflagged"
    )
}

.assign_collapsed_color <- function(is_ctrl, area_col, dapi_col) {
    color <- rep("unflagged", length(is_ctrl))
    color[is_ctrl] <- "ctrl/total ratio > 0.1"
    color[area_col == "> area um higher thr."] <- "> area um higher thr."
    color[area_col == "< area um lower thr."] <- "< area um lower thr."
    color[dapi_col == "> DAPI higher thr."] <- "> DAPI higher thr."
    color[dapi_col == "< DAPI lower thr."] <- "< DAPI lower thr."
    color
}

.make_outlier_plot <- function(polygons, fov, fillvar, pal, title = NULL,
                                leg = FALSE) {
    filtered_polygons <- polygons[polygons$fov %in% fov, ]

    if (nrow(filtered_polygons) == 0 || !fillvar %in% names(filtered_polygons)){
        # Create an empty plot with appropriate theme
        p <- ggplot2::ggplot() +
            ggplot2::theme_void()
        if (!is.null(title)) {
            p <- p + ggplot2::ggtitle(title)
        }
        return(p)
    }
    newpal <- pal[which(names(pal)%in%unique(filtered_polygons[[fillvar]]))]
    p <- ggplot2::ggplot() +
        ggplot2::geom_sf(
            data = filtered_polygons,
            mapping = ggplot2::aes(
                fill  = .data[[fillvar]],
                color = .data[[fillvar]]
            ),
            lwd = 0,
            show.legend = TRUE
        ) +
        ggplot2::scale_fill_manual(values = newpal) +
        ggplot2::scale_color_manual(values = newpal)

    if (!leg) {
        p <- p + ggplot2::theme(legend.position = "none")
    }
    if (!is.null(title)) {
        p <- p + ggplot2::ggtitle(title)
    }
    return(p)
}



#' qcFlagPlots
#' @name qcFlagPlots
#' @rdname qcFlagPlots
#' @description
#' Plots the flagged cells identified with first filter, based on control count
#' on total count ratio, area in um and DAPI signal.
#'
#' This function generates a plot that shows selected (FOVs)
#' within a `SpatialExperiment` object, with cells flagged in different colors
#' over a light or dark layout chosen by the user.
#'
#' @param spe A `SpatialExperiment` object containing spatial transcriptomics
#' data.
#' @param fov An integer or numeric vector specifying the FOVs to be plotted
#' Must match values in the `fov` column of `colData(spe)`.
#' @param theme A character string among "light" or "dark".
#' @param custom A boolean value. If TRUE, custom polygons derived metrics will
#' be used.
#'
#' @return A panel with multiple plots showing flagged cells for different
#' variables.
#'
#' @importFrom dplyr case_when
#' @importFrom ggplot2 ggplot geom_sf aes scale_fill_manual scale_color_manual
#' @importFrom ggplot2 ggtitle theme theme_void
#' @importFrom cowplot get_legend plot_grid
#' @export
#'
#' @examples
#'
#' example(readAndAddPolygonsToSPE)
#' spe <- spatialPerCellQC(spe)
#' spe <- computeThresholdFlags(spe)
#' p <- qcFlagPlots(spe, fov=16, theme="dark")
#' print(p)
qcFlagPlots <- function(spe, fov=unique(spe$fov),
                            theme=c("light","dark"), custom=FALSE) {
    if (!all(c("is_zero_counts", "is_ctrl_tot_outlier") %in%
            names(colData(spe))))
        stop("Fixed thresholds flag cells not found. Run computeFixedFlags().")
    spe <- computeSpatialOutlier(spe, computeBy="Area_um", method="both")
    spe <- computeSpatialOutlier(spe, computeBy="Mean.DAPI", method="both")
    spe$polygons$fixed_flags_color <- dplyr::case_when(
        spe$is_zero_counts ~ "0 counts",
        spe$is_ctrl_tot_outlier ~ "ctrl/total ratio > 0.1",
        TRUE ~ "unflagged"
    )
    if(all(spe$polygons$fixed_flags_color=="unflagged"))
        warning("No 0 counts or control/total ratio > 0.1 found")
    dapi_col <- .assign_outlier_color(spe$Mean.DAPI, spe$Mean.DAPI_outlier_mc,
                                    "> DAPI higher thr.", "< DAPI lower thr.")
    spe$polygons$dapi_outlier_color <- dapi_col
    area_vals <- if (custom) spe$cust_Area_um else spe$Area_um
    area_mc   <- if (custom) spe$cust_Area_um_outlier_mc else
        spe$Area_um_outlier_mc
    area_col  <- .assign_outlier_color(area_vals, area_mc,
        "> area um higher thr.", "< area um lower thr.")
    spe$polygons$area_outlier_color <- area_col
    spe$polygons$collapsed_color <-
        .assign_collapsed_color(spe$is_ctrl_tot_outlier, area_col, dapi_col)
    outlier_palette <- c(
        "unflagged"="#c0c8cf", "ctrl/total ratio > 0.1"="magenta",
        "< area um lower thr."="darkturquoise", "> area um higher thr."="red",
        "< DAPI lower thr."="purple", "> DAPI higher thr."="greenyellow"
    )
    plot_func <- if(theme[1]=="light") .light_theme else .dark_theme
    ggp1 <- .make_outlier_plot(spe$polygons, fov, "fixed_flags_color",
                            outlier_palette, "Control counts ratio") +
        plot_func()
    ggp2 <- .make_outlier_plot(spe$polygons, fov, "area_outlier_color",
                            outlier_palette, "Area in um") + plot_func()
    ggp3 <- .make_outlier_plot(spe$polygons, fov, "dapi_outlier_color",
                            outlier_palette, "Mean DAPI") + plot_func()
    # legp <- .make_outlier_plot(spe$polygons, fov, "collapsed_color",
    #                         outlier_palette, leg=TRUE) + plot_func() +
    #     ggplot2::theme(legend.title=element_blank())
    # ggp4 <- cowplot::get_legend(legp)
    # final <- cowplot::plot_grid(ggp1, ggp2, ggp3, ggp4, ncol=2)
    # if(theme[1]=="dark")
    # final <- final +
    #     ggplot2::theme(panel.background=element_rect(fill="black"))
    # final
    legp <- .make_outlier_plot(spe$polygons, fov, "collapsed_color",
                                outlier_palette, leg = TRUE) +
        plot_func() +
        ggplot2::theme(legend.title = ggplot2::element_blank()) +
        # scale "robuste" per evitare il drop delle classi assenti
        ggplot2::scale_fill_manual(
            values = outlier_palette,
            limits = names(outlier_palette),
            drop   = FALSE,
            na.value = "grey80"
        ) +
        ggplot2::scale_color_manual(
            values = outlier_palette,
            limits = names(outlier_palette),
            drop   = FALSE,
            na.value = "grey80"
        )

    # Estrazione sicura della legenda
    ggp4 <- tryCatch(
        cowplot::get_legend(legp),
        error = function(e) NULL
    )

    # Composizione finale
    if (!is.null(ggp4)) {
        final <- cowplot::plot_grid(ggp1, ggp2, ggp3, ggp4, ncol = 2)
    } else {
        final <- cowplot::plot_grid(ggp1, ggp2, ggp3, ncol = 2)
    }

    if (theme[1] == "dark") {
        final <- final +
            ggplot2::theme(panel.background = ggplot2::element_rect(fill="black"))
    }
    final


}

#' .fov_image_theme
#' @name .fov_image_theme
#' @rdname dot-fov_image_theme
#' @description
#' internal function to setup the theme for the fov background on the whole
#' image
#'
#' @param backColor not used
#' @param backBorder color for the borders of the background (default=NA)
#' @param titleCol character indicating the color of the title
#'
#' @return a ggplot2 theme object
#' @importFrom ggplot2 theme element_blank element_rect element_text
#' @keywords internal
.fov_image_theme <- function(backColor="black", backBorder=NA,
                            titleCol="white")
{
    theme(panel.border=element_blank(),
        legend.key=element_blank(),
        axis.title.x=element_blank(),
        axis.title.y=element_blank(),
        axis.ticks=element_blank(),
        axis.text.y=element_blank(),
        axis.text.x=element_blank(),
        panel.grid=element_blank(),
        panel.grid.minor=element_blank(),
        panel.grid.major=element_blank(),
        panel.background=element_rect(fill="transparent", colour=NA),
        plot.title=element_text(color=titleCol, hjust=0.5, face="bold"),
        plot.background=element_rect(fill="transparent", colour=backBorder))
}

#' .centroid_image_theme
#' @name .centroid_image_theme
#' @rdname dot-centroid_image_theme
#' @description
#' internal function to setup the theme for the centroid plot background
#'
#' @param backBorder color for the borders of the background (default=NA)
#'
#' @return a ggplot2 theme object
#' @importFrom ggplot2 theme element_blank element_rect
#' @keywords internal
.centroid_image_theme <- function(backBorder=NA) {
    theme(panel.border=element_rect(color="black"),
        legend.key=element_blank(),
        axis.line=element_blank(),
        axis.title.x=element_blank(),
        axis.title.y=element_blank(),
        axis.ticks=element_blank(),
        axis.text.y=element_blank(),
        axis.text.x=element_blank(),
        panel.grid=element_blank(),
        panel.grid.minor=element_blank(),
        panel.grid.major=element_blank(),
        panel.background=element_rect(fill="transparent", colour=NA),
        plot.title=element_blank(),
        plot.background=element_rect(fill="transparent", colour=backBorder))
}


#' .negative_image_theme
#' @name .negative_image_theme
#' @rdname dot-negative_image_theme
#' @description
#' internal function to setup the theme for the negative background for
#' negative plots
#'
#' @param fillColor color to fill the element_rect (default is "black")
#' @param foreColor color for all the other elements (default is "white")
#'
#' @return a ggplot2 theme object
#' @importFrom ggplot2 theme element_line element_rect element_text
#' element_blank
#' @keywords internal
.negative_image_theme <- function(fillColor="black", foreColor="white") {
    theme(panel.border = element_rect(color = foreColor),
        panel.background=element_rect(fill=fillColor, color=NA),
        plot.background=element_rect(fill=fillColor, color=NA),
        title=element_blank(),
        axis.line=element_blank(),
        axis.title.x=element_blank(),
        axis.title.y=element_blank(),
        axis.ticks=element_blank(),
        axis.text.y=element_blank(),
        axis.text.x=element_blank(),
        legend.background=element_rect(fill=fillColor, color=NA),
        legend.text=element_text(color=foreColor),
        legend.title=element_text(color=foreColor),
        panel.grid=element_blank(),
        panel.grid.minor=element_blank(),
        panel.grid.major=element_blank())
}

#' .light_theme
#' @name .light_theme
#' @rdname dot-light_theme
#' @description
#' internal function to setup the white background theme for the First Filter
#' plot
#'
#' @return a ggplot2 theme object
#' @importFrom ggplot2 theme element_blank element_rect
#' @keywords internal
.light_theme <- function(fillColor="white", foreColor="black") {
    theme(panel.border=element_rect(color = foreColor, fill = NA,
        linewidth = 0.1),
        panel.background=element_rect(fill=fillColor, color=NA),
        plot.background=element_rect(fill=fillColor, color=NA),
        axis.line=element_blank(),
        axis.title.x=element_blank(),
        axis.title.y=element_blank(),
        axis.ticks=element_blank(),
        axis.text.y=element_blank(),
        axis.text.x=element_blank(),
        panel.grid=element_blank(),
        panel.grid.minor=element_blank(),
        panel.grid.major=element_blank())
}

#' .dark_theme
#' @name .dark_theme
#' @rdname dot-dark_theme
#' @description
#' internal function to setup the black background theme for the First Filter
#' plot
#' @param fillColor color to fill the element_rect (default is "black")
#' @param foreColor color for all the other elements (default is "white")
#'
#' @return a ggplot2 theme object
#' @importFrom ggplot2 theme element_blank element_rect element_text
#' @keywords internal
.dark_theme <- function(fillColor="black", foreColor="white") {
    theme(panel.background=element_rect(fill=fillColor, color=NA),
        plot.background=element_rect(fill=fillColor, color=NA),
        panel.border=element_rect(color=foreColor, fill=NA, linewidth=0.1),
        title=element_text(color=foreColor),
        axis.line=element_blank(),
        axis.title.x=element_blank(),
        axis.title.y=element_blank(),
        axis.ticks=element_blank(),
        axis.text.y=element_blank(),
        axis.text.x=element_blank(),
        legend.background=element_rect(fill=fillColor, color=NA),
        legend.text=element_text(color=foreColor),
        panel.grid=element_blank(),
        panel.grid.minor=element_blank(),
        panel.grid.major=element_blank())
}

#' createPaletteFromColData
#' @name createPaletteFromColData
#' @rdname createPaletteFromColData
#' @description
#' Create a Palette from colData in a SpatialExperiment Object
#'
#' This function generates a palette mapping based on specified columns in
#' the `colData` of a `SpatialExperiment` object.
#'
#' @param spe A `SpatialExperiment` object with spatial transcriptomics data.
#' @param paletteNames A character string specifying the column in
#' `colData(spe)` to be used for the names in the palette.
#' @param paletteColors A character string specifying the column in
#' `colData(spe)` to be used for the colors in the palette.
#'
#' @return A character vector representing the palette mapping, where each
#' element is a string in the format `"name=color"`.
#'
#' @details The function creates a new palette based on the unique combinations
#' of values in the specified `paletteNames` and `paletteColors` columns in
#' `colData(spe)`.
#'
#' @importFrom SummarizedExperiment colData
#' @keywords internal
createPaletteFromColData <- function(spe, paletteNames, paletteColors)
{
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot(all(c(paletteNames, paletteColors) %in% names(colData(spe))))

    tb <- table(spe[[paletteNames]], spe[[paletteColors]])
    newpal <- colnames(tb)[which(tb!=0, arr.ind=TRUE)[,2]]
    names(newpal) <- rownames(tb)[which(tb!=0, arr.ind=TRUE)[,1]]
    return(newpal)
}

#' firstFlagPalette
#' @name firstFlagPalette
#' @rdname firstFlagPalette
#' @description
#' neon color palette for firstFlagPlot
#'
#' @return a palette for firstFlagPlot
#' @keywords internal
firstFlagPalette <- c(
    "unflagged" = "#7f7f7f",
    "ctrl/total ratio > 0.1"     = "magenta",
    "< area um lower thr."       = "darkturquoise",
    "> area um higher thr."      = "red",
    "< logged aspect ratio lower thr."  = "purple",
    "> logged aspect ratio higher thr." = "greenyellow"
)

#' spatialPerCellQC
#' @name spatialPerCellQC
#' @rdname spatialPerCellQC
#' @description
#' Computes quality‐control metrics for each cell and adds them to `colData`.
#'
#' @param spe A `SpatialExperiment` object containing spatial omics data.
#' @param micronConvFact Numeric factor to convert pixels to microns. Default
#' is 0.12.
#' @param rmZeros logical for removing zero total counts (protein intensities) 
#' cells (default is TRUE).
#' @param negProbList Character vector of patterns to identify negative probes.
#' Defaults include:
#'   Nanostring CosMx: `"NegPrb"`, `"Negative"`, `"SystemControl"`
#'   Xenium: `"NegControlProbe"`, `"NegControlCodeword"`, `"UnassignedCodeword"`
#'   MERFISH: `"Blank"`
#' @param use_altexps logical for `use_altexps` in `scuttle` package.
#' If TRUE uses the altexps for computing some metrics on it.
#' Useful for interoperability with `SpatialExperimentIO`.
#' (See \link[scuttle]{addPerCellQC} for additional details).
#'
#' @return A `SpatialExperiment` object with added QC metrics in `colData`.
#'
#' @details
#' The function computes and appends per‑cell QC metrics to `colData(spe)` and
#' may subset the returned SpatialExperiment.
#'
#' Key behaviours and expectations:
#' - Feature detection: negative‑probe patterns supplied in `negProbList` are
#' passed to `scater::addPerCellQC()` (via `use_altexps` when requested). 
#' `addPerCellQC()` must be able to find matching control feature names 
#' (`rownames` of spe) and will create `subsets_*_sum` and `subsets_*_detected` 
#' columns used below for each of control feature types.
#' 
#' The function appends `sum`, `detected` and `total` to the SPE colData by
#' calling internally `addPerCellQC()`; these are used to compute `control_sum`, 
#' `control_detected`, `target_sum` and `target_detected.`
#' 
#' - Control metrics: `control_sum` and `control_detected` are computed by
#' summing matching `subsets_*` columns; `target_*` metrics are computed as
#' sum(detected) - control_sum(_detected).
#' 
#' - Ratios and logs: `ctrl_total_ratio` (`control_sum`/`sum`) and its stabilized
#' log2-transformed `log2Ctrl_total_ratio` are added.
#' 
#' - Coordinate and area handling:
#' - For CosMx technologies (Nanostring_CosMx and Nanostring_CosMx_Protein)
#' spatial coordinates are converted from pixels to microns using
#' `micronConvFact` and appended to `colData` (column names have px -> um).
#' 
#' - For CosMx, `Area_um` is derived from an existing Area column scaled by
#' micronConvFact^2 and `.computeBorderDistanceCosMx()` is invoked to
#' compute `dist_border`.
#' 
#' - For Nanostring_CosMx_Protein, a legacy `Area.um2` column (if present)
#' is renamed to `Area_um` to standardize naming.
#' 
#' - For Xenium (10X_Xenium), if `Area_um` is missing the function will
#' attempt to use `cell_area` as a fallback and issue a warning.
#' 
#' - Aspect ratio: if `AspectRatio` exists it is logged (`log2AspectRatio`); if
#' missing a warning is emitted.
#' 
#' - Signal density: `SignalDensity` is computed as `sum` / `Area_um` for most
#' technologies; for Nanostring_CosMx_Protein it coincides with `total`. A
#' log2-transformed `log2SignalDensity` is also added.
#' 
#' - Zero‑count (protein intensity) removal: when `rmZeros = TRUE` cells with 
#' `sum == 0` are removed from the returned SpatialExperiment (message printed).
#' 
#' - Side effects: the function modifies `colData(spe)` (adds multiple new
#' columns). It may add spatial coordinates into `colData` if missing, and may
#' subset the SPE to remove zero‑count cells. It issues warnings when expected
#' inputs (e.g. area, aspect ratio, polygon‑derived fields) are missing or
#' when fallbacks are used.
#'
#' Use this information to ensure the input SpatialExperiment contains the
#' necessary assays and fields (feature names, `sum`/`detected`/`total`,
#' `Area`/`cell_area`, log2AspectRatio, dist_border when available) so 
#' metrics are computed and assigned correctly.
#'
#' @importFrom SummarizedExperiment colData
#' @importFrom scater addPerCellQC
#' @importFrom S4Vectors cbind.DataFrame
#' @export
#' @examples
#' example(readCosmxSPE)
#' spe <- spatialPerCellQC(spe)
spatialPerCellQC <- function(spe, micronConvFact=0.12, rmZeros=TRUE,
    negProbList=c("NegPrb", "Negative", "SystemControl", "Ms IgG1", "Rb IgG",
    "BLANK_", "NegControlProbe", "NegControlCodeword", "UnassignedCodeword",
    "Blank"),
    use_altexps=NULL) {
    stopifnot(is(object=spe, "SpatialExperiment"))
    idxlist <- lapply(negProbList, function(ng) {
        grep(paste0("^", ng), rownames(spe))
    })
    names(idxlist) <- negProbList
    idxlist <- idxlist[which(lengths(idxlist)!=0)]
    spe <- addPerCellQC(spe, subsets=idxlist, use_altexps=use_altexps)
    idx <- grep("^subsets_.*_sum$", colnames(colData(spe)))
    npc <- npd <- 0
    if ( length(idx) !=0 ) {
        npc <- rowSums(as.matrix(colData(spe)[ , idx, drop=FALSE]))
        # TODO: not robust at all! the +1 is not a really good choice
        npd <- rowSums(as.matrix(colData(spe)[ , idx+1, drop=FALSE]))
    }
    spe$control_sum <- npc
    spe$control_detected <- npd
    spe$target_sum <- spe$sum - npc
    spe$target_detected <- spe$detected - npd
    if(!all(spatialCoordsNames(spe) %in% names(colData(spe)))) {
        # TODO: CHANGE SPE constructor WITH COORDINATES IN COLDATA
        colData(spe) <- cbind.DataFrame(colData(spe), spatialCoords(spe))
    }

    spe$ctrl_total_ratio <- spe$control_sum/spe$total
    spe$ctrl_total_ratio[which(is.na(spe$ctrl_total_ratio))] <- 0
    # using eps avoids NaNs and 0 in further computations
    eps <- 0.0001
    spe$log2Ctrl_total_ratio <- log2(spe$ctrl_total_ratio+eps)
    if(metadata(spe)$technology == "Nanostring_CosMx_Protein") {
        # Only for proteins will be included in QScore
        idx <- which(names(colData(spe)) == "Area.um2")
        if(length(idx)!=0) { names(colData(spe))[idx] <- "Area_um" }
    }

    if(any(metadata(spe)$technology %in%
            c("Nanostring_CosMx", "Nanostring_CosMx_Protein"))) {
        spnc <- spatialCoords(spe) * micronConvFact
        colnames(spnc) <- gsub("px", "um", spatialCoordsNames(spe))
        colData(spe) <- cbind.DataFrame(colData(spe), spnc)
        spe$Area_um <- spe$Area * (micronConvFact^2)
        spe <- .computeBorderDistanceCosMx(spe)
    }

    if (metadata(spe)$technology == "10X_Xenium") {
        if(!"Area_um" %in% colnames(colData(spe))) {
            warning("Missing Area_um in colData for Xenium data.\n",
                "Computing from cell_area column.")
            spe$Area_um <- spe$cell_area # standardized across other techs
        }
    }
    if ("AspectRatio" %in% colnames(colData(spe))) {
        spe$log2AspectRatio <- log2(spe$AspectRatio) # not cosmx
    } else { warning("Missing aspect ratio in colData") }

    spe$SignalDensity <- spe$sum/spe$Area_um

    if (metadata(spe)$technology == "Nanostring_CosMx_Protein") {
        spe$SignalDensity <- spe$total
    }

    spe$log2SignalDensity <- log2(spe$SignalDensity)
    if (rmZeros) {
        if (sum(spe$sum==0) > 0) {
            message("Removing ", dim(spe[,spe$sum==0])[2],
                    " cells with 0 counts!")
            spe <- spe[,!spe$sum==0]
        }
    }
    return(spe)
}

#' .computeBorderDistanceCosMx
#' @name .computeBorderDistanceCosMx
#' @rdname dot-computeBorderDistanceCosMx
#' @description
#' Calculates the minimum distance of each cell to the Field‐Of‐View (FOV) 
#' border and adds it to `colData`.
#'
#' @param spe A `SpatialExperiment` object with CosMx data.
#' @param xwindim Width of FOV along x axis (default from `metadata(spe)$fov_dim`).
#' @param ywindim Height of FOV along y axis (default from `metadata(spe)$fov_dim`).
#'
#' @return A `SpatialExperiment` object with `dist_border` column in
#' `colData`.
#'
#' @importFrom dplyr left_join
#' @importFrom SummarizedExperiment colData
#' @importFrom S4Vectors metadata
#' @keywords internal
.computeBorderDistanceCosMx <- function(spe,
                                    xwindim=metadata(spe)$fov_dim[["xdim"]],
                                    ywindim=metadata(spe)$fov_dim[["ydim"]]) {
    stopifnot(is(spe, "SpatialExperiment"))
    cd <- colData(spe)
    cdf <- left_join(as.data.frame(cd), metadata(spe)$fov_positions, by="fov")
    spcn <- spatialCoordsNames(spe)
    fovpn <- colnames(metadata(spe)$fov_positions)[colnames(
        metadata(spe)$fov_positions) %in% c("x_global_px", "y_global_px")]
    cd$dist_border_x <- pmin(cdf[,spcn[1]] - cdf[,fovpn[1]],
                            (cdf[,fovpn[1]] + xwindim) - cdf[,spcn[1]])
    cd$dist_border_y <- pmin(cdf[,spcn[2]] - cdf[,fovpn[2]],
                            (cdf[,fovpn[2]] + ywindim) - cdf[,spcn[2]])
    cd$dist_border <- pmin(cd$dist_border_x, cd$dist_border_y)
    colData(spe) <- cd
    return(spe)
}

#' computeSpatialOutlier
#' @name computeSpatialOutlier
#' @rdname computeSpatialOutlier
#' @description
#' Computes outliers for quantitative cell metadata based on a statistical test:
#' medcouple (`mc` method argument) or the Median Absolute Deviation (MAD, `scuttle` 
#' method argument), when the metric distribution is respectively asymmetric or symmetric.
#'
#' @details
#' The medcouple method is a measure for the skeweness of univariate
#' distribution as described in Hubert M. et al. (2008).
#' In particular, the computed medcouple value must be in a range between -0.6
#' and 0.6 to compute adjusted boxplots and perform the outlier detection.
#' For MAD method we just wrap the isOutlier function in the scuttle package. 
#' Please see McCarthy DJ et al (2017) for further details.
#'
#' @param spe a SpatialExperiment object with quantitative cell metrics in the `colData`.
#' @param computeBy character indicating a `colData` column name to compute the outliers on.
#' @param method one of `mc`, `scuttle`, `both`.
#' Use `mc` for medcouple, `scuttle` for MAD as computed in `scuttle`, `both` 
#' for computing both of them.
#' @param mcDoScale logical indicating if the metric values should be scaled 
#' to compute medcouple (default is FALSE, as suggested by the original medcouple 
#' authors). See \link[robustbase]{mc} for further readings.
#' @param scuttleType One of `"both"`, `"lower"`, `"higher"` for scuttle method.
#'
#' @return a SpatialExperiment object with additional column(s) (named as
#' the column name indicated in `computeBy` followed by the outlier_sc/mc).
#' Resulting columns are `outlier.filter` logical class objects, allowing to store 
#' the thresholds as attributes of the columns.
#' Use attr(,"thresholds") to retrieve them.
#'
#' @export
#' @importFrom robustbase mc adjbox
#' @importFrom e1071 skewness
#' @importFrom scuttle isOutlier outlier.filter
#'
#' @examples
#' example(spatialPerCellQC)
#' spe <- computeSpatialOutlier(spe, computeBy="log2SignalDensity", method="both")
#' table(spe$log2SignalDensity_outlier_mc)
#' table(spe$log2SignalDensity_outlier_sc)
computeSpatialOutlier <- function(spe, computeBy=NULL,
    method=c("mc", "scuttle", "both"), mcDoScale=FALSE,
    scuttleType=c("both", "lower", "higher")) {
    stopifnot(all(is(spe, "SpatialExperiment"), !is.null(computeBy)))
    stopifnot(computeBy %in% names(colData(spe)))
    options(mc_doScale_quiet=TRUE)
    method <- match.arg(method)
    scuttleType <- match.arg(scuttleType)
    cd <- colData(spe)
    cdcol <- cd[[computeBy]]
    mcfl <- scuttlefl <- FALSE
    switch(method, both={ mcfl <- scuttlefl <- TRUE },
            mc={ mcfl <- TRUE }, scuttle={ scuttlefl <- TRUE },
            {stop("Method is not one of allowed methods")} )
    if (mcfl) {
        skw <- e1071::skewness(cdcol, na.rm = TRUE) # NAs arise problems
        if (skw >- 1 & skw < 1) warning("Distribution is symmetric: ",
                "mc is for asymmetric distributions. Use scuttle instead.")
        mcval <- robustbase::mc(cdcol, doScale=mcDoScale, na.rm=TRUE)
        if ( any( (mcval <= -0.6), (mcval >= 0.6) ) )
            stop("mc is: ",round(mcval, digits=4),"outliers reqs not satisfied")
        names(cdcol) <- colnames(spe)
        outl <- robustbase::adjbox(cdcol, plot=FALSE)
        outsmc <- rep("NO", dim(cd)[1])
        outsmc[rownames(cd) %in% names(outl$out)] <-
            ifelse(outl$out <= outl$fence[1], "LOW", "HIGH")
        outlier_mc <- scuttle::outlier.filter(outsmc) #using scuttle class
        thrs <- as.numeric(outl$fence)
        names(thrs) <- c("lower", "higher")
        attr(outlier_mc, "thresholds") <- thrs
        cd$outlier_mc <- outlier_mc
        names(cd)[names(cd) =="outlier_mc"] <- paste0(computeBy, "_outlier_mc")
        # TODO: compute distributions in the adjusted boxplots to store in cd
    }
    if (scuttlefl) {
        outssc <- scuttle::isOutlier(cdcol, type=scuttleType)
        sctri <- rep("NO", dim(cd)[1])
        sctri <- ifelse(outssc == TRUE & cdcol <= attr(outssc, "thresholds")[1],
                        "LOW", sctri)
        outlier_sc <- ifelse(outssc == TRUE &
                            cdcol >= attr(outssc, "thresholds")[2],
                            "HIGH", sctri)
        outlier_sc <- scuttle::outlier.filter(outlier_sc)
        attr(outlier_sc, "thresholds") <- attr(outssc, "thresholds")
        cd$outlier_sc <- outlier_sc
        names(cd)[names(cd)=="outlier_sc"] <- paste0(computeBy, "_outlier_sc")
    }
    colData(spe) <- cd
    return(spe)
}

#' computeThresholdFlags
#' @name computeThresholdFlags
#' @rdname computeThresholdFlags
#' @description
#' Computes flagged cells using fixed thresholds for SpatialExperiment.
#'
#' This function calculates flagged cells only for total counts (protein intensities) 
#' and control-to-total probe counts ratio using fixed thresholds for a `SpatialExperiment`
#' object.
#'
#' @param spe A `SpatialExperiment` object with spatial omics data.
#' @param totalThreshold A numeric value for the threshold of total counts to
#' identify cells with low counts. Default is `0`.
#' @param ctrlTotRatioThreshold A numeric value for the threshold of
#' control-to-total ratio to flag cells over a certain threshold. Default is
#' `0.1`.
#'
#' @return The `SpatialExperiment` object with added flagged cells as TRUE/FALSE
#'  in `colData`.
#'
#' @details The function flags cells basing on zero counts and control-to-total
#' ratio to identify junk cells.
#' It also combines these two flags into a single flag metric named `threshold_flags`.
#'
#' @importFrom SummarizedExperiment colData
#' @export
#' @examples
#' example(readCosmxSPE)
#' spe <- spatialPerCellQC(spe)
#' spe <- computeThresholdFlags(spe)
#' table(spe$threshold_flags)
computeThresholdFlags <- function(spe, totalThreshold=0,
                            ctrlTotRatioThreshold=0.1)
{
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot("total" %in% names(colData(spe)))
    stopifnot("ctrl_total_ratio" %in% names(colData(spe)))

    spe$is_zero_counts <- ifelse(spe$total == totalThreshold, TRUE, FALSE)
    #flagging cells with probe counts on total counts ratio > 0.1
    spe$is_ctrl_tot_outlier <- ifelse(spe$ctrl_total_ratio >
                                        ctrlTotRatioThreshold, TRUE, FALSE)

    spe$threshold_flags <- (spe$is_ctrl_tot_outlier &
                                spe$is_zero_counts)
    return(spe)
}


#' computeLambda
#' @description
#' Computes optimal ridge regularization parameter \eqn{\lambda} via
#' k-fold cross-validation.
#'
#' \code{computeLambda} performs ridge (L2) logistic regression with
#' k-fold cross-validation to identify the optimal regularization parameter
#' \eqn{\lambda} for Quality Score (QS) model training.
#'
#' @param trainDF  `data.frame`
#'   A data frame for QS model training that must include:
#'     Predictor columns: All columns referenced in the formula returned by `getModelFormula()`.
#'     `QScore_train` or historical `qcscore_train`, containing the binary
#'     (0/1) response to be modeled.
#' @param modelFormula A one-sided formula or character string describing the
#'   supported predictor terms.
#'
#' @return
#' `numeric`
#'   The value of \eqn{\lambda} (i.e., `lambda.min`) from
#'   `cv.glmnet` that minimizes the cross-validation error.
#'
#' @details
#' Complete cases for the formula variables and response are retained. The
#' function constructs a model matrix and passes it to an internal
#' matrix-based implementation that runs ridge logistic-regression
#' cross-validation using `cv.glmnet` with `alpha = 0`.
#'
#' @examples
#' example(computeTrainDF)
#' modform <- getModelFormula(names(metadata(spe)$formula_variables))
#' best_lambda <- computeLambda(df_train, modform)
#' print(best_lambda)
#'
#'
#' @export

computeLambda <- function(trainDF, modelFormula) {
    stopifnot(is.data.frame(trainDF))
    model_formula <- stats::as.formula(modelFormula)
    response <- .getQScoreResponse(trainDF)
    train_ok <- .filterCompleteModelCases(
        df=trainDF,
        modelFormula=model_formula,
        response=response,
        context="training cells for lambda selection"
    )
    train_df <- trainDF[train_ok, , drop=FALSE]
    response <- response[train_ok]
    model_matrix <- stats::model.matrix(
        model_formula,
        data=train_df
    )
    .computeLambda(modelMatrix=model_matrix, response=response)
}

.computeLambda <- function(modelMatrix, response) {
    ridge_cv <- glmnet::cv.glmnet(
        modelMatrix,
        response,
        family="binomial",
        alpha=0,
        lambda=NULL
    )
    ridge_cv$lambda.min
}

#' computeQScore
#' @name computeQScore
#' @rdname computeQScore
#' @description
#' Computes Quality Score (QS) and automatically define weights of the QS model
#' terms through glm training. This function computes QS with a formula
#' defined on the metrics "log2SignalDensity", "Area_um",
#' "log2AspectRatio", "log2Ctrl_total_ratio", as computed from
#' the `spatialPerCellQC` function.
#' It automatically computes the number of outliers for each available
#' metric, which are needed for the glm training.
#' See Details for further information.
#'
#' @details For all the technologies, the QS formula can contain
#' only the following terms at most:
#'
#' Quality Score ~ signal density - size - border effect - background signal 
#'
#' Where signal density is the total counts/protein intensities-to-size ratio, 
#' border effect is represented by the aspect ratio (i.e. width/height of the cell) 
#' and is considered only for cells with a distance from the FOV border less than 
#' 50 px and only for CosMx datasets. Background signal is represented by the ratio 
#' of control probe counts to total counts (protein intensities) and cell size is 
#' the area for CosMx and Xenium, while it is the volume for Merfish.
#' For each couple of variables interaction terms are computed.
#'
#' Note that the function is responsible for automatically including/excluding
#' metrics in the formula based on their availability in the `colData` of the
#' `SpatialExperiment` object.
#'
#' For the default formula, inclusion of metrics also depends on the number of
#' available outliers. If the number of outliers for a metric is less than
#' 0.1\% of the dataset, that metric is excluded from the default QS formula.
#'
#' - Model fitting: ridge (L2) logistic regression is fitted (via `glmnet`) on
#' the balanced training set. The function uses `trainModel()` for fitting
#' and `computeLambda()` to select lambda unless `bestLambda` is supplied.
#'
#' - Lambda details: because of the randomness in the training set selection,
#' results may vary. The easiest way to ensure reproducibility is to let lambda 
#' be computed internally, just set a seed with `set.seed()` before running 
#' `computeQScore`. Otherwise, a fixed value of lambda previously computed with 
#' `computeLambda` preceeded by `computeTrainDF` and `getModelFormula` can be set.
#' 
#' - Model formula details: the most complete default formula is:
#' `~(log2SignalDensity + Area_um + I(abs(log2AspectRatio) * as.numeric(dist_border < 50)) + log2Ctrl_total_ratio)^2`.
#' A user-supplied formula is fitted as supplied: additive structures, selected
#' interactions, `:` and `*` are preserved. The formula must include
#' `log2SignalDensity` and may use any subset of the other supported predictors:
#' `Area_um`,
#' `log2AspectRatio`, and `log2Ctrl_total_ratio`. Arbitrary predictors and
#' transformations are rejected. The supported CosMx-only border expression is
#' `I(abs(log2AspectRatio) * as.numeric(dist_border < 50))`.
#'
#' When a custom `modelFormula` is supplied, the supported base QC metrics in
#' that formula are also used to compute outliers, define good and bad training
#' cells, and construct `QScore_train`. Interactions do not create additional
#' training metrics. In the supported border expression, `dist_border` is an
#' auxiliary variable and `log2AspectRatio` is the corresponding training QC
#' metric. Supported metrics omitted from the custom formula do not contribute
#' to training-label construction. When `modelFormula=NULL`, the established
#' default metric-selection and training-label workflow is unchanged.
#'
#' The computed model output is stored in `metadata(spe)$QScore_model`.
#' For inspection and model coefficient transfer between datasets
#' see also \code{\link{applyQScoreModel}}. However, transferring models between
#' datasets is not recommended. As described in the paper, QS model training is
#' dataset-specific and does not generalize well across datasets. Moreover,
#' this step is computationally efficient. Please, refer to the paper for  
#' details on scalability and coefficient transfer.
#'
#' @param spe A `SpatialExperiment` object with spatial omics data.
#' @param verbose logical for having a verbose output. Default is FALSE.
#' @param bestLambda the best lambda typically computed using `computeLambda`.
#' @param modelFormula A one-sided formula or character string. If `NULL`, the
#'   default formula is generated from available supported metrics and their
#'   outliers. If supplied, its supported terms and interactions are fitted
#'   without rewriting. See Details.
#' @return The `SpatialExperiment` object with added Quality Score in `colData`.
#' @export
#' @importFrom dplyr case_when filter mutate distinct pull
#' @importFrom glmnet glmnet cv.glmnet
#' @importFrom stats as.formula model.matrix quantile predict coef terms
#' @examples
#' example(spatialPerCellQC)
#' set.seed(1998)
#' spe <- computeQScore(spe)
#' summary(spe$QScore)
computeQScore <- function(spe, bestLambda=NULL, modelFormula=NULL, verbose=FALSE) {
    stopifnot(is(spe, "SpatialExperiment"))
    if (dim(spe[,spe$total == 0])[2] != 0) {
        warning(paste0(dim(spe[,spe$total == 0])[2],
            " cells with 0 counts were found. These cells will be removed."))
        spe <- spe[,spe$total > 0]
    }
    formula_info <- NULL
    if (is.null(modelFormula)) {
        metric_list <- intersect(
            .qscoreSupportedPredictors(),
            names(colData(spe))
        )
    } else {
        formula_info <- .validateQScoreFormula(
            modelFormula=modelFormula,
            dataNames=names(colData(spe)),
            technology=metadata(spe)$technology
        )
        metric_list <- formula_info$training_metrics
    }

    if (!"log2SignalDensity" %in% metric_list) {
        stop(
            "'log2SignalDensity' is required to construct Quality Score ",
            "training labels."
        )
    }

    ctx <- .prepQCContext(spe, metric_list, verbose)
    df <- ctx$df; out_var <- ctx$out_var; tech <- ctx$tech

    if (is.null(modelFormula)) {
        model_formula <- getModelFormula(names(out_var))
        model_formula_object <- stats::as.formula(model_formula)
    } else {
        model_formula <- formula_info$text
        model_formula_object <- formula_info$formula
    }

    if (verbose) {
        message("Using final model formula:")
        message(model_formula)
    }

    train_df <- computeTrainDF(df, out_var, tech, verbose)

    train_ok <- .filterCompleteModelCases(
        df=train_df,
        modelFormula=model_formula_object,
        response=train_df$QScore_train,
        context="training cells"
    )

    train_df <- train_df[train_ok, , drop=FALSE]

    model_matrix <- stats::model.matrix(
        model_formula_object,
        data=train_df
    )
    model <- trainModel(model_matrix, train_df)
    
    if(is.null(bestLambda)) {
        bestLambda <- .computeLambda(
            modelMatrix=model_matrix,
            response=train_df$QScore_train
        )
    }

    coefs <- coef(model, s=bestLambda)

    if (verbose) {
        message("Model coefficients for every term used in the formula:")
        message( paste( rownames(coefs), round(as.numeric(coefs), 2), sep="=",
            collapse=" "))
    }

    full_ok <- .filterCompleteModelCases(
        df=df,
        modelFormula=model_formula_object,
        context="cells")

    full_matrix <- stats::model.matrix(
        model_formula_object,
        data=df[full_ok, , drop=FALSE]
    )

    ## NAs may come from model variables used in `model_formula`, e.g. cells with
    ## missing `log2AspectRatio` when aspect ratio cannot be computed from polygons.
    ## Predict only complete cases; incomplete cells keep QScore = NA.
    qscore <- rep(NA_real_, nrow(df))
    qscore[full_ok] <- as.vector(
        predict(model, s=bestLambda, newx=full_matrix, type="response")
    )

    spe$QScore <- qscore

    metadata(spe)$QScore_model <- list(
        model=model,
        bestLambda=bestLambda,
        model_formula=model_formula,
        formula_variables=out_var,
        model_matrix_colnames=colnames(model_matrix),
        coefficients=coefs,
        coefficients_table=data.frame(
            term=rownames(coefs),
            coefficient=as.numeric(coefs),
            row.names=NULL
        )
    )
    return(spe)
}

#' trainModel
#' @name trainModel
#' @rdname trainModel
#' @description
#' Fits a ridge logistic regression model.
#'
#' \code{trainModel} fits an L2-regularized (ridge) logistic regression
#' using \pkg{glmnet}, given a design matrix and a training data frame.
#'
#' @param trainDF `data.frame`
#'   A data frame containing `QScore_train` or historical `qcscore_train`,
#'   coded as 0/1. If both are present, they must agree.
#' @param modelMatrix a matrix describing the model variables, tipically created
#' with `getModelFormula` and `model.matrix` functions.
#'
#' @return
#' A \code{\link[glmnet]{glmnet}} model object fitted with
#' \code{family="binomial"}, \code{alpha=0} (ridge), and a sequence of
#' \eqn{\lambda} values.
#'
#' @export
#' @examples
#' example(computeTrainDF)
#' model_formula <- getModelFormula(metadata(spe)$formula_variables)
#' model_matrix <- model.matrix(as.formula(model_formula), data=df_train)
#' fit <- trainModel(model_matrix, df_train)
#' coef(fit, s = 0.01)
trainModel <- function(modelMatrix, trainDF)
{
    response <- .getQScoreResponse(trainDF)
    model <- glmnet(x=modelMatrix, y=response,
                    family="binomial", lambda=NULL, alpha=0)
    return(model)
}

#' computeTrainDF
#'
#' @rdname computeTrainDF
#'
#' @description
#' Builds a balanced training data frame from a SpatialExperiment.
#'
#' \code{computeTrainDF} takes a \code{SpatialExperiment} object
#' and assembles a balanced training set of “good” vs “bad” cells for
#' subsequent model fitting.
#' @param colData A per-cell metadata table. Typically
#'   `as.data.frame(colData(spe))`. Must include at least:
#'   `cell_id`, raw metric columns named as in `formulaVars` (e.g.
#'   `log2SignalDensity`, `Area_um`, `log2Ctrl_total_ratio`,
#'   optionally `log2AspectRatio` and `dist_border`), and the corresponding outlier-label
#'   columns referenced by `formulaVars`.
#' @param formulaVars A named character vector mapping variable name to
#'   its outlier label column name, e.g.
#'   `c(log2SignalDensity="log2SignalDensity_outlier_train", ...)`.
#' @param tech Character string with the acquisition technology. Used to
#'   enable CosMx-specific inclusion of border effect. Expected
#'   values include `"Nanostring_CosMx"` or `"Nanostring_CosMx_Protein"`.
#' @param verbose Logical; print progress messages.
#'
#' @param verbose `logical(1)` (default \code{FALSE})
#'   If \code{TRUE}, prints the number of “bad” and “good” cells selected.
#'
#' @return
#' A \code{data.frame} with one row per cell, including:
#'   \code{QScore_train} and its compatibility alias
#'   \code{qcscore_train} (0/1) indicating “bad” vs “good”,
#'   relevant \code{colData} columns used for modeling.
#'   Deduplicates and down-samples “good” cells to match the number of “bad” cells.
#'
#' @details The function builds a training set using the variables specified
#' in the `metadata` of the `SpatialExperiment` object.
#'
#' @importFrom SummarizedExperiment colData
#' @importFrom dplyr filter mutate distinct pull
#' @importFrom glmnet glmnet cv.glmnet
#' @importFrom stats as.formula model.matrix quantile predict
#'
#' @examples
#' example(spatialPerCellQC)
#' spe <- computeOutliersQScore(spe)
#' spe <- checkOutliers(spe)
#' df_train <- computeTrainDF(colData(spe), metadata(spe)$formula_variables,
#'     metadata(spe)$technology)
#' table(df_train$QScore_train)
#'
#' @export
computeTrainDF <- function(colData, formulaVars, tech, verbose=FALSE) {
    out_var <- formulaVars
    df <- as.data.frame(colData)
    train_bad_var <- character()
    train_good_var <- character()

    stopifnot("log2SignalDensity is not included in the Quality Score formula.\nQuality Score cannot be computed"="log2SignalDensity" %in% names(out_var))

    cfg <- list(
        log2SignalDensity=list(bad="LOW", good=c(0.90,0.99)),
        Area_um=list(bad="HIGH", good=c(0.25,0.75)),
        log2Ctrl_total_ratio=list(bad="HIGH", good=NULL)
    )

    vars <- intersect(names(cfg), names(out_var))

    for (v in vars) {
        r <- cfg[[v]]
        out_col <- out_var[names(out_var) == v]
        if (length(out_col) == 0L) next

        # BAD ids
        bad_ids <- df |>
            dplyr::filter(.data[[out_col]] == r$bad) |>
            dplyr::pull(cell_id)
        train_bad_var <- unique(c(train_bad_var, bad_ids))

        # GOOD ids (if a band is defined)
        if (!is.null(r$good)) {
            qs <- stats::quantile(df[[v]], probs=r$good)
            good_ids <- df |>
                dplyr::filter(.data[[v]] > qs[1] & .data[[v]] < qs[2]) |>
                dplyr::pull(cell_id)
            train_good_var <- unique(c(train_good_var, good_ids))
        }
    }

    # CosMx-specific handling for log2AspectRatio
    is_cosmx <- tech %in% c("Nanostring_CosMx", "Nanostring_CosMx_Protein")

    if (is_cosmx && "log2AspectRatio" %in% names(out_var)) {

        if (!("dist_border" %in% colnames(df))) {
            stop("dist_border column is required for CosMx handling")
        }

        out_col <- out_var[names(out_var) == "log2AspectRatio"]

        bad_ids <- df |>
            dplyr::filter(.data[[out_col]] %in% c("HIGH","LOW") &
                            dist_border < 50) |>
            dplyr::pull(cell_id)
        train_bad_var <- unique(c(train_bad_var, bad_ids))

        qs <- stats::quantile(df$log2AspectRatio, probs=c(0.25,0.75))
        good_ids <- df |>
            dplyr::filter(log2AspectRatio > qs[1] &
                            log2AspectRatio < qs[2] &
                            dist_border > 50) |>
            dplyr::pull(cell_id)
        train_good_var <- unique(c(train_good_var, good_ids))

        idx <- grep(pattern="log2AspectRatio_outlier", x=out_var)
        if (length(idx)) {
            names(out_var)[idx] <-
                "I(abs(log2AspectRatio) * as.numeric(dist_border < 50))"
        }
    }

    train_bad <- df |>
        dplyr::filter(cell_id %in% train_bad_var) |>
        dplyr::mutate(QScore_train=0)

    train_good <- df |>
        dplyr::filter(cell_id %in% train_good_var) |>
        dplyr::mutate(
            QScore_train=1,
            is_a_bad_boy=cell_id %in% train_bad$cell_id
        )

    train_bad <- train_bad |>
        dplyr::distinct(cell_id, .keep_all=TRUE)

    if (verbose) message(paste0("Chosen low qual examples: ", nrow(train_bad)))

    # good example duplicates removal without any warning to the user
    train_good <- train_good |>
        dplyr::distinct(cell_id, .keep_all=TRUE)

    train_good <- train_good[!train_good$is_a_bad_boy, ]

    n_bad <- nrow(train_bad)
    stopifnot("Not enough good examples to match bad examples"=
                nrow(train_good) > n_bad)
    idx <- sample(seq_len(nrow(train_good)), n_bad, replace=FALSE)
    train_good <- train_good[idx, ]

    if (verbose) {
        message(paste0(
            "Chosen good quality examples (should match bad): ",
            nrow(train_good)
        ))
    }

    train_good$is_a_bad_boy <- NULL
    train_df <- rbind(train_bad, train_good)
    train_df <- train_df |>
        dplyr::distinct(cell_id, .keep_all=TRUE)
    ## `qcscore_train` remains available because `computeTrainDF()` is a
    ## published API. Canonical internals use `QScore_train`.
    train_df$qcscore_train <- train_df$QScore_train
    return(train_df)
}

#' getModelFormula
#' @name getModelFormula
#' @rdname getModelFormula
#' @description
#' Returns the right‐hand side of a model formula string based on a vector of
#' metric names.
#' @param formulaVars A named character vector mapping metric names to outlier
#'   columns, as historically stored in
#'   `metadata(spe)$formula_variables`. An unnamed character vector of metric
#'   names is also accepted.
#' @param verbose Logical. If `TRUE`, prints the generated formula.
#' @param ... May contain the named replacement `metricList`. If both
#'   `formulaVars` and `metricList` are supplied, `metricList` takes precedence
#'   with a warning.
#' @return `character`
#'   A one‐sided formula as a string (e.g. "~ log2SignalDensity + ...").
#' @export
#' @examples
#' example(checkOutliers)
#' getModelFormula(metadata(spe)$formula_variables)
getModelFormula <- function(formulaVars, verbose=FALSE, ...)
{
    has_formula_vars <- !missing(formulaVars)
    dots <- list(...)
    unknown_dots <- setdiff(names(dots), "metricList")
    if (length(unknown_dots) > 0L) {
        stop(
            "Unused argument(s): ",
            paste(unknown_dots, collapse=", "),
            "."
        )
    }
    has_metric_list <- "metricList" %in% names(dots)
    if (!has_formula_vars && !has_metric_list) {
        stop("'formulaVars' or 'metricList' must be supplied.")
    }
    if (has_formula_vars && has_metric_list) {
        warning(
            "Both 'formulaVars' and 'metricList' were supplied; ",
            "using 'metricList'."
        )
    }
    out_var <- if (has_metric_list) dots$metricList else formulaVars
    if (!is.character(out_var)) {
        stop("Model formula variables must be supplied as a character vector.")
    }
    if (!is.null(names(out_var)) && all(nzchar(names(out_var)))) {
        out_var <- names(out_var)
    }
    if ("log2AspectRatio" %in% out_var) {
        out_var[grep("log2AspectRatio", out_var)] <-
            "I(abs(log2AspectRatio) * as.numeric(dist_border < 50))"
    }
    model_formula <- paste0("~(", paste(out_var, collapse = " + "),
                        ")^2", sep = "")

    if (verbose) {
        message("Final formula used for Quality Score computation:")
        message(model_formula)
    }

    return(model_formula)
}

.qscoreSupportedPredictors <- function() {
    c(
        "log2SignalDensity",
        "Area_um",
        "log2AspectRatio",
        "log2Ctrl_total_ratio"
    )
}

.qscoreBorderTerm <- function() {
    paste0(
        "I(abs(log2AspectRatio)*",
        "as.numeric(dist_border<50))"
    )
}

.validateQScoreFormula <- function(modelFormula, dataNames, technology) {
    model_formula <- tryCatch(
        stats::as.formula(modelFormula),
        error=function(e) {
            stop("Unable to parse 'modelFormula': ", conditionMessage(e))
        }
    )
    if (length(model_formula) != 2L) {
        stop("'modelFormula' must be a one-sided formula.")
    }

    required_variables <- base::all.vars(model_formula)
    supported <- .qscoreSupportedPredictors()
    allowed_variables <- c(supported, "dist_border")
    unsupported_variables <- setdiff(
        required_variables,
        allowed_variables
    )
    if (length(unsupported_variables) > 0L) {
        stop(
            "Unsupported Quality Score predictor(s): ",
            paste(unsupported_variables, collapse=", "),
            ". Formula structure may be customised, but predictors and ",
            "transformations are restricted to: ",
            paste(supported, collapse=", "),
            "."
        )
    }

    term_labels <- attr(stats::terms(model_formula), "term.labels")
    if (length(term_labels) == 0L) {
        stop("'modelFormula' must contain at least one supported predictor.")
    }
    normalise <- function(x) gsub("[[:space:]]+", "", x)
    border_term <- .qscoreBorderTerm()
    unsupported_terms <- character()
    for (term_label in term_labels) {
        components <- strsplit(
            normalise(term_label),
            ":",
            fixed=TRUE
        )[[1]]
        valid_components <- components %in% supported |
            components == border_term
        if (!all(valid_components)) {
            unsupported_terms <- c(unsupported_terms, term_label)
        }
    }
    unsupported_terms <- unique(unsupported_terms)
    if (length(unsupported_terms) > 0L) {
        stop(
            "Unsupported Quality Score formula term(s) or transformation(s): ",
            paste(unsupported_terms, collapse=", "),
            ". Formula structure and interactions may be customised, but ",
            "predictors and transformations may not be extended. Supported ",
            "predictors are: ",
            paste(supported, collapse=", "),
            "; the supported border-effect expression is ",
            "'I(abs(log2AspectRatio) * as.numeric(dist_border < 50))'."
        )
    }

    missing_variables <- setdiff(required_variables, dataNames)
    if (length(missing_variables) > 0L) {
        stop(
            "Missing variables required by 'modelFormula': ",
            paste(missing_variables, collapse=", "),
            ". Run 'spatialPerCellQC()' before computing Quality Score."
        )
    }

    is_cosmx <- technology %in%
        c("Nanostring_CosMx", "Nanostring_CosMx_Protein")
    uses_border_predictor <- any(
        c("log2AspectRatio", "dist_border") %in% required_variables
    )
    if (uses_border_predictor && !is_cosmx) {
        stop(
            "'log2AspectRatio' and the border-effect expression are ",
            "supported only for Nanostring CosMx datasets; detected ",
            "technology: '", technology, "'."
        )
    }

    formula_text <- if (is.character(modelFormula)) {
        paste(modelFormula, collapse=" ")
    } else {
        paste(deparse(model_formula, width.cutoff=500L), collapse=" ")
    }
    training_metrics <- required_variables[
        required_variables %in% supported
    ]
    list(
        formula=model_formula,
        text=formula_text,
        variables=required_variables,
        terms=term_labels,
        training_metrics=training_metrics
    )
}

#' .computeXenMerTrainSet
#' @name dot-computeXenMerTrainSet
#' @rdname dot-computeXenMerTrainSet
#' @description
#' Internal: builds Training Set for Xenium & MERFISH.
#' Splits a SpatialExperiment into “bad” vs “good” cells based on
#' pre-computed outlier labels on log2SignalDensity.
#' @param spe \code{SpatialExperiment}
#' @return
#' A list with elements \code{bad} and \code{good}, each a data.frame
#' with \code{qscore_train} and (for “good”) an \code{is_a_bad_boy} flag.
#' @keywords internal
.computeXenMerTrainSet <- function(spe)
{
    train_bad <- data.frame(colData(spe)) |>
        filter(log2SignalDensity_outlier_train=="LOW") |> mutate(qscore_train=0)
    train_good <- data.frame(colData(spe)) |>
        filter((log2SignalDensity > quantile(log2SignalDensity, probs = 0.90) &
                    log2SignalDensity < quantile(log2SignalDensity, probs = 0.99))) |>
        mutate(qscore_train=1, is_a_bad_boy=cell_id %in% train_bad$cell_id)
    return(list(bad=train_bad, good=train_good))
}

#' .computeCosmxTrainSet
#' @name dot-computeCosmxTrainSet
#' @rdname dot-computeCosmxTrainSet
#' @description
#' Internal: builds Training Set for CosMx.
#' Splits a SpatialExperiment into “bad” vs “good” cells based on
#' outliers in aspect ratio near FOV border or low signal density.
#' @param spe \code{SpatialExperiment}
#' @return
#' A list with elements \code{bad} and \code{good}, each a data.frame
#' with \code{qscore_train} and (for “good”) an \code{is_a_bad_boy} flag.
#' @keywords internal
.computeCosmxTrainSet <- function(spe)
{
    spe <- computeSpatialOutlier(spe, "log2AspectRatio", "scuttle")
    train_bad <- data.frame(colData(spe)) |>
        filter((log2AspectRatio_outlier_sc == "HIGH" & dist_border < 50) |
                (log2AspectRatio_outlier_sc == "LOW" & dist_border < 50) |
                log2SignalDensity_outlier_train == "LOW") |>
        mutate(qscore_train = 0)

    train_good <- data.frame(colData(spe)) |>
        filter((log2AspectRatio > quantile(log2AspectRatio, probs = 0.25) &
                    log2AspectRatio < quantile(log2AspectRatio, probs = 0.75) &
                    dist_border > 50) |
                    (log2SignalDensity > quantile(log2SignalDensity, probs = 0.90) &
                    log2SignalDensity < quantile(log2SignalDensity, probs = 0.99))) |>
        mutate(qscore_train=1, is_a_bad_boy=cell_id %in% train_bad$cell_id)
    return(list(bad=train_bad, good=train_good))
}

#' .computeCosmxProteinTrainSet
#' @name dot-computeCosmxProteinTrainSet
#' @rdname dot-computeCosmxProteinTrainSet
#' @description
#' Internal: builds Training Set for CosMx-Protein.
#' Splits a SpatialExperiment into “bad” vs “good” cells based on
#' outliers in aspect ratio near FOV border or low signal density.
#' @param spe \code{SpatialExperiment}
#' @return
#' A list with elements \code{bad} and \code{good}, each a data.frame
#' with \code{qscore_train} and (for “good”) an \code{is_a_bad_boy} flag.
#' @keywords internal
.computeCosmxProteinTrainSet <- function(spe)
{
    spe <- computeSpatialOutlier(spe, "log2AspectRatio", method="both")
    spe <- computeSpatialOutlier(spe, "log2Ctrl_total_ratio", method="both")
    train_bad <- data.frame(colData(spe)) |>
        filter((log2AspectRatio_outlier_mc == "HIGH" & dist_border < 50) |
                (log2AspectRatio_outlier_mc == "LOW" & dist_border < 50) |
                log2SignalDensity_outlier_train == "LOW" |
                log2Ctrl_total_ratio_outlier_sc == "HIGH") |>
        mutate(qscore_train = 0)

    train_good <- data.frame(colData(spe)) |>
        filter((log2AspectRatio > quantile(log2AspectRatio, probs = 0.25) &
                log2AspectRatio < quantile(log2AspectRatio, probs = 0.75) &
                dist_border > 50) |
                (log2SignalDensity > quantile(log2SignalDensity, probs = 0.90) &
                log2SignalDensity < quantile(log2SignalDensity, probs = 0.99))) |>
        mutate(qscore_train=1, is_a_bad_boy=cell_id %in% train_bad$cell_id)
    return(list(bad=train_bad, good=train_good))
}


#' computeQScoreFlags
#' @name computeQScoreFlags
#' @rdname computeQScoreFlags
#' @description
#' Computes flagged cells based on a manually chosen threshold on Quality Score (QS).
#'
#' This function computes flagged cells based on a manually chosen threshold on
#' QS stored in `SpatialExperiment` object.
#'
#' @param spe A `SpatialExperiment` object with spatial omics data.
#' @param qsThreshold Numeric threshold or quantile for QS. Default is `0.5`.
#' @param useQSQuantiles Logical; if `TRUE`, treat `qsThreshold` as a percentile.
#'
#' @return The `SpatialExperiment` object with added low-quality flagged cells as
#' TRUE/FALSE in `colData`.
#'
#' @importFrom SummarizedExperiment colData
#' @export
#' @examples
#' example(computeQScore)
#' spe <- computeQScoreFlags(spe)
#' table(spe$low_QScore)
#' # if fixed filters are defined we have an additional column
#' spe <- computeThresholdFlags(spe)
#' spe <- computeQScoreFlags(spe)
#' table(spe$low_threshold_QScore)
computeQScoreFlags <- function(spe, qsThreshold=0.5, useQSQuantiles=FALSE) {
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot("QScore" %in% names(colData(spe)))

    if(useQSQuantiles) {
        spe$low_QScore <- ifelse(
            spe$QScore < quantile(spe$QScore, probs=qsThreshold),
            TRUE, FALSE)
    } else {
        spe$low_QScore <- spe$QScore < qsThreshold
    }

    if("threshold_flags" %in% names(colData(spe))) {
        spe$low_threshold_QScore <- (spe$low_QScore &
                                        spe$threshold_flags)
    }
    return(spe)
}

#' .checkSkw
#' @name dot-checkSkw
#' @rdname dot-checkSkw
#' @description
#' Internal: checks skewness of metrics to choose outlier detection method.
#'
#' @param cd colData of `SpatialExperiment` object.
#' @param metricList A character vector specifying the metrics to include in
#' the Quality Score formula. Defaults are "log2SignalDensity", "Area_um",
#' "log2AspectRatio", "log2Ctrl_total_ratio".
#'
#' @return
#' A vector containing the list of chosen outlier detection method for each
#' metric.
#' @importFrom e1071 skewness
#' @keywords internal
.checkSkw <- function(cd, metricList=c("log2SignalDensity", "Area_um",
    "log2AspectRatio", "log2Ctrl_total_ratio")) {
    stopifnot(is(cd, "DataFrame"))
    stopifnot(all(metricList %in% names(cd)))
    method <- c()
    for (i in metricList) {
        skw <- e1071::skewness(cd[[i]], na.rm = TRUE)
        method[i] <- ifelse((skw>-1 & skw<1), "sc", "mc")
    }
    for (submeth in c("log2SignalDensity", "log2Ctrl_total_ratio"))
    {
        if (!submeth %in% names(method)) next
        idx <- switch(submeth,
                    "log2SignalDensity"        = cd[["total"]] > 0,
                    "log2Ctrl_total_ratio" = cd[["ctrl_total_ratio"]] != 0,
                    rep(TRUE, nrow(cd))
        )
        if (!any(idx, na.rm = TRUE)) next
        skw <- e1071::skewness(cd[[submeth]][idx], na.rm=TRUE)
        newm <- ifelse((skw>-1 & skw<1), "sc", "mc")
        if (!is.na(newm)) method[[submeth]] <- newm
    }
    return(method)
}

#' computeOutliersQScore
#' @name computeOutliersQScore
#' @rdname computeOutliersQScore
#' @description
#' Computes outlier cells for each metric that can be used in Quality Score formula
#' for SpatialExperiment.
#'
#' This function calculates outlier cells for each variable specified
#' in `metricList` for a `SpatialExperiment`. log2SignalDensity must be present in
#' the `colData` of the `SpatialExperiment` object as a minimum requirement.
#' The user can choose which metrics to include among the following: Area_um,
#' log2Ctrl_total_ratio, log2AspectRatio. For Xenium and Merfish datasets,
#' even if outliers for log2AspectRatio are detected, neither the metric will be 
#' included in the Quality Score formula nor nor the outliers will be used for
#' training.
#' 
#' @param spe A `SpatialExperiment` object with spatial omics data.
#' @param metricList A character vector specifying the metrics to include in
#' the Quality Score formula. Default is `c("log2SignalDensity", "Area_um",
#' "log2AspectRatio", "log2Ctrl_total_ratio")`.
#'
#' @return The `SpatialExperiment` object with added outlier variables in
#' `colData` and the temporary metrics considered for Quality Score computation
#' stored in the object `metadata`.
#'
#' @details The function computes outliers for each specified metric after
#' automatically choosing the appropriate method according to the skewness of
#' the distribution.
#' Internally the function:
#' \enumerate{
#'    \item Calls \code{.checkSkw()} to choose the proper outlier detection
#'     method according to the variable skewness;
#'    \item Calls \code{computeSpatialOutlier()} on each included metric to get
#'    fences;
#'    \item Labels cells as “LOW”/“HIGH” outliers or “NO”.
#' }
#'
#' @importFrom SummarizedExperiment colData
#' @importFrom dplyr case_when
#' @importFrom scuttle outlier.filter
#' @importFrom stats quantile
#' @export
#' @examples
#' example(readCosmxSPE)
#' spe <- spatialPerCellQC(spe)
#' spe <- computeOutliersQScore(spe)
#' table(spe$log2SignalDensity_outlier_train)
computeOutliersQScore <- function(spe, metricList=c("log2SignalDensity","Area_um",
    "log2AspectRatio", "log2Ctrl_total_ratio")) {

    stopifnot(is(spe, "SpatialExperiment"))
    unsupported_metrics <- setdiff(
        metricList,
        .qscoreSupportedPredictors()
    )
    if (length(unsupported_metrics) > 0L) {
        stop(
            "Unsupported Quality Score metric(s): ",
            paste(unsupported_metrics, collapse=", "),
            ". Supported metrics are: ",
            paste(.qscoreSupportedPredictors(), collapse=", "),
            "."
        )
    }

    cd <- colData(spe)
    method <- .checkSkw(cd, metricList)

    spec <- list(
        log2SignalDensity = list(
            idx  = spe$total > 0,
            zero = spe$total == 0,
            tweak_lower = TRUE
        ),
        log2Ctrl_total_ratio = list(
            idx  = spe$ctrl_total_ratio != 0,
            zero = spe$ctrl_total_ratio == 0,
            tweak_lower = FALSE
        )
    )
    if(any(is.na(method)))
    {
        idx <- which(is.na(method))
        for (i in idx) {
            message("Metric ", names(method)[i], " produced NA skewness values.",
                "It will be removed from the Quality Score formula.")
            method <- method[-i]
        }
    }
    for (var in intersect(names(method), names(spec))) {
        s <- spec[[var]]
        spe_temp <- computeSpatialOutlier(spe[, s$idx], computeBy=var,
                                        method=method[[var]])
        out_var <- paste0(var, "_outlier_", method[[var]])
        low_thr  <- getFencesOutlier(spe_temp, out_var, "lower")
        high_thr <- getFencesOutlier(spe_temp, out_var, "higher")
        if (isTRUE(s$tweak_lower)) {
            if (low_thr < min(spe_temp[[var]], na.rm = TRUE)) {
                low_thr <- stats::quantile(spe[[var]], probs=0.01)
            }
        }
        tgt <- paste0(var, "_outlier_train")
        spe[[tgt]] <- dplyr::case_when(
            s$zero ~ "NO",
            spe[[var]] < low_thr  ~ "LOW",
            spe[[var]] > high_thr ~ "HIGH",
            TRUE ~ "NO"
        )
        spe[[tgt]] <- scuttle::outlier.filter(spe[[tgt]])#is this really needed?
        thr <- getFencesOutlier(spe_temp, out_var)
        if (isTRUE(s$tweak_lower)) thr[1] <- low_thr
        attr(spe[[tgt]], "thresholds") <- thr
    }

    submethod <- method[!names(method) %in%
        c("log2SignalDensity", "log2Ctrl_total_ratio")]

    for(j in names(submethod)){
        spe <- computeSpatialOutlier(spe, computeBy=j, method=submethod[j])
    }

    out_var <- paste0(names(method), "_outlier_", method)
    names(out_var) <- names(method)
    # gives warning if one of the variables is missing, but still works!
    # out_var[names(out_var) %in% c("log2SignalDensity", "log2Ctrl_total_ratio")] <-
    #     c("log2SignalDensity_outlier_train", "log2Ctrl_total_ratio_outlier_train")
    mapping <- c(
        log2SignalDensity = "log2SignalDensity_outlier_train",
        log2Ctrl_total_ratio = "log2Ctrl_total_ratio_outlier_train"
    )
    present <- intersect(names(out_var), names(mapping))
    if (length(present) > 0L) {
        for (nm in present) out_var[nm] <- mapping[nm]
    }
    metadata(spe)$formula_variables <- out_var
    return(spe)
}


#' checkOutliers
#' @name checkOutliers
#' @rdname checkOutliers
#' @description
#' Checks if computed outliers meet the minimum numerical requirement, being
#' at least 0.1\% of total cells for each metric to be used in Quality Score formula.
#' If the requirement is not met, the variable is removed from the formula.
#'
#' @param spe A `SpatialExperiment` object with spatial omics data.
#' @param verbose Logical. If `TRUE`, prints how many outliers were found for
#' each metric.
#'
#' @return The `SpatialExperiment` object with added metric variables that
#' will be used for Quality Score computation stored in the object` metadata`.
#'
#' @details The function checks if computed outliers for each metric meet
#' the minimum number to get the metric included in the Quality Score formula.
#' If verbose is TRUE, it also prints how many outliers were found for each
#' metric.
#'
#' @importFrom SummarizedExperiment colData
#' @export
#' @examples
#' example(computeOutliersQScore)
#' spe <- checkOutliers(spe, verbose = TRUE)
#' metadata(spe)$formula_variables
checkOutliers <- function(spe, verbose=FALSE) {
    warnstopmsg <- function(var, warnstop=c("w","s")) {
        warnstop <- match.arg(warnstop)
        m1 <- paste0("Not enough outlier cells for ", var, ".\n")
        m2 <- switch(warnstop,
                s="In this version of SpaceTrooper, presence of outliers
                for log2SignalDensity is required to compute Quality Score.",
                w="This variable will not be used in the final formula")
        return(paste0(m1, m2))
    }
    out_var <- metadata(spe)$formula_variables
    cd <- colData(spe)
    if (verbose) {
        for (i in names(out_var)) {
            message("Outliers found for ", i, ":")
            outlier_counts <- table(cd[[out_var[i]]])
            for (k in seq_along(outlier_counts)) {
                message(names(outlier_counts)[k], ": ", outlier_counts[k])
            }
        }
    }
    stopifnot(
        "log2SignalDensity is not included in the Quality Score formula.\n
        Quality Score cannot be computed"=
            "log2SignalDensity" %in% names(out_var)
    )
    cfg <- list(
        log2SignalDensity=list(pattern="log2SignalDensity_outlier",
            remove="log2SignalDensity_outlier_train", label="LOW", act=stop,
            code="s"),
        Area_um=list(pattern="Area_um_outlier", remove="Area_um_outlier",
            label="HIGH", act=warning, code="w"),
        log2Ctrl_total_ratio=list(pattern="log2Ctrl_total_ratio_outlier",
            remove="log2Ctrl_total_ratio_outlier_train", label="HIGH",
            act=warning, code="w")
    )
    for (v in intersect(names(cfg), names(out_var))) {
        r <- cfg[[v]]
        gi <- grep(r$pattern, out_var)
        if (length(gi)==0L) next
        col <- out_var[gi]
        labs <- factor(cd[[col]], levels=c("LOW","HIGH","NO"))
        tab <- table(labs)
        cnt <- tab[r$label]
        if (is.null(cnt) || is.na(cnt)) cnt <- 0L
        if (cnt < ncol(spe)*0.001) {
            r$act(warnstopmsg(v, r$code))
            out_var <- out_var[-grep(r$remove, out_var)]
        }
    }
    var <- "log2AspectRatio"
    pat <- paste0(var, "_outlier")
    is_cosmx <- metadata(spe)$technology %in%
                c("Nanostring_CosMx", "Nanostring_CosMx_Protein")
    idx <- grep(pat, out_var)            
    if (is_cosmx && (var %in% names(out_var))) {
        if (length(idx)) {
            col <- out_var[idx]
            labs <- factor(cd[[col]], levels=c("LOW","HIGH","NO"))
            tab <- table(labs)
            nmin <- ncol(spe) * 0.001

            low  <- tab[["LOW"]];  if (is.na(low))  low  <- 0L
            high <- tab[["HIGH"]]; if (is.na(high)) high <- 0L

            if (low < nmin && high < nmin) {
                warning(warnstopmsg(var, "w"))
                out_var <- out_var[-idx]
            }
        }
    } else {
        if (length(idx)) {
            out_var <- out_var[-idx]
        }
    }
    metadata(spe)$formula_variables <- out_var
    return(spe)
}


.prepQCContext <- function(spe, metricList=c("log2SignalDensity", "Area_um",
    "log2AspectRatio", "log2Ctrl_total_ratio"), verbose=FALSE) {

    stopifnot(is(spe,"SpatialExperiment"))

    # remove zero-count cells once
    zerocells <- spe$total==0
    if (sum(zerocells) > 0) {
        warning(sum(zerocells),
            " cells with 0 counts were found. These cells will be removed.")
        spe <- spe[, !zerocells]
    }
    if("log2CountArea" %in% names(colData(spe)))
    {
        stop("log2CountArea found in colData.\n",
        "Please updated the object with the latest version of spatialPerCellQC.")
    }
    spe1 <- computeOutliersQScore(spe, metricList)
    spe1 <- checkOutliers(spe1, verbose)

    out_var <- metadata(spe1)$formula_variables
    df <- as.data.frame(colData(spe1))
    tech <- metadata(spe1)$technology

    return(list(df=df, out_var=out_var, tech=tech))
}

#' applyQScoreModel
#' @name applyQScoreModel
#' @rdname applyQScoreModel
#' 
#' @description
#' Applies a previously trained Quality Score model to a new SpatialExperiment
#' object.
#' See details for important considerations when applying a model to a
#' different dataset.
#' @details
#' The authors do not recommend applying a Quality Score (QS) model trained on one
#' dataset to a different dataset, but this could be useful in some cases,
#' for example to transfer a model trained on a dataset sourced from a tissue 
#' slide sample and then applied to other slides from the same experiment.
#' When transferring the model from a reference dataset of the same technology
#' of the query dataset, there is no need to set a common formula. However,
#' when transferring from a different technology, reference QS computation 
#' should be performed providing a formula that excludes untransferable or unwanted 
#' variables (e.g. computing QS on a reference CosMx dataset omitting the border 
#' effect term to transfer the QS model to a Xenium query dataset. This is done by
#' providing a custom formula in computeQScore, see link computeQScore). The function will check 
#' for the presence of required model variables in the new dataset and will handle 
#' missing or extra variables accordingly.
#'
#' @param spe A `SpatialExperiment` object with QC metrics already computed.
#' @param qsModel A Quality Score model object, usually stored in
#'   `metadata(spe)$QScore_model`.
#' @param scoreName Name of the output column in `colData`.
#'
#' @return A `SpatialExperiment` object with added Quality Score in `colData`.
#'
#' @export
#' @examples
#'
#' example(readCosmxSPE)
#'
#' ## Compute per-cell quality control metrics
#' spe <- spatialPerCellQC(spe)
#'
#' ## Split the object only for example purposes
#' set.seed(1998)
#' idx <- sample(seq_len(ncol(spe)), floor(ncol(spe) / 2))
#' spe_train <- spe[, idx]
#' spe_test <- spe[, -idx]
#'
#' ## Train the Quality Control (QC) score model on one dataset
#' spe_train <- computeQScore(spe_train)
#' qs_model <- metadata(spe_train)$QScore_model
#'
#' ## Apply the trained model to another dataset
#' spe_test <- applyQScoreModel(
#'     spe=spe_test,
#'     qsModel=qs_model,
#'     scoreName="QScore_transferred"
#' )
#'
#' summary(spe_test$QScore_transferred)
applyQScoreModel <- function(spe, qsModel, scoreName="QScore") {
    .applyQScoreModel(spe=spe, qsModel=qsModel, scoreName=scoreName)
}

.applyQScoreModel <- function(spe, qsModel, scoreName="QScore") {
    stopifnot(is(spe, "SpatialExperiment"))

    required <- c(
        "model",
        "bestLambda",
        "model_formula",
        "model_matrix_colnames"
    )

    stopifnot(
        "qsModel is missing required fields" =
            all(required %in% names(qsModel))
    )

    df <- as.data.frame(colData(spe))
    .validateQScoreFormula(
        modelFormula=qsModel$model_formula,
        dataNames=colnames(df),
        technology=metadata(spe)$technology
    )

    ok <- .filterCompleteModelCases(
        df=df,
        modelFormula=qsModel$model_formula,
        context="cells"
    )

    new_matrix <- model.matrix(
        as.formula(qsModel$model_formula),
        data=df[ok, , drop=FALSE]
    )

    missing_cols <- setdiff(
        qsModel$model_matrix_colnames,
        colnames(new_matrix)
    )

    if (length(missing_cols) > 0L) {
        zero_matrix <- matrix(
            0,
            nrow=nrow(new_matrix),
            ncol=length(missing_cols),
            dimnames=list(rownames(new_matrix), missing_cols)
        )

        new_matrix <- cbind(new_matrix, zero_matrix)
    }

    extra_cols <- setdiff(
        colnames(new_matrix),
        qsModel$model_matrix_colnames
    )

    if (length(extra_cols) > 0L) {
        warning(
            "New model matrix contains columns not present in the ",
            "training matrix. These columns will be ignored: ",
            paste(extra_cols, collapse=", ")
        )
    }

    new_matrix <- new_matrix[, qsModel$model_matrix_colnames, drop=FALSE]

    score <- rep(NA_real_, nrow(df))

    score[ok] <- as.vector(
        predict(
            qsModel$model,
            s=qsModel$bestLambda,
            newx=new_matrix,
            type="response"
        )
    )

    cd <- colData(spe)
    cd[[scoreName]] <- score
    colData(spe) <- cd

    metadata(spe)$QScore_model_applied <- qsModel

    return(spe)
}

.dropModelIntercept <- function(modelMatrix) {
    if ("(Intercept)" %in% colnames(modelMatrix)) {
        modelMatrix <- modelMatrix[, colnames(modelMatrix)!="(Intercept)",
            drop=FALSE]
    }

    return(modelMatrix)
}

.getQScoreResponse <- function(trainDF) {
    has_canonical <- "QScore_train" %in% colnames(trainDF)
    has_legacy <- "qcscore_train" %in% colnames(trainDF)
    if (!has_canonical && !has_legacy) {
        stop(
            "Training data must contain 'QScore_train' or ",
            "'qcscore_train'."
        )
    }
    if (has_canonical && has_legacy) {
        canonical <- trainDF$QScore_train
        legacy <- trainDF$qcscore_train
        same <- (is.na(canonical) & is.na(legacy)) |
            (!is.na(canonical) & !is.na(legacy) & canonical == legacy)
        if (any(!same)) {
            stop(
                "'QScore_train' and 'qcscore_train' are both present ",
                "but contain different values."
            )
        }
    }
    if (has_canonical) trainDF$QScore_train else trainDF$qcscore_train
}

#' @importFrom stats complete.cases
.filterCompleteModelCases <- function(df, modelFormula, response=NULL,
    context="cells") {

    vars <- base::all.vars(stats::as.formula(modelFormula))

    missing_vars <- setdiff(vars, colnames(df))

    if (length(missing_vars) > 0L) {
        stop(
            "Missing variables required by model formula: ",
            paste(missing_vars, collapse=", ")
        )
    }
    ## Missing values may come from variables used in `modelFormula`, for
    ## example `log2AspectRatio` when aspect ratio cannot be computed from
    ## polygons.
    ok <- complete.cases(df[, vars, drop=FALSE])

    if (!is.null(response)) {
        ok <- ok & !is.na(response)
    }

    if (!all(ok)) {
        warning(
            sum(!ok),
            " ", context,
            " contain missing model variables."
        )
    }

    return(ok)
}

## ---- Deprecated functions -----------------------------------------------

#' Deprecated functions in SpaceTrooper
#'
#' @name SpaceTrooper-deprecated
#' @description
#' These functions are retained for compatibility with published SpaceTrooper
#' APIs. They issue a deprecation warning and direct users to the canonical
#' Quality Score functions.
#'
#' @details
#' \itemize{
#'   \item `computeQCScore()` replaces canonical `QScore` and
#'     `metadata(spe)$QScore_model` with historical `QC_score` and
#'     `metadata(spe)$QCScore_model`.
#'   \item `computeQCScoreFlags()` consumes `QC_score` and creates
#'     `low_qcscore` and, when applicable, `low_threshold_qcscore`.
#'   \item `computeOutliersQCScore()` is the historical name for
#'     `computeOutliersQScore()`.
#'   \item `applyQCScoreModel()` retains the `qcModel` argument, the default
#'     `QC_score` output, and `metadata(spe)$QCScore_model_applied`.
#' }
#'
#' Use `computeQScore()`, `computeQScoreFlags()`,
#' `computeOutliersQScore()`, and `applyQScoreModel()` for canonical names.
NULL

#' computeQCScore (deprecated)
#' @rdname SpaceTrooper-deprecated
#'
#' @param spe A `SpatialExperiment` object.
#' @param bestLambda Passed to \code{\link{computeQScore}}.
#' @param modelFormula Passed to \code{\link{computeQScore}}.
#' @param verbose Passed to \code{\link{computeQScore}}.
#' @return A `SpatialExperiment` object; see \code{\link{computeQScore}}.
#' @export
computeQCScore <- function(spe, bestLambda=NULL, modelFormula=NULL,
                           verbose=FALSE) {
    .Deprecated(
        new="computeQScore",
        package="SpaceTrooper",
        msg=paste0(
            "'computeQCScore' is deprecated.\n",
            "Use 'computeQScore' instead.\n",
            "See help('computeQScore') for details."
        )
    )
    spe <- computeQScore(
        spe,
        bestLambda=bestLambda,
        modelFormula=modelFormula,
        verbose=verbose
    )
    spe$QC_score <- spe$QScore
    spe$QScore <- NULL
    metadata(spe)$QCScore_model <- metadata(spe)$QScore_model
    metadata(spe)$QScore_model <- NULL
    return(spe)
}

#' computeQCScoreFlags (deprecated)
#' @rdname SpaceTrooper-deprecated
#'
#' @param spe A `SpatialExperiment` object.
#' @param qsThreshold Passed to \code{\link{computeQScoreFlags}}.
#' @param useQSQuantiles Passed to \code{\link{computeQScoreFlags}}.
#' @return A `SpatialExperiment` object; see \code{\link{computeQScoreFlags}}.
#' @export
computeQCScoreFlags <- function(spe, qsThreshold=0.5, useQSQuantiles=FALSE) {
    .Deprecated(
        new="computeQScoreFlags",
        package="SpaceTrooper",
        msg=paste0(
            "'computeQCScoreFlags' is deprecated.\n",
            "Use 'computeQScoreFlags' instead.\n",
            "See help('computeQScoreFlags') for details."
        )
    )
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot("QC_score" %in% names(colData(spe)))

    if (useQSQuantiles) {
        spe$low_qcscore <- spe$QC_score <
            quantile(spe$QC_score, probs=qsThreshold, na.rm=TRUE)
    } else {
        spe$low_qcscore <- spe$QC_score < qsThreshold
    }

    if ("threshold_flags" %in% names(colData(spe))) {
        spe$low_threshold_qcscore <- spe$low_qcscore &
            spe$threshold_flags
    }
    return(spe)
}

#' computeOutliersQCScore (deprecated)
#' @rdname SpaceTrooper-deprecated
#'
#' @param spe A `SpatialExperiment` object.
#' @param metricList Passed to \code{\link{computeOutliersQScore}}.
#' @return A `SpatialExperiment` object; see
#'   \code{\link{computeOutliersQScore}}.
#' @export
computeOutliersQCScore <- function(spe, metricList=c(
    "log2SignalDensity", "Area_um", "log2AspectRatio",
    "log2Ctrl_total_ratio"
)) {
    .Deprecated(
        new="computeOutliersQScore",
        package="SpaceTrooper"
    )
    computeOutliersQScore(spe=spe, metricList=metricList)
}

#' applyQCScoreModel (deprecated)
#' @rdname SpaceTrooper-deprecated
#'
#' @param spe A `SpatialExperiment` object with QC metrics already computed.
#' @param qcModel A historical QC score model object, usually stored in
#'   `metadata(spe)$QCScore_model`.
#' @param scoreName Name of the legacy output column in `colData`.
#' @return A `SpatialExperiment` object with the applied score in `colData`
#'   and the model in `metadata(spe)$QCScore_model_applied`.
#' @export
applyQCScoreModel <- function(spe, qcModel, scoreName="QC_score") {
    .Deprecated(
        new="applyQScoreModel",
        package="SpaceTrooper"
    )
    spe <- .applyQScoreModel(
        spe=spe,
        qsModel=qcModel,
        scoreName=scoreName
    )
    metadata(spe)$QCScore_model_applied <-
        metadata(spe)$QScore_model_applied
    metadata(spe)$QScore_model_applied <- NULL
    return(spe)
}

library(testthat)
library(SpaceTrooper)

# Load the example SpatialExperiment directly so tests do not depend on an
# installed help database.
cosmx_path <- system.file(
    "extdata",
    "CosMx_DBKero_Tiny",
    package="SpaceTrooper"
)
spe0 <- readCosmxSPE(cosmx_path, sampleName="DBKero_Tiny")

test_that("QC functions are exported", {
    expect_true(exists("spatialPerCellQC",   mode = "function"))
    expect_true(exists("computeQScore",       mode = "function"))
    expect_true(exists("computeSpatialOutlier", mode = "function"))
    expect_true(exists("computeQScoreFlags",  mode = "function"))
    expect_true(exists("computeThresholdFlags",  mode = "function"))
    ## deprecated wrappers should still be exported
    expect_true(exists("computeQCScore",      mode = "function"))
    expect_true(exists("computeQCScoreFlags", mode = "function"))
    expect_true(exists("computeOutliersQCScore", mode = "function"))
    expect_true(exists("applyQScoreModel", mode = "function"))
    expect_true(exists("applyQCScoreModel", mode = "function"))
})


test_that("spatialPerCellQC adds per-cell metrics to colData", {
    spe <- spatialPerCellQC(spe0, micronConvFact = 0.15)
    expect_s4_class(spe, "SpatialExperiment")
    cd <- colData(spe)
    # at least these should now be in colData()
    required <- c("sum", "detected", "total", "control_sum", "target_sum",
                  "Area_um", "log2AspectRatio", "log2SignalDensity")
    expect_true(all(required %in% colnames(cd)))
})

test_that("computeQScore adds a QScore between 0 and 1", {
    spe <- spatialPerCellQC(spe0)
    set.seed(42)
    spe2 <- computeQScore(spe)
    cd2 <- colData(spe2)
    expect_true("QScore" %in% colnames(cd2))
    fs <- cd2$QScore
    expect_true(is.numeric(fs))
    expect_true(all(fs[!is.na(fs)] >= 0 & fs[!is.na(fs)] <= 1))
})

test_that("computeQCScore preserves legacy score and model names", {
    spe <- spatialPerCellQC(spe0)
    set.seed(42)
    expect_warning(
        spe2 <- computeQCScore(spe),
        "deprecated"
    )
    expect_true("QC_score" %in% colnames(colData(spe2)))
    expect_false("QScore" %in% colnames(colData(spe2)))
    expect_true("QCScore_model" %in% names(metadata(spe2)))
    expect_false("QScore_model" %in% names(metadata(spe2)))
})

test_that("computeQCScoreFlags preserves legacy flag names", {
    spe <- spe0
    spe$QC_score <- seq(0, 1, length.out=ncol(spe))
    spe$threshold_flags <- rep(c(TRUE, FALSE), length.out=ncol(spe))

    expect_warning(
        flagged <- computeQCScoreFlags(spe, qsThreshold=0.5),
        "deprecated"
    )

    expect_true(all(c(
        "low_qcscore",
        "low_threshold_qcscore"
    ) %in% colnames(colData(flagged))))
    expect_false("low_QScore" %in% colnames(colData(flagged)))
})

test_that("computeSpatialOutlier flags outliers for a chosen metric", {
    spe <- spatialPerCellQC(spe0)
    # use 'Area_um' and both methods
    out <- computeSpatialOutlier(spe, computeBy="Area_um", method="scuttle")
    cd_out <- colData(out)
    expect_true(is.logical(cd_out$Area_um_outlier_sc) ||
                    is(cd_out$Area_um_outlier_sc, "outlier.filter"))
})


test_that("computeQScoreFlags combines filters and returns filter_out", {
    spe <- spatialPerCellQC(spe0)
    set.seed(42)
    spe <- computeQScore(spe)
    ff <- computeThresholdFlags(spe,
                            totalThreshold = 10,
                            ctrlTotRatioThreshold = 0.2)
    cd_ff <- colData(ff)
    # expect logical columns and combined filter_out
    expect_true(all(c("is_zero_counts", "is_ctrl_tot_outlier",
                      "threshold_flags") %in% colnames(cd_ff)))
    expect_true(is.logical(cd_ff$is_zero_counts))
    expect_true(is.logical(cd_ff$is_ctrl_tot_outlier))
    expect_true(is.logical(cd_ff$threshold_flags))
    # filter_out should only be TRUE where all three flags are TRUE
    combined <- (ff$is_zero_counts & ff$is_ctrl_tot_outlier)
    expect_identical(combined, cd_ff$threshold_flags)
})

test_that("getModelFormula preserves historical calls", {
    formula_vars <- c(
        log2SignalDensity="log2SignalDensity_outlier_train",
        Area_um="Area_um_outlier_sc"
    )

    expect_identical(
        getModelFormula(formula_vars),
        "~(log2SignalDensity + Area_um)^2"
    )
    expect_message(
        getModelFormula(formulaVars=formula_vars, verbose=TRUE),
        "Final formula"
    )
    expect_identical(
        getModelFormula(metricList=names(formula_vars)),
        "~(log2SignalDensity + Area_um)^2"
    )
    expect_warning(
        getModelFormula(
            formulaVars=formula_vars,
            metricList="log2SignalDensity"
        ),
        "using 'metricList'"
    )
})

test_that("plotCellsFovs old and canonical arguments have equal effects", {
    old <- plotCellsFovs(
        spe0,
        size=2,
        alpha=0.4,
        alphaNumbers=0.3,
        scaleBar=FALSE
    )
    canonical <- plotCellsFovs(
        spe0,
        pointSize=2,
        pointAlpha=0.4,
        numbersAlpha=0.3,
        scaleBar=FALSE
    )

    expect_equal(old$layers[[1]]$aes_params, canonical$layers[[1]]$aes_params)
    expect_equal(old$layers[[3]]$aes_params, canonical$layers[[3]]$aes_params)

    historical_positional <- plotCellsFovs(
        spe0,
        unique(spe0$sample_id),
        "firebrick",
        "black",
        0.3,
        metadata(spe0)$fov_dim,
        2,
        0.4,
        FALSE,
        0.12
    )
    expect_equal(
        historical_positional$layers[[1]]$aes_params$size,
        2
    )
    expect_warning(
        plotCellsFovs(
            spe0,
            size=1,
            pointSize=2,
            scaleBar=FALSE
        ),
        "using 'pointSize'"
    )
})

test_that("plotZoomFovsMap old and canonical arguments are equivalent", {
    cosmx_polygons <- readCosmxSPE(
        cosmx_path,
        sampleName="DBKero_Tiny",
        keepPolygons=TRUE
    )
    fov <- unique(cosmx_polygons$fov)[1]

    old <- plotZoomFovsMap(
        cosmx_polygons,
        fovs=fov,
        mapNumbersCol="blue",
        mapAlphaNumbers=0.3,
        csize=0.7,
        calpha=0.4,
        scaleBars=FALSE
    )
    canonical <- plotZoomFovsMap(
        cosmx_polygons,
        fovs=fov,
        fovNumbersCol="blue",
        fovNumbersAlpha=0.3,
        mapPointSize=0.7,
        mapPointAlpha=0.4,
        scaleBars=FALSE
    )

    expect_true(isTRUE(all.equal(
        old,
        canonical,
        check.environment=FALSE
    )))
    expect_warning(
        plotZoomFovsMap(
            cosmx_polygons,
            fovs=fov,
            csize=0.5,
            mapPointSize=0.7,
            scaleBars=FALSE
        ),
        "using 'mapPointSize'"
    )
})

library(testthat)
library(SpaceTrooper)

qscore_transfer_data <- function() {
    path <- system.file(
        "extdata",
        "CosMx_DBKero_Tiny",
        package="SpaceTrooper"
    )
    spe <- spatialPerCellQC(
        readCosmxSPE(path, sampleName="DBKero_Tiny")
    )
    split <- floor(ncol(spe) * 0.7)
    list(
        train=spe[, seq_len(split)],
        query=spe[, seq.int(split + 1L, ncol(spe))]
    )
}

test_that("applyQScoreModel transfers a supported custom model", {
    data <- qscore_transfer_data()
    supplied <- paste(
        "~ log2SignalDensity + Area_um +",
        "log2SignalDensity:log2Ctrl_total_ratio"
    )
    set.seed(301)
    trained <- computeQScore(
        data$train,
        modelFormula=supplied
    )
    qscore_model <- metadata(trained)$QScore_model

    applied <- applyQScoreModel(
        data$query,
        qsModel=qscore_model
    )

    expect_true("QScore" %in% colnames(colData(applied)))
    expect_true(all(
        applied$QScore[!is.na(applied$QScore)] >= 0 &
            applied$QScore[!is.na(applied$QScore)] <= 1
    ))
    expect_identical(
        metadata(applied)$QScore_model_applied$model_matrix_colnames,
        qscore_model$model_matrix_colnames
    )
    expect_identical(
        qscore_model$model_matrix_colnames,
        c(
            "(Intercept)",
            "log2SignalDensity",
            "Area_um",
            "log2SignalDensity:log2Ctrl_total_ratio"
        )
    )
})

test_that("deprecated applyQCScoreModel preserves legacy output names", {
    data <- qscore_transfer_data()
    set.seed(302)
    trained <- computeQScore(
        data$train,
        modelFormula=~ log2SignalDensity + Area_um
    )
    qscore_model <- metadata(trained)$QScore_model

    expect_warning(
        applied <- applyQCScoreModel(
            data$query,
            qcModel=qscore_model
        ),
        "deprecated"
    )

    expect_true("QC_score" %in% colnames(colData(applied)))
    expect_false("QScore" %in% colnames(colData(applied)))
    expect_true("QCScore_model_applied" %in% names(metadata(applied)))
    expect_false("QScore_model_applied" %in% names(metadata(applied)))
})

test_that("model transfer rejects technology-incompatible border terms", {
    data <- qscore_transfer_data()
    border_formula <- paste0(
        "~ log2SignalDensity + ",
        "I(abs(log2AspectRatio) * as.numeric(dist_border < 50))"
    )
    set.seed(303)
    trained <- computeQScore(
        data$train,
        modelFormula=border_formula
    )

    metadata(data$query)$technology <- "10X_Xenium"
    expect_error(
        applyQScoreModel(
            data$query,
            qsModel=metadata(trained)$QScore_model
        ),
        "supported only for Nanostring CosMx"
    )
})

test_that("computeTrainDF and trainModel accept legacy training response", {
    data <- qscore_transfer_data()
    outliers <- computeOutliersQScore(data$train)
    outliers <- checkOutliers(outliers)
    set.seed(304)
    train_df <- computeTrainDF(
        colData(outliers),
        metadata(outliers)$formula_variables,
        metadata(outliers)$technology
    )

    expect_identical(
        train_df$QScore_train,
        train_df$qcscore_train
    )

    matrix <- stats::model.matrix(
        ~ log2SignalDensity + Area_um,
        data=train_df
    )
    legacy_only <- train_df
    legacy_only$QScore_train <- NULL
    expect_s3_class(
        trainModel(matrix, legacy_only),
        "glmnet"
    )

    conflicting <- train_df
    conflicting$qcscore_train[1] <-
        1L - conflicting$QScore_train[1]
    expect_error(
        trainModel(matrix, conflicting),
        "different values"
    )
})

test_that("stored coefficient metadata contains the complete vector", {
    data <- qscore_transfer_data()
    set.seed(305)
    trained <- computeQScore(
        data$train,
        modelFormula=~ log2SignalDensity + Area_um
    )
    model <- metadata(trained)$QScore_model

    expect_equal(
        nrow(model$coefficients),
        length(model$model_matrix_colnames) + 1L
    )
    expect_equal(
        nrow(model$coefficients_table),
        nrow(model$coefficients)
    )
})

library(testthat)
library(SpaceTrooper)

lambda_train_data <- function(response_name="QScore_train") {
    set.seed(101)
    train_df <- data.frame(
        x1=rnorm(80),
        x2=rnorm(80),
        x3=rnorm(80)
    )
    response <- as.integer(
        train_df$x1 - 0.5 * train_df$x2 + rnorm(80) > 0
    )
    train_df[[response_name]] <- response
    train_df
}

test_that("computeLambda preserves historical positional and named calls", {
    train_df <- lambda_train_data("qcscore_train")
    model_formula <- "~ x1 + x2"

    set.seed(11)
    positional <- computeLambda(train_df, model_formula)
    set.seed(11)
    named <- computeLambda(
        trainDF=train_df,
        modelFormula=model_formula
    )

    expect_type(positional, "double")
    expect_length(positional, 1L)
    expect_true(is.finite(positional))
    expect_identical(positional, named)
})

test_that("computeLambda accepts canonical and legacy responses", {
    canonical <- lambda_train_data("QScore_train")
    legacy <- canonical
    names(legacy)[names(legacy) == "QScore_train"] <- "qcscore_train"

    set.seed(22)
    canonical_lambda <- computeLambda(canonical, ~ x1 + x2)
    set.seed(22)
    legacy_lambda <- computeLambda(legacy, ~ x1 + x2)
    expect_identical(canonical_lambda, legacy_lambda)

    both <- canonical
    both$qcscore_train <- both$QScore_train
    set.seed(22)
    expect_identical(
        computeLambda(both, ~ x1 + x2),
        canonical_lambda
    )

    both$qcscore_train[1] <- 1L - both$QScore_train[1]
    expect_error(
        computeLambda(both, ~ x1 + x2),
        "different values"
    )
})

test_that("computeLambda filters complete cases", {
    train_df <- lambda_train_data()
    train_df$x2[c(2, 7)] <- NA_real_
    complete_df <- train_df[complete.cases(
        train_df[, c("x1", "x2", "QScore_train")]
    ), ]

    set.seed(33)
    expect_warning(
        with_missing <- computeLambda(train_df, ~ x1 + x2),
        "2 training cells"
    )
    set.seed(33)
    without_missing <- computeLambda(complete_df, ~ x1 + x2)

    expect_identical(with_missing, without_missing)
})

test_that("computeLambda preserves additive and selected interaction formulas", {
    train_df <- lambda_train_data()

    set.seed(44)
    additive <- computeLambda(train_df, ~ x1 + x2 + x3)
    set.seed(44)
    selected_interaction <- computeLambda(
        train_df,
        ~ x1 + x2 + x1:x3
    )

    expect_true(is.finite(additive))
    expect_true(is.finite(selected_interaction))
})

test_that("public computeLambda agrees with matrix implementation and glmnet", {
    train_df <- lambda_train_data()
    model_formula <- ~ x1 + x2 + x1:x2
    model_matrix <- stats::model.matrix(model_formula, data=train_df)
    response <- train_df$QScore_train

    set.seed(55)
    public <- computeLambda(train_df, model_formula)
    set.seed(55)
    internal <- SpaceTrooper:::.computeLambda(
        modelMatrix=model_matrix,
        response=response
    )
    set.seed(55)
    reference <- glmnet::cv.glmnet(
        model_matrix,
        response,
        family="binomial",
        alpha=0,
        lambda=NULL
    )$lambda.min

    expect_identical(public, internal)
    expect_identical(public, reference)
})

library(testthat)
library(SpaceTrooper)

qscore_formula_spe <- function() {
    path <- system.file(
        "extdata",
        "CosMx_DBKero_Tiny",
        package="SpaceTrooper"
    )
    spatialPerCellQC(
        readCosmxSPE(path, sampleName="DBKero_Tiny")
    )
}

model_terms <- function(spe) {
    attr(
        stats::terms(stats::as.formula(
            metadata(spe)$QScore_model$model_formula
        )),
        "term.labels"
    )
}

captured_training_metrics <- function(spe, model_formula, seed=206) {
    seen <- new.env(parent=emptyenv())
    original_prep <- SpaceTrooper:::.prepQCContext
    testthat::local_mocked_bindings(
        .prepQCContext=function(spe, metricList, verbose=FALSE) {
            seen$metricList <- metricList
            original_prep(spe, metricList, verbose)
        },
        .package="SpaceTrooper"
    )
    set.seed(seed)
    scored <- computeQScore(
        spe,
        bestLambda=0.01,
        modelFormula=model_formula
    )
    list(metrics=seen$metricList, scored=scored)
}

test_that("supported Quality Score predictors are explicit and stable", {
    expect_identical(
        SpaceTrooper:::.qscoreSupportedPredictors(),
        c(
            "log2SignalDensity",
            "Area_um",
            "log2AspectRatio",
            "log2Ctrl_total_ratio"
        )
    )
})

test_that("default formula keeps the established pairwise interactions", {
    spe <- qscore_formula_spe()
    set.seed(201)
    scored <- computeQScore(spe)
    terms <- model_terms(scored)

    expect_true("log2SignalDensity" %in% terms)
    expect_true("Area_um" %in% terms)
    expect_true(any(grepl(":", terms, fixed=TRUE)))
})

test_that("additive custom formula remains additive", {
    spe <- qscore_formula_spe()
    supplied <- "~ log2SignalDensity + Area_um"
    set.seed(202)
    scored <- computeQScore(spe, modelFormula=supplied)

    expect_identical(
        metadata(scored)$QScore_model$model_formula,
        supplied
    )
    expect_identical(
        model_terms(scored),
        c("log2SignalDensity", "Area_um")
    )
    expect_identical(
        metadata(scored)$QScore_model$model_matrix_colnames,
        c("(Intercept)", "log2SignalDensity", "Area_um")
    )
})

test_that("custom formula preserves selected interactions and column order", {
    spe <- qscore_formula_spe()
    supplied <- paste(
        "~ log2SignalDensity + Area_um +",
        "log2SignalDensity:log2Ctrl_total_ratio"
    )
    set.seed(203)
    scored <- computeQScore(spe, modelFormula=supplied)

    expect_identical(
        model_terms(scored),
        c(
            "log2SignalDensity",
            "Area_um",
            "log2SignalDensity:log2Ctrl_total_ratio"
        )
    )
    expect_identical(
        metadata(scored)$QScore_model$model_matrix_colnames,
        c(
            "(Intercept)",
            "log2SignalDensity",
            "Area_um",
            "log2SignalDensity:log2Ctrl_total_ratio"
        )
    )
})

test_that("custom formula is not rebuilt by getModelFormula", {
    spe <- qscore_formula_spe()
    testthat::local_mocked_bindings(
        getModelFormula=function(...) stop("getModelFormula was called"),
        .package="SpaceTrooper"
    )

    set.seed(204)
    expect_no_error(computeQScore(
        spe,
        modelFormula=~ log2SignalDensity + Area_um
    ))
})

test_that("custom formulas define metrics passed to QScore preparation", {
    spe <- qscore_formula_spe()

    single <- captured_training_metrics(
        spe,
        ~ log2SignalDensity,
        seed=206
    )
    expect_identical(single$metrics, "log2SignalDensity")

    additive <- captured_training_metrics(
        spe,
        ~ log2SignalDensity + Area_um,
        seed=207
    )
    expect_identical(
        additive$metrics,
        c("log2SignalDensity", "Area_um")
    )

    interaction <- captured_training_metrics(
        spe,
        ~ log2SignalDensity * Area_um,
        seed=208
    )
    expect_identical(
        interaction$metrics,
        c("log2SignalDensity", "Area_um")
    )
    expect_identical(
        model_terms(interaction$scored),
        c(
            "log2SignalDensity",
            "Area_um",
            "log2SignalDensity:Area_um"
        )
    )

    selected <- captured_training_metrics(
        spe,
        ~ log2SignalDensity +
            log2SignalDensity:log2Ctrl_total_ratio,
        seed=209
    )
    expect_identical(
        selected$metrics,
        c("log2SignalDensity", "log2Ctrl_total_ratio")
    )
})

test_that("default formula keeps the established training metrics", {
    spe <- qscore_formula_spe()
    default <- captured_training_metrics(spe, NULL, seed=210)

    expect_identical(
        default$metrics,
        SpaceTrooper:::.qscoreSupportedPredictors()
    )
})

test_that("supported CosMx border expression is accepted", {
    spe <- qscore_formula_spe()
    supplied <- paste0(
        "~ log2SignalDensity + ",
        "I(abs(log2AspectRatio) * as.numeric(dist_border < 50))"
    )
    set.seed(205)
    scored <- computeQScore(spe, modelFormula=supplied)

    expect_identical(
        metadata(scored)$QScore_model$model_formula,
        supplied
    )
    formula_info <- SpaceTrooper:::.validateQScoreFormula(
        modelFormula=supplied,
        dataNames=names(colData(spe)),
        technology=metadata(spe)$technology
    )
    expect_identical(
        formula_info$training_metrics,
        c("log2SignalDensity", "log2AspectRatio")
    )
    expect_false("dist_border" %in% formula_info$training_metrics)
})

test_that("custom metric subsets change deterministic training labels", {
    n_cells <- 100L
    training_data <- data.frame(
        cell_id=paste0("cell", seq_len(n_cells)),
        log2SignalDensity=seq_len(n_cells),
        Area_um=seq_len(n_cells),
        log2SignalDensity_outlier_train=c(
            "LOW",
            rep("NO", n_cells - 1L)
        ),
        Area_um_outlier_sc=c(
            "NO",
            "HIGH",
            rep("NO", n_cells - 2L)
        )
    )
    outlier_columns <- c(
        log2SignalDensity="log2SignalDensity_outlier_train",
        Area_um="Area_um_outlier_sc"
    )
    formula_metrics <- function(model_formula) {
        SpaceTrooper:::.validateQScoreFormula(
            modelFormula=model_formula,
            dataNames=names(training_data),
            technology="10X_Xenium"
        )$training_metrics
    }

    signal_metrics <- formula_metrics(~ log2SignalDensity)
    subset_metrics <- formula_metrics(
        ~ log2SignalDensity + Area_um
    )

    set.seed(211)
    signal_train <- computeTrainDF(
        training_data,
        outlier_columns[signal_metrics],
        tech="10X_Xenium"
    )
    set.seed(211)
    subset_train <- computeTrainDF(
        training_data,
        outlier_columns[subset_metrics],
        tech="10X_Xenium"
    )

    expect_identical(
        sort(signal_train$cell_id[signal_train$QScore_train == 0]),
        "cell1"
    )
    expect_identical(
        sort(subset_train$cell_id[subset_train$QScore_train == 0]),
        c("cell1", "cell2")
    )
})

test_that("unsupported predictors and transformations are rejected", {
    spe <- qscore_formula_spe()

    expect_error(
        computeQScore(
            spe,
            modelFormula=~ log2SignalDensity + customMetric
        ),
        "Unsupported Quality Score predictor.*customMetric"
    )
    expect_error(
        computeQScore(
            spe,
            modelFormula=~ log2SignalDensity + log1p(Area_um)
        ),
        "Unsupported Quality Score formula term.*log1p"
    )

    spe$customMetric <- seq_len(ncol(spe))
    expect_error(
        computeOutliersQScore(
            spe,
            metricList=c("log2SignalDensity", "customMetric")
        ),
        "Unsupported Quality Score metric.*customMetric"
    )
})

test_that("custom formulas retain the signal-density requirement", {
    spe <- qscore_formula_spe()

    expect_error(
        computeQScore(spe, modelFormula=~ Area_um),
        "'log2SignalDensity' is required.*training labels"
    )
})

test_that("missing and technology-incompatible variables are reported", {
    spe <- qscore_formula_spe()
    spe$Area_um <- NULL
    expect_error(
        computeQScore(
            spe,
            modelFormula=~ log2SignalDensity + Area_um
        ),
        "Missing variables required.*Area_um"
    )

    xenium_like <- qscore_formula_spe()
    metadata(xenium_like)$technology <- "10X_Xenium"
    expect_error(
        computeQScore(
            xenium_like,
            modelFormula=~ log2SignalDensity + log2AspectRatio
        ),
        "supported only for Nanostring CosMx"
    )
})

addLabelsDBKero <- function(spe)
{
    labf <- system.file(file.path("extdata", "CosMx_DBKero_Tiny",
                            "labels_tiny.tsv"), package="SpaceTrooper")
    labs <-  read.table(file=labf, sep="\t", header=TRUE)
    spe$labels <- NA
    spe$labels_colors <- "black"
    spe$labels[match(labs$cell_id, spe$cell_id)] <- labs$label
    spe$labels_colors[match(labs$cell_id, spe$cell_id)] <- labs$lab_color
    spe$labels <- as.factor(spe$labels)
    spe$labels_colors <- as.factor(spe$labels_colors)
    spe
}

#' .getActiveGeometryName
#' @name .getActiveGeometryName
#' @rdname dot-getActiveGeometryName
#' @param sf an sf object
#'
#' @return character with the name of the active geometry
#' @export
#'
#' @examples
#' example(readPolygonsCosmx)
#' .getActiveGeometryName(polygons)
.getActiveGeometryName <- function(sf)
{
    stopifnot(is(sf, "sf"))
    cn <- attr(sf, "sf_column")
    return(cn)
}

#' .setActiveGeometry
#' @name .setActiveGeometry
#' @rdname dot-setActiveGeometry
#' @param sf an sf object
#' @param name character for the geometry to activate
#'
#' @return an sf object
#' @export
#'
#' @examples
#' example(readPolygonsCosmx)
#' .setActiveGeometry(polygons, "local")
.setActiveGeometry <- function(sf, name)
{
    stopifnot(is(sf, "sf"), name %in% names(sf))
    sf::st_geometry(sf) <- name
    return(sf)
}

#' .renameGeometry
#' @name .renameGeometry
#' @rdname dot-renameGeometry
#' @description renames the `from` to `to` geometry of the `sf` object.
#' If `activate` is `TRUE` it set as the active geometry the new geometry name.
#' Default behaviour is to check if the renamed geometry is already active and
#' leave it as active with the new name.
#'
#' @param sf an sf object with the `from` geometry
#' @param from character indicating the name of the geometry to change
#' @param to character indicating the new name of the geometry
#' @param activate logical indicating if the renamed geometry has to be
#' activated
#'
#' @return an sf object
#' @export
#' @examples
#' example(readPolygonsCosmx)
#' .renameGeometry(polygons, "global", "global1")
.renameGeometry <- function(sf, from, to, activate=FALSE)
{
    stopifnot(all(is(sf, "sf"), from %in% names(sf)))
    act <- .getActiveGeometryName(sf)

    names(sf)[which(names(sf)==from)] <- to
    if(from==act) { sf <- .setActiveGeometry(sf, to) }
    if(activate) { sf <- .setActiveGeometry(sf, to) }
    return(sf)
}

#' getFencesOutlier
#' @name getFencesOutlier
#' @rdname getFencesOutlier
#' @description
#'
#' Retrieve Threshold (Fence) Values from a SpatialExperiment Object
#'
#' This function extracts the threshold values, also known as fences,
#' from a specified column in the `colData` of a `SpatialExperiment` object.
#'
#' @param spe A `SpatialExperiment` object containing spatial transcriptomics
#' data.
#' @param fencesOf A character string specifying the name of the column in
#' `colData(spe)` from which to extract the fence values. This column should
#' contain an `outlier.filter` object (see `computeSpatialOutlier```).
#' @param highLow character indicating which fence to get if "higher", "lower"
#' or "both" (default is "both").
#' @param decimalRound An optional integer specifying the number of decimal
#' places to which the fence values should be rounded. If `NULL`, no rounding is
#'  applied. Default is `NULL`.
#'
#' @return A numeric vector containing the lower and upper threshold values
#' extracted from the specified column.
#'
#' @importFrom SummarizedExperiment colData
#' @export
#' @examples
#' example(computeSpatialOutlier)
#' getFencesOutlier(spe, fencesOf="log2CountArea_outlier_mc")
getFencesOutlier <- function(spe, fencesOf,
                    highLow=c("both", "lower", "higher"), decimalRound=NULL)
{
    stopifnot(is(spe, "SpatialExperiment"))
    stopifnot(fencesOf %in% names(colData(spe)))
    stopifnot(is(colData(spe)[[fencesOf]], "outlier.filter"))
    highLow <- match.arg(highLow)
    fences <- attr(colData(spe)[[fencesOf]], "thresholds")
    if(!is.null(decimalRound)) {fences <- round(fences, decimalRound)}
    fences <- switch (highLow,
        both = {fences},
        lower = {fences[1]},
        higher = {fences[2]}
    )
    return(fences)
}

outfolder <- "personal_folder"

outfolder <- "~/Library/CloudStorage/GoogleDrive-drighelli@gmail.com/My\ Drive/works/coding"
## creating data for cosmx starting from DBKero
cospath <- "~/Downloads/CosMx_data/DBKero/CosMx_Breast/CosMx_data_Case2"
# debug(readCosmxSPE)
spe<- readCosmxSPE(cospath)
f=c(16)
# c=c(1:50)
# gc<- 1:12

countmat_file <- list.files(cospath, "exprMat_file.csv", full.names=TRUE)
countmat <- data.table::fread(countmat_file, showProgress=FALSE) # cell count matrix
c1 <- countmat[countmat$fov%in%f,]
# c1 <- c1[c1$cell_ID%in%c,1:12]
write.csv(x=c1, file=file.path(outfolder, "/SpaceTrooper/inst/extdata/CosMx_DBKero_Tiny/DBKero_exprMat_file16.csv"), row.names=FALSE)

# spe32 <- spe[,spe$fov==32]
# spe32 <- spe32[1:22, 1:100]
# spe32 <- spatialPerCellQC(spe32)
# spe32 <- computeQScore(spe32)
# spe32$quality_score
# spe32 <- computeFixedFlags(spe32)
# FirstFilterPlot(spe32)
metadata_file <- list.files(cospath, "metadata_file.csv", full.names=TRUE)
metadata <- data.table::fread(metadata_file, showProgress=FALSE) # cell metadata
m1 <- metadata[metadata$fov==f,]
# m1 <- m1[m1$cell_ID%in%c,]
write.csv(x=m1, file=file.path(outfolder,"/SpaceTrooper/inst/extdata/CosMx_DBKero_Tiny/DBKero_metadata_file16.csv"), row.names=FALSE)

fovpos_file <- list.files(cospath, "fov_positions_file.csv", full.names=TRUE)
fov_positions <- as.data.frame(data.table::fread(fovpos_file, header=TRUE))
f1 <- fov_positions[fov_positions$fov==f,]
write.csv(x=f1, file=file.path(outfolder,"/SpaceTrooper/inst/extdata/CosMx_DBKero_Tiny/DBKero_fov_positions_file16.csv"), row.names=FALSE)

pol_file <- metadata(spe)$polygons

spat_obj <- data.table::fread(pol_file)
s1 <- spat_obj[spat_obj$fov==f, ]
write.csv(x=s1, file=file.path(outfolder,"/SpaceTrooper/inst/extdata/CosMx_DBKero_Tiny/DBKero-polygons16.csv"), row.names=FALSE)

## data creation for merscope

library(devtools)
load_all()
# folder <- "~/Downloads/MERSCOPE_data/Human_brain"
folder <- "~/Downloads/Merfish_data/human_uterine_cancer_patient2"

debug(readMerfishSPE)
spe <- readMerfishSPE(folder, compute_missing_metrics=FALSE, keep_polygons=FALSE)
countmat_file <- list.files(dirname, countmatfpattern, full.names=TRUE)
metadata_file <- list.files(dirname, metadatafpattern, full.names=TRUE)
# Read in
countmat <- data.table::fread(countmat_file)
c1 <- countmat[1:50,1:11]
write.csv(x=c1, file=file.path(outfolder, "/SpaceTrooper/inst/extdata/Merfish_Tiny/HumanUterineCancerPatient2_cell_by_gene.csv"), row.names=FALSE)

metadata <- data.table::fread(metadata_file) # cell metadata
m1 <- metadata[1:50,]
write.csv(x=m1, file=file.path(outfolder, "/SpaceTrooper/inst/extdata/Merfish_Tiny/HumanUterineCancerPatient2_cell_metadata.csv"), row.names=FALSE)

debug(readPolygonsMerfish)
pols <- readPolygonsMerfish(metadata(spe)$polygons, type="parquet")
mpols <- pols$cell_id[pols$is_multi]
polfile <- metadata(spe)$polygons
polygons <- arrow::read_parquet(polfile, as_data_frame=FALSE)
polygons$metadata
m1 <- data.table::fread(file.path(outfolder, "/SpaceTrooper/inst/extdata/Merfish_Tiny/HumanUterineCancerPatient2_cell_metadata.csv")) # cell metadata
library(dplyr)
polygons
p1 <- polygons %>%
    filter(EntityID %in% m1$EntityID)
p2 <- polygons[polygons$EntityID %in% m1$EntityID,]
p1$Geometry
arrow::write_parquet(p1, sink=file.path(outfolder, "/SpaceTrooper/inst/extdata/Merfish_Tiny/HumanUterineCancerPatient2_cell_boundaries.parquet"))
p3 <- arrow::read_parquet(file.path(outfolder, "/SpaceTrooper/inst/extdata/Merfish_Tiny/HumanUterineCancerPatient2_cell_boundaries.parquet"), as_data_frame=TRUE)
polfiles <- file.path(outfolder, "/SpaceTrooper/inst/extdata/Merfish_Tiny/HumanUterineCancerPatient2_cell_boundaries.parquet")
readPolygonsMerfish(polfiles, "parquet")


#### CosMx Protein
spe <- readCosmxProteinSPE("~/Downloads/CosMx_data/S0_prot")
# spe <- spe1
spe1 <- spe
# x_shift_px <- 1
y_shift_px <- 80
fov_px <- 4239.291

metadata(spe)$fov_positions <- metadata(spe)$fov_positions |>
    dplyr::mutate(
        # x_global_px = x_global_mm / 0.12028 * 1000 + x_shift_px,
        y_global_px = y_global_mm / 0.12028 * 1000 - fov_px + y_shift_px
    )
fovs <- c(16,17,28,29)
fovs <- c(30:100)
fovs <- c(59,60,71,72)
fovs <- 201
spe10<-spe[, spe$fov%in%fovs]
metadata(spe10)$fov_positions <- metadata(spe10)$fov_positions[metadata(spe10)$fov_positions$fov%in%fovs,]
plotCellsFovs(spe10)

fovs <- c(200:201)
spe11<-spe[, spe$fov%in%fovs]
spe11
metadata(spe11)$fov_positions <- metadata(spe11)$fov_positions[metadata(spe11)$fov_positions$fov%in%fovs,]
plotCellsFovs(spe11)

spe11 <- spatialPerCellQC(spe11)
spe11 <- computeSpatialOutlier(spe11, "log2CountArea", method="both")
spe11 <- computeQCScore(spe11)

table(spe10$fov)
fovs <- c(60)
cospath <- "~/Downloads/CosMx_data/S0_prot"
countmat_file <- list.files(cospath, "exprMat_file.csv", full.names=TRUE)
countmat <- data.table::fread(countmat_file, showProgress=FALSE) # cell count matrix
c1 <- countmat[countmat$fov%in%fovs,]
# c1 <- c1[c1$cell_ID%in%c,1:12]
write.csv(x=c1, file=file.path(outfolder, "/SpaceTrooper/inst/extdata/S0_prot/S01_exprMat_file.csv.gz"), row.names=FALSE)

metadata_file <- list.files(cospath, "metadata_file.csv", full.names=TRUE)
metadata <- data.table::fread(metadata_file, showProgress=FALSE) # cell metadata
m1 <- metadata[metadata$fov==fovs,]
write.csv(x=m1, file=file.path(outfolder,"/SpaceTrooper/inst/extdata/S0_prot/S01_metadata_file.csv"), row.names=FALSE)

fovpos_file <- list.files(cospath, "fov_positions_file.csv", full.names=TRUE)
fov_positions <- as.data.frame(data.table::fread(fovpos_file, header=TRUE))
f1 <- fov_positions[fov_positions$FOV %in%fovs,]
y_shift_px <- 80
fov_px <- 4239.291

f1 <- f1 |>
    dplyr::mutate(
        # x_global_px = x_global_mm / 0.12028 * 1000 + x_shift_px,
        y_global_px = y_global_mm / 0.12028 * 1000 - fov_px + y_shift_px
    )
write.csv(x=f1, file=file.path(outfolder,"/SpaceTrooper/inst/extdata/S0_prot/S01_fov_positions_file.csv"), row.names=FALSE)

pol_file <- metadata(spe)$polygons

spat_obj <- data.table::fread(pol_file)
s1 <- spat_obj[spat_obj$fov %in% fovs, ]
write.csv(x=s1, file=file.path(outfolder,"/SpaceTrooper/inst/extdata/S0_prot/S01-polygons.csv"), row.names=FALSE)

spe200 <- readCosmxProteinSPE("~/Library/CloudStorage/GoogleDrive-drighelli@gmail.com/My\ Drive/works/coding/SpaceTrooper/inst/extdata/S01_prot")

## ----setup, include=FALSE-----------------------------------------------------
# Set chunk options: suppress echo, messages, and warnings in code output
knitr::opts_chunk$set(echo = TRUE, message = FALSE, warning = FALSE)

## ----install, eval=FALSE------------------------------------------------------
# # Install BiocManager if not already installed, then install SpaceTrooper
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# BiocManager::install("drighelli/SpaceTrooper")

## ----load-data----------------------------------------------------------------
# Load the SpaceTrooper library
library(SpaceTrooper)

# Load Xenium data into a Spatial Experiment object (SPE)
protfolder <- system.file( "extdata", "S01_prot", package="SpaceTrooper")
(spe <- readCosmxProteinSPE(protfolder))
colData(spe)

## ----plot-fovs----------------------------------------------------------------
# Plot the cells within their respective Field of Views (FOVs)
plotCellsFovs(spe)

## ----cosmx-analysis-qc--------------------------------------------------------
# Perform per-cell quality control checks
spe <- spatialPerCellQC(spe, negProbList=c("Ms IgG1", "Rb IgG"))
names(colData(spe))

## ----plot-metrics-------------------------------------------------------------
# Plot a histogram of counts (sum)
plotMetricHist(spe, metric="sum")
# Plot a histogram of cell areas (Area_um)
plotMetricHist(spe, metric="Area_um")
# Plot a histogram of proportion of counts respect to the cell area in micron 
plotMetricHist(spe, metric="log2CountArea")
# Plot a histogram of proportion of negative probe counts respect to the total
# counts in cells
plotMetricHist(spe, metric="log2Ctrl_total_ratio")

## ----cosmx-analysis-outlier, warning=FALSE------------------------------------
# Identify spatial outliers based on cell area (Area_um)
spe <- computeSpatialOutlier(spe, computeBy="Area_um", method="both")

# Identify spatial outliers based on mean DAPI intensity
spe <- computeSpatialOutlier(spe, computeBy="Mean.DAPI", method="both")
names(colData(spe))

## ----plot-metrics-2-----------------------------------------------------------
# Plot a histogram with fences to identify outliers using the medcouple
plotMetricHist(spe, metric="Area_um", useFences="Area_um_outlier_mc")

# Plot a histogram with fences to identify outliers using scuttle
plotMetricHist(spe, metric="Area_um", useFences="Area_um_outlier_sc")

# Plot a histogram with fences to identify outliers using the medcouple
plotMetricHist(spe, metric="Mean.DAPI", useFences="Mean.DAPI_outlier_mc")

# Plot a histogram with fences to identify outliers using scuttle
plotMetricHist(spe, metric="Mean.DAPI", useFences="Mean.DAPI_outlier_sc")

## ----cosmx-analysis-score-----------------------------------------------------
# Calculate quality scores for each cell
spe <- computeQCScore(spe)
names(colData(spe))

## ----cosmx-analysis-score2----------------------------------------------------
# Compute flags to identify cells for filtering
spe <- computeQCScoreFlags(spe, qsThreshold=0.5)
names(colData(spe))
table(spe$low_qcscore)

## ----thresholds-outliers------------------------------------------------------
spe <- computeThresholdFlags(spe, totalThreshold=0, 
                                ctrlTotRatioThreshold=0.1)
table(spe$threshold_flags)

## ----plot-polygons------------------------------------------------------------
# Read polygon data associated with cells in the SPE
# the polygon file path is stored in the spe metadata
pols <- readPolygonsCosmx(metadata(spe)$polygons)

# Add the polygon data to the SPE object
spe <- addPolygonsToSPE(spe, pols)

## ----plot-polygons-fov-1------------------------------------------------------
# Plot the polygons of the selected cells
plotPolygons(spe, bgColor="white")

## ----plot-polygons-fov-2------------------------------------------------------
# Plot polygons colored by cell area
plotPolygons(spe, colourBy="log2AspectRatio")
plotPolygons(spe, colourBy="Area_um")

## ----plot-polygons-fov-3------------------------------------------------------
plotPolygons(spe, colourBy="QC_score")
plotPolygons(spe, colourBy="low_qcscore")

## ----plot-zoom-fov------------------------------------------------------------
plotZoomFovsMap(spe, fovs=c(201), colourBy="QC_score")
plotZoomFovsMap(spe, fovs=c(201), colourBy="low_qcscore")

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

#' readMerfishSPE
#' @name readMerfishSPE
#' @rdname readMerfishSPE
#' @description
#' `readMerfishSPE()` imports MERFISH/Merscore outputs (counts, metadata,
#' and optionally cell boundary polygons) from a directory and builds a
#' SpatialExperiment object.
#'
#' @param dirName `character(1)`
#'   Path to a folder containing MERFISH output files.
#' @param sampleName `character(1)`
#'   Identifier to assign to the `sample_id` field in the returned object.
#'   Default: `"sample01"`.
#' @param computeMissingMetrics `logical(1)`
#'   If `TRUE`, compute area and aspect‐ratio metrics from the cell boundary
#'   polygons. Default: `TRUE`.
#'   In particular for area, if a "volume" column is present in the colData, it
#'   will be used as area value, otherwise area will be computed from polygons.
#'   This is relevant for MERFISH as the volume is computed on the entire 3D
#'   cell available data, while polygons are 2D sections.
#' @param keepPolygons `logical(1)`
#'   If `TRUE`, attach raw polygon geometries as extra columns in `colData`.
#'   Default: `FALSE`.
#' @param boundariesType `character(1)`
#' One of `"parquet"` or `"HDF5"`. If `"parquet"`, reads a single Parquet
#' file of boundaries; if `"HDF5"`, uses a folder of HDF5 polygon files.
#' @param countmatFPattern `character(1)`
#'   Regex passed to `list.files()` to find the count matrix CSV. Default:
#'   `"cell_by_gene.csv"`.
#' @param metadataFPattern `character(1)`
#'   Pattern to find the cell metadata CSV. Default: `"cell_metadata.csv"`.
#' @param polygonsFPattern `character(1)`
#'   Pattern to find the cell boundaries file. Default:
#'   `"cell_boundaries.parquet"`.
#' @param coordNames `character(2)`
#'   Names of the columns in `colData` that store X/Y spatial coordinates.
#'   Default: `c("center_x", "center_y")`.
#' @param polygonsCol character indicating the name of the polygons column to
#' add into the colData (default is `polygons`).
#' @param useVolume `logical(1)` If `TRUE`, prefer a "volume" column in the
#' metadata for area (when present). Default: `TRUE`.
#' @return A `SpatialExperiment` object with:
#'   - `assays$counts`: gene × cell count matrix
#'   - `colData`: per‐cell metadata (including computed metrics)
#'   - spatial coordinates named by `coordNames`
#'   - `metadata$polygons`: path to the boundaries file
#'   - `metadata$technology`: `"Vizgen_MERFISH"`.
#'
#' @details
#' The function searches `dirName` for three resources using the provided
#' filename patterns: the count matrix (`countmatFPattern`), the cell metadata
#' (`metadataFPattern`), and the polygon resource (`polygonsFPattern`). The
#' count matrix and metadata are expected to contain a `cell_id` column
#' (aliases `V1`, `cell`, or `EntityID` are renamed). If `computeMissingMetrics`
#' is TRUE, `computeMissingMetricsMerfish()` is invoked and receives `useVolume`.
#'
#' The returned `SpatialExperiment` contains:
#' - `assays$counts`: gene × cell count matrix (rows = genes, cols = `cell_id`);
#' - `colData`: per-cell metadata (may be reordered by the function);
#' - spatial coordinates named by `coordNames`;
#' - `metadata$polygons`: path(s) matched by `polygonsFPattern`;
#' - `metadata$technology`: `"Vizgen_MERFISH"`.
#'
#' Side effects: the function may reorder `colData` columns, attach polygon
#' file paths to `metadata(spe)$polygons`, and (when requested) append
#' `Area_um` and `AspectRatio` to `colData`.
#'
#'
#' @author Dario Righelli, Benedetta Banzi
#' @export
#' @importFrom data.table fread
#' @importFrom S4Vectors DataFrame
#' @importFrom dplyr left_join
#' @importFrom SpatialExperiment SpatialExperiment
#' @examples
#' path <- system.file("extdata", "Merfish_Tiny",
#'                     package = "SpaceTrooper")
#' spe <- readMerfishSPE(
#'   dirName = path,
#'   sampleName = "Patient2",
#'   keepPolygons = TRUE,
#'   boundariesType = "parquet"
#' )
#' spe
readMerfishSPE <- function(dirName, sampleName="sample01",
    computeMissingMetrics=TRUE, keepPolygons=FALSE,
    boundariesType=c("parquet", "HDF5"), countmatFPattern="cell_by_gene.csv",
    metadataFPattern="cell_metadata.csv",
    polygonsFPattern="cell_boundaries.parquet",
    coordNames=c("center_x", "center_y"), polygonsCol="polygons",
    useVolume=TRUE) {

    countmat_file <- list.files(dirName, countmatFPattern, full.names=TRUE)
    metadata_file <- list.files(dirName, metadataFPattern, full.names=TRUE)
    countmat <- data.table::fread(countmat_file)
    # spe <- readMerscopeSXE(dirName=dirName, returnType="SPE",
    #     countMatPattern=countmatFPattern, metaDataPattern=metadataFPattern,
    #     coordNames=coordNames)
    pol_file <- list.files(dirName, polygonsFPattern, full.names=TRUE) #parquet?
    names(countmat)[names(countmat) %in% c("V1", "cell")] <- "cell_id"
    metadata <- data.table::fread(metadata_file) # cell metadata
    names(metadata)[names(metadata) %in% c("EntityID", "V1")] <- "cell_id"
    countmat <- left_join(countmat, metadata[, "cell_id"], by = "cell_id")
    cn <- countmat$cell_id
    counts <- subset(countmat, select = -cell_id)
    features <- colnames(counts)
    counts <- t(as.matrix(counts))
    rownames(counts) <- features
    colnames(counts) <- as.character(cn)
    # TODO: rowData (does not exist) tx file use readSparseCSV from sparseArray
    colData <- metadata[match(countmat$cell_id, metadata$cell_id), ]
    rownames(colData) <- colData$cell_id
    cd <- subset(colData, select = c(2,1,3:dim(colData)[2]))
    if (computeMissingMetrics) {
        message("Computing missing metrics, this could take a while...")
        cd <- computeMissingMetricsMerfish(pol_file, colData, boundariesType,
                                            keepPolygons, polygonsCol, useVolume)
    }
    spe <- SpatialExperiment::SpatialExperiment(sample_id=sampleName,
        assays=list(counts=counts), colData=cd,
        spatialCoordsNames=coordNames,
        metadata=list(polygons=pol_file, technology="Vizgen_MERFISH"))
    colnames(spe) <- spe$cell_id
    return(spe)
}

#' computeMissingMetricsMerfish
#' @name computeMissingMetricsMerfish
#' @rdname computeMissingMetricsMerfish
#' @description
#' `computeMissingMetricsMerfish()` takes cell metadata and boundary
#' polygons, calculates per‐cell area and aspect‐ratio, and optionally
#' appends the raw polygon geometries.
#' @param polFile character for the path to the polygon file. Tipically a
#' parquet file or a folder of HDF5 files in the `metadata(spe)$polygons`.
#' @param coldata `DataFrame` or `data.frame`
#' Cell metadata with at least a `cell_id` column.
#' @param boundariesType `character(1)`
#' One of `"HDF5"` or `"parquet"`—passed on to `readPolygonsMerfish()`.
#' @param keepPolygons `logical(1)`
#' If `TRUE`, cbinds the raw polygon `sf` columns onto `coldata`.
#' @param polygonsCol character indicating the name of the polygons column to
#' add into the colData (default is `polygons`).
#' @param useVolume `logical(1)` it assigns the area from the "volume" column.
#' If the column is not present it computes the area from the polygons.
#' Default: `TRUE`.
#'
#' @return A `DataFrame` (or `data.frame`) with:
#'   - all columns of `coldata`
#'   - `Area_um`: area/volume of each cell’s polygon
#'   - `AspectRatio`: width/height aspect ratio
#'   - (optionally) the polygon geometries
#'
#' @details
#' `computeMissingMetricsMerfish()` reads polygon geometries via
#' `readPolygonsMerfish(polFile, type=boundariesType)` where `polFile`
#' is a Parquet file (parquet) or a folder of HDF5 files (HDF5). It expects
#' `coldata` to contain at least `cell_id`.
#'
#' Behavior:
#' - If `useVolume=TRUE` and `coldata` contains a `volume` column, the
#'   returned `Area_um` is set to `volume`. Otherwise `Area_um` is computed
#'   from the polygons using `computeAreaFromPolygons()`.
#' - `AspectRatio` is computed from polygons using `computeAspectRatioFromPolygons()`.
#' - If `keepPolygons=TRUE`, geometries are attached to the returned
#'   `DataFrame` under the name given by `polygonsCol`.
#'
#' Important: polygons must align (in order or by matching logic) with the
#' rows of `coldata`; otherwise area/aspect values may be assigned incorrectly.
#'
#' @export
#' @importFrom S4Vectors DataFrame
#' @examples
#' example(readMerfishSPE)
#' cd <- computeMissingMetricsMerfish(metadata(spe)$polygons, colData(spe),
#'     boundariesType="parquet")
#' colData(spe) <- cd
#' cd
computeMissingMetricsMerfish <- function(polFile, coldata,
    boundariesType=c("parquet", "HDF5"), keepPolygons=FALSE,
    polygonsCol="polygons", useVolume=TRUE) {

    boundariesType <- match.arg(boundariesType)
    polygons <- readPolygonsMerfish(polFile, type=boundariesType)
    cd <- DataFrame(coldata)

    if (useVolume)
    {
        if(!"volume" %in% colnames(cd))
        {
            warning("Volume column not found in colData.\nComputing area from polygons instead.")
            area <- computeAreaFromPolygons(polygons)
        } else {
            warning("Volume is used to compute Quality Score for MERFISH technology.
                For simplicity, it is renamed as Area_um.")
            area <- cd$volume
        }
    }else {
        area <- computeAreaFromPolygons(polygons)
    }
    cd$Area_um <- area
    cd$AspectRatio <- computeAspectRatioFromPolygons(polygons)
    if (keepPolygons) cd <- .addPolygonsToCD(cd, polygons, polygonsCol)
    return(cd)
}

utils::globalVariables(c(
    "log2Ctrl_total_ratio_outlier_sc",
    "log2AspectRatio_outlier_mc",
    "log2AspectRatio_outlier_sc",
    "dist_border",
    "log2CountArea_outlier_train",
    "log2CountArea",
    "log2AspectRatio",
    "cell_id",
    "cell_ID",
    "fixed_flags_color",
    "area_outlier_color",
    "dapi_outlier_color",
    "collapsed_color",
    "polygonsFolder",
    "pol_file",
    "fov",
    "x_mm",
    "y_mm",
    ".data",
    "fov_name",
    "n_points"
))

#' readXeniumSPE
#' @name readXeniumSPE
#' @rdname readXeniumSPE
#' @title Load data from a 10x Genomics Xenium experiment
#'
#' @description
#' Creates a [`SpatialExperiment`] from an unzipped Xenium Output Bundle
#' directory containing spatial gene expression data.
#'
#' @param dirName `character(1)`
#'   Path to a Xenium Output Bundle directory.
#' @param sampleName `character(1)`
#'   Sample identifier to assign to `sample_id`. Default: `"sample01"`.
#' @param type `character(1)`
#'   One of `"HDF5"` or `"sparse"`; method to read the feature matrix.
#' @param coordNames `character(2)`
#'   Names of X/Y spatial coordinate columns. Default:
#'   `c("x_centroid", "y_centroid")`.
#' @param boundariesType `character(1)`
#'   One of `"parquet"` or `"csv"`; format of the polygon file.
#' @param computeMissingMetrics `logical(1)`
#'   If `TRUE`, compute area and aspect‐ratio from boundary polygons.
#' @param keepPolygons `logical(1)`
#'   If `TRUE`, append raw polygon geometries to `colData`.
#' @param countsFilePattern `character(1)`
#'   Pattern to locate the feature matrix file. Default:
#'   `"cell_feature_matrix"`.
#' @param metadataFPattern `character(1)`
#'   Pattern to locate the cell metadata file. Default: `"cells"`.
#' @param polygonsFPattern `character(1)`
#'   Pattern to locate the cell boundaries file. Default:
#'   `"cell_boundaries"`.
#' @param polygonsCol `character(1)`
#'   Name of the polygons column to add to `colData`. Default:
#'   `"polygons"`.
#' @param txPattern `character(1)`
#'   Pattern (base filename, without extension) to locate the transcript file
#'   (usually a `.parquet` file) from which to extract Field-Of-View (FOV)
#'   information for each cell. Default: `"transcripts"`.
#' @param addFOVs `logical(1)`
#'   If `TRUE`, extract Field-Of-View (FOV) information from the transcript file
#'   (as located by `txPattern`) and append it to cell metadata (`colData`).
#'   Default: `FALSE`.

#'
#' @details
#' Expects the unzipped bundle to contain an `outs/` folder with:
#' - `cell_feature_matrix.h5` or `cell_feature_matrix/`
#' - `cells.csv.gz`
#'
#' @return A [`SpatialExperiment`] object with assays, `colData`, spatial
#'   coordinates, and `metadata$polygons` & `metadata$technology`.
#'
#' @author Dario Righelli, Benedetta Banzi
#'
#' @importFrom DropletUtils read10xCounts
#' @importFrom data.table fread
#' @importFrom SpatialExperiment SpatialExperiment
#' @importFrom SummarizedExperiment assays rowData
#' @importFrom SpatialExperimentIO readXeniumSXE
#' @export
#' @examples
#' xepath <- system.file(
#'   "extdata", "Xenium_small", package = "SpaceTrooper"
#' )
#' (spe <- readXeniumSPE(
#'   dirName = xepath,
#'   keepPolygons = TRUE
#' ))
readXeniumSPE <- function(dirName, sampleName="sample01",
    type=c("HDF5", "sparse"), coordNames=c("x_centroid", "y_centroid"),
    boundariesType=c("parquet", "csv"), computeMissingMetrics=TRUE,
    keepPolygons=FALSE, countsFilePattern="cell_feature_matrix",
    metadataFPattern="cells.csv.gz", polygonsFPattern="cell_boundaries",
    polygonsCol="polygons", txPattern="transcripts", addFOVs=FALSE) {

    stopifnot(file.exists(dirName))
    type <- match.arg(type)
    boundariesType <- match.arg(boundariesType)
    if(basename(dirName) != "outs") { # add "outs/" dir if not already included
        dirbkup <- dirName
        dirName <- file.path(dirName, "outs")
        if (!file.exists(dirName)) {
            dirName <- dirbkup
        } else {
            warning("automatically detected/added outs dir in the 10x filepath")
        }
    }

    cfm <- paste0(countsFilePattern, switch(type, HDF5=".h5", ""))
    spe <- SpatialExperimentIO::readXeniumSXE(dirName=dirName,
            countMatPattern=cfm, metaDataPattern=metadataFPattern,
            coordNames=coordNames, returnType="SPE", addExperimentXenium=FALSE,
            altExps=NULL, addParquetPaths=FALSE)
    spe <- .setupXeniumSPE(spe, dirName, sampleName,
        boundariesType, computeMissingMetrics, keepPolygons,
        polygonsFPattern, polygonsCol, txPattern, addFOVs)
    return(spe)
}

#' computeMissingMetricsXenium
#' @name computeMissingMetricsXenium
#' @rdname computeMissingMetricsXenium
#'
#' @description
#' Compute Missing Metrics for Xenium Data
#'
#' This function computes missing metrics, such as the aspect ratio, from
#' polygon data in a Xenium dataset and optionally appends the polygon data to
#' the resulting `colData`.
#'
#' @param polFile A character string specifying the file path to the polygon
#' data.
#' @param colData A `DataFrame` containing the `colData` for the Xenium dataset.
#' @param keepPolygons A logical value indicating whether to keep the polygon
#' data in the resulting `colData`. Default is `FALSE`.
#' @param polygonsCol character indicating the name of the polygons column to
#' add into the colData (default is `polygons`).
#'
#' @return A `DataFrame` containing the updated `colData` with computed metrics.
#' If `keepPolygons` is `TRUE`, the polygon data is also included.
#'
#' @details The function reads the polygon data from the specified file,
#' computes the aspect ratio for each polygon, and merges these metrics with
#' the provided `colData`. Optionally, the polygon data can be kept in the
#' returned `colData`.
#'
#' @importFrom S4Vectors cbind.DataFrame
#' @export
#'
#' @examples
#' example(readXeniumSPE)
#' colData(spe) <- computeMissingMetricsXenium(metadata(spe)$polygons,
#'     colData(spe), keepPolygons=TRUE)
computeMissingMetricsXenium <- function(polFile, colData, keepPolygons=FALSE,
                                polygonsCol="polygons") {
    stopifnot(file.exists(polFile))
    polygons <- readPolygonsXenium(polFile, keepMultiPol=TRUE)
    cd <- colData
    cd$AspectRatio <- computeAspectRatioFromPolygons(polygons)
    cd <- .checkAndFixAreaXenium(cd, polygons)
    if(keepPolygons) cd <- .addPolygonsToCD(cd, polygons, polygonsCol)
    return(cd)
}

#' .addFovFromTx
#' @name dot-addFovFromTx
#' @rdname dot-addFovFromTx
#'
#' @description
#' Add FOV information from transcript file to cell metadata.
#'
#' This function retrieves FOV information from transcript file and appends
#' the data to the resulting `colData`.
#'
#' @param txFile `character(1)` path to a Xenium Output tx file.
#' @param colData A `DataFrame` containing the `colData` for the Xenium dataset.
#'
#' @return A `DataFrame` containing the updated `colData` with FOV information.
#'
#' @details The function reads the transcript file then groups it by cell_id
#' and merges the FOV information to the cell metadata in `colData`. Only
#' parquet file is supported for this operation
#'
#' @importFrom S4Vectors cbind.DataFrame
#' @importFrom arrow read_parquet
#' @importFrom dplyr group_by select distinct left_join
#' @keywords internal
.addFovFromTx <- function(txFile, colData) {
    stopifnot(file.exists(txFile))
    df <- data.frame(colData)
    tx <- arrow::read_parquet(txFile)
    if (!"fov_name" %in% colnames(tx)) {
        stop("No fov_name column was found in tx file. \r\n",
            "Rerun readXeniumSPE without adding FOV information.")}
    g_tx <- group_by(tx, cell_id) |> select(cell_id, fov=fov_name) |>
        distinct(cell_id, .keep_all = TRUE)
    df <- left_join(df, g_tx, by="cell_id")
    colData$fov <- df$fov
    return(colData)
}

#' .checkAndFixArea
#' @description
#' Check and fix cell area column in colData
#'
#' This internal helper verifies whether a \code{colData} table
#' contains a column named \code{"cell_area"}. If found, it copies that
#' column into a new standardized column named \code{"Area_um"}.
#' If the column is missing, the function computes cell areas using
#' \code{\link{computeAreaFromPolygons}}.
#'
#' @param cd A \linkS4class{DataFrame} (typically \code{colData(spe)})
#' containing per-cell metadata.
#' @param polygons An `sf` object with matching `fov` and `cellID` columns.
#'
#' @return The same \code{cd} object with a column \code{Area_um} added
#' or replaced.
#'
#' @details
#' This function ensures consistent naming of the cell area field.
#' It is intended for internal use within quality control or preprocessing
#' steps of \code{SpatialExperiment} objects.
#'
#' @seealso [computeAreaFromPolygons()]
#'
#' @keywords internal
#' @examples
#' cd <- DataFrame(cell_area = c(10, 20, 30))
#' cd <- .checkAndFixArea(cd)
#' head(cd$Area_um)
#' @noRd
.checkAndFixAreaXenium <- function(cd, polygons) {
    idx <- which(names(cd) == "cell_area")
    if (length(idx) != 0) {
        cd$Area_um <- cd[[idx]]
        cd$cell_area <- NULL
    } else {
        cd$Area_um <- computeAreaFromPolygons(polygons)
    }
    return(cd)
}

#' @keywords internal
#' @noRd
.setupXeniumSPE <- function(spe, dirName, sampleName="sample01",
    boundariesType=c("parquet", "csv"), computeMissingMetrics=TRUE,
    keepPolygons=FALSE, polygonsFPattern="cell_boundaries",
    polygonsCol="polygons", txPattern="transcripts", addFOVs=FALSE) {
    stopifnot("spe is not a SpatialExperiment"=is(spe, "SpatialExperiment"))
    stopifnot("dirName not existing"=file.exists(dirName))
    boundariesType <- match.arg(boundariesType)
    spe$sample_id <- sampleName
    rownames(colData(spe)) <- spe$cell_id
    pex <- paste0(polygonsFPattern, switch(boundariesType, parquet=".parquet",
                                                            csv=".csv.gz"))
    polfile <- list.files(dirName, pex, full.names=TRUE)
    cd <- colData(spe)
    if (computeMissingMetrics) {
        message("Computing missing metrics, this could take some time...")
        cd <- computeMissingMetricsXenium(polfile, cd, keepPolygons,
                                        polygonsCol)
    }
    if (addFOVs) {
        cd <- .addFovFromTx(file.path(dirName, paste0(txPattern, ".parquet")), cd)
    }

    if (!identical(colData(spe), cd)) colData(spe) <- cd
    metadata(spe) <- list(polygons=polfile, technology="10X_Xenium")
    return(spe)
}

#' updateXeniumSPE
#'
#' @description
#' Update a SpatialExperiment created from 10x Genomics Xenium outputs by
#' wiring polygons/boundaries, computing optional QC metrics, and standardizing
#' metadata and column names. This is a thin wrapper that delegates to the
#' internal helper `.setupXeniumSPE()`.
#'
#' @param spe SpatialExperiment object to update.
#' @param dirName Directory containing Xenium outputs.
#' @param sampleName Sample identifier to store (default \code{"sample01"}).
#' @param polygonsFPattern Character pattern used to locate polygon files when
#'     boundaries are provided as CSVs (default \code{"polygons.csv"}).
#' @param boundariesType One of \code{c("parquet","csv")} indicating the source
#'     format of cell boundaries.
#' @param computeMissingMetrics Logical; if \code{TRUE}, compute metrics that are
#'     not already present from transcripts/polygons.
#' @param keepPolygons Logical; if \code{TRUE}, keep polygons in the resulting
#'     object (e.g., in metadata or colData, depending on implementation).
#' @param polygonsCol Name of the geometry/column storing polygons when reading
#'     from parquet (default \code{"polygons"}).
#' @param txPattern Pattern (file/glob) for transcript-level files
#'     (default \code{"transcripts"}).
#' @param addFOVs Logical; if \code{TRUE}, derive and attach FOV identifiers
#'     from transcript resources.
#'
#' @return Updated SpatialExperiment object.
#'
#' @details
#' This function performs input checks and then calls `.setupXeniumSPE()`,
#' which does the heavy lifting (I/O, renaming, metadata updates, metrics).
#'
#' @export
#' @examples
#' xepath <- system.file("extdata", "Xenium_small", package="SpaceTrooper")
#' (spe <- SpatialExperimentIO::readXeniumSXE(dirName=xepath))
#' spe <- updateXeniumSPE(spe, dirName=xepath, boundariesType="parquet",
#'     computeMissingMetrics=TRUE, keepPolygons=TRUE)
updateXeniumSPE <- function(spe, dirName, sampleName="sample01",
    boundariesType=c("parquet", "csv"), computeMissingMetrics=TRUE,
    keepPolygons=FALSE, polygonsFPattern="cell_boundaries",
    polygonsCol="polygons", txPattern="transcripts", addFOVs=FALSE) {
    stopifnot("spe is not a SpatialExperiment"=is(spe, "SpatialExperiment"))
    stopifnot("dirName not existing"=file.exists(dirName))
    spe <- .setupXeniumSPE(spe=spe, dirName=dirName, sampleName=sampleName,
        boundariesType=boundariesType, computeMissingMetrics=computeMissingMetrics,
        keepPolygons=keepPolygons, polygonsFPattern=polygonsFPattern,
        polygonsCol=polygonsCol, txPattern=txPattern, addFOVs=addFOVs)
    return(spe)
}

#' readPolygons
#' @name readPolygons
#' @rdname readPolygons
#' @description
#' Read and Validate Polygons from a File
#'
#' This function reads polygon data from a specified file, validates the
#' polygons, and returns them as an `sf` object. It supports multiple file
#' formats and can handle both global and local coordinates.
#'
#' @param polygonsFile A character string specifying the path to the polygon
#' file.
#' @param type A character string specifying the file type. Supported types are
#' `"csv"`, `"parquet"`, and `"h5"`. Default is `"csv"`.
#' @param x A character vector specifying the column names for the x-coordinates
#' in the polygon data. Default is `c("x_global_px", "vertex_x")`.
#' @param y A character vector specifying the column names for the y-coordinates
#' in the polygon data. Default is `c("y_global_px", "vertex_y")`.
#' @param xloc A character string specifying the column name for the local
#' x-coordinates. Default is `"x_local_px"`.
#' @param yloc A character string specifying the column name for the local
#' y-coordinates. Default is `"y_local_px"`.
#' @param keepMultiPol A logical value indicating whether to keep multipolygons
#' during validation. Default is `TRUE`.
#' @param verbose A logical value indicating whether to print additional
#' information during processing. Default is `FALSE`.
#'
#' @return An `sf` object with the loaded and validated polygons.
#'
#' @details The function reads polygon data from the specified file and formats.
#' It validates the polygons and handles both global and local coordinates if
#' provided. If the file type is `"h5"`, the function currently does not handle
#' the data, as this part of the code is not implemented.
#'
#' @importFrom data.table fread
#' @importFrom arrow read_parquet
#' @importFrom sf st_geometry
#' @importFrom dplyr group_by mutate ungroup filter n
#' @export
#' @examples
#' example(readCosmxSPE)
#' polygons <- readPolygons(metadata(spe)$polygons)
#' polygons
readPolygons <- function(polygonsFile, type=c("csv", "parquet", "h5"),
    x=c("x_global_px", "vertex_x"), y=c("y_global_px", "vertex_y"),
    xloc="x_local_px", yloc="y_local_px", #micronConvFact=0.12,
    keepMultiPol=TRUE, verbose=FALSE){
    stopifnot(file.exists(polygonsFile))
    type <- match.arg(type)
    x <- match.arg(x)
    y <- match.arg(y)

    if(type=="h5") {
        stop("h5 support is momentarily disabled, please report an issue on GH")
    } else {
        spat_obj <- switch(type, csv=fread(polygonsFile),
                                parquet=read_parquet(polygonsFile))
        if (! "cell_id" %in% colnames(spat_obj)) {
            spat_obj$cell_id <- paste0("f", spat_obj$fov, "_c", spat_obj$cellID)
        }
        spat_obj$cell_id <- as.factor(spat_obj$cell_id)
        # removing polygons with less than 4 points
        if(any(as.vector(table(spat_obj$cell_id)<4)))
        {
            warning(paste0("Removing ",table(table(spat_obj$cell_id)<4)["TRUE"],
                            " polygons with less than 4 points."))
            spat_obj <- spat_obj |> group_by(cell_id) |> mutate(n_points = n()) |>
                ungroup() |>  filter(n_points >= 4)
        }
        polygons <- .createPolygons(spat_obj, x=x, y=y,
                                        polygon_id="cell_id")
        polygons <- .renameGeometry(polygons, "geometry", "global") ##for cosmx
        if(all(c(xloc, yloc) %in% colnames(spat_obj))) {
            idxs <- which(colnames(polygons) %in% c(xloc, yloc))
            polygons <- polygons[,-idxs]
            polygons_loc <- .createPolygons(spat_obj, x=xloc, y=yloc,
                                            polygon_id="cell_id")[,-c(4,5)]
            polygons <- cbind(polygons, polygons_loc$geometry)
            polygons <- .renameGeometry(polygons, "geometry", "local")
        }
        if(verbose) message("Polygons detected: ", dim(polygons)[1])
        polygons <- .checkPolygonsValidity(polygons, keepMultiPol=keepMultiPol,
                                            verbose=verbose)
        rownames(polygons) <- polygons$cell_id #needed x check addPolygonsToSPE
        if(verbose) message("Polygons after validity: ", dim(polygons)[1])
        return(polygons)
    }
}


#' .checkPolygonsValidity
#' @name .checkPolygonsValidity
#' @rdname dot-checkPolygonsValidity
#' @description checks validity on a geometry of `sf` object.
#' It removes multipolygons when `keepMultiPol` is `FALSE`
#'
#' @details
#' In case geometry is NULL validity is checked on the active geometry,
#' otherwise it is checked on the passed geometry without changing the active
#' geometry of the sf object.
#' In case of not valid polygons, these are removed.
#' If keeMultiPol is FALSE, possible detected multipolygons are removed.
#'
#'
#' @param sf An `sf` class object containing the spatial data.
#' @param geometry character for the geometry to check validity, if `NULL`
#' it checks the active geometry (default is `NULL`)
#' @param keepMultiPol logical for keeping/removing moltipolygons, if any
#' (default is `TRUE`, so keeping the multipolygons)
#' @param verbose logical to print verbose output (default is `FALSE`)
#'
#' @return An `sf` object with valid geometries, possibly with multipolygons
#' removed.
#' @keywords internal
#' @importFrom sf st_is_valid st_buffer st_geometry_type
.checkPolygonsValidity <- function(sf, geometry=NULL, keepMultiPol=TRUE,
    verbose=FALSE) {
    stopifnot(is(sf, "sf"))
    if(is.null(geometry)) { geometry <- .getActiveGeometryName(sf)
    } else {
        act <- .getActiveGeometryName(sf)
        sf <- .setActiveGeometry(sf, geometry)
    }
    sf_tf <- st_is_valid(sf) # TODO: parallelize splitting sf ?
    if(sum(sf_tf)!=dim(sf)[1]) sf <- st_buffer(sf, dist=0)
    is_multi <- sf::st_geometry_type(sf[[geometry]]) == "MULTIPOLYGON" ## <<<<
    ## TRUE is merscope - FALSE is cosmx and xenium
    merscopeFl <- (sum(is_multi) == dim(sf)[1])
    funct <- ifelse(merscopeFl, "lengths", "length")
    ll <- lapply(sf[[geometry]], funct)
    sums <- lapply(ll, sum)
    idx <- which(sums!=1)
    sf$is_multi <- FALSE
    sf$multi_n <- 1
    if(length(idx)!=0) {
        sf$is_multi[idx] <- TRUE
        sf$multi_n[idx] <- unlist(sums[idx])
    }
    if( merscopeFl ) {
        if( length(idx)!=0 ) {
            sf$global[-idx] <- st_cast(sf$global[-idx], "POLYGON")
        } else {
            sf$global <- st_cast(sf$global, "POLYGON")
        }
    }

    if(verbose) message("Detected ", sum(sf$is_multi), " multipolygons.")
    if(!keepMultiPol) {
        if(verbose) message("Removing ", sum(sf$is_multi), " multipolygons.")
        sf <- sf[!sf$is_multi,]
    }
    if(exists("act")) sf <- .setActiveGeometry(sf, act)
    return(sf)
}

#' readAndAddPolygonsToSPE
#' @name readAndAddPolygonsToSPE
#' @rdname readAndAddPolygonsToSPE
#' @description Read and Add Polygons to a SpatialExperiment Object
#'
#' This function reads polygon boundary data based on the technology associated
#' with the provided SpatialExperiment (SPE) object and adds the polygons to the
#' SPE.
#'
#' @param spe A \code{SpatialExperiment} object. The object should contain
#' metadata with the field \code{"technology"}, specifying the technology used
#' (e.g., "Nanostring_CosMx", "Vizgen_MERFISH", or "10X_Xenium").
#' @param keepMultiPol Logical. If \code{TRUE}, multi-polygon features will be
#' kept when reading the boundary data. Defaults to \code{TRUE}.
#' @param boundariesType Character. Specifies the type of boundary file format
#' to read. Options are \code{"HDF5"} or \code{"parquet"}. Defaults to
#' \code{"HDF5"}.
#' @param polygonsCol character indicating the name of the polygons column to
#' add into the colData (default is `polygons`).
#'
#' @return A \code{SpatialExperiment} object with the added polygon data.
#'
#' @export
#'
#' @examples
#' example(readCosmxSPE)
#' spe <- readAndAddPolygonsToSPE(spe)
#' colData(spe)
readAndAddPolygonsToSPE <- function(spe, polygonsCol="polygons",
    keepMultiPol=TRUE, boundariesType=c("csv", "HDF5", "parquet"))
{
    boundariesType <- match.arg(boundariesType)
    stopifnot("technology" %in% names(metadata(spe)))
    tech <- metadata(spe)$technology
    switch(tech,
            "Nanostring_CosMx_Protein"=,
            "Nanostring_CosMx"={
                polygons <- readPolygonsCosmx(metadata(spe)$polygons)
            },
            "Vizgen_MERFISH"={
                ##### NEED TO HANDLE THE DIFFERENCES BETWEEN HDF5 FILES
                ##### AND PARQUET, TO PROPAGATE TO READING FUNCTION
                # ifelse(boundariesType=="HDF5", merpol=)
                polygons <- readPolygonsMerfish(metadata(spe)$polygons,
                                keepMultiPol=TRUE, type=boundariesType)
            },
            "10X_Xenium"={
                polygons <- readPolygonsXenium(metadata(spe)$polygons,
                                            keepMultiPol=TRUE)
            },
            stop("Unrecognized technology, please use an SPE from one of ",
                "Nanostring_CosMx, Vizgen_MERFISH and 10x_Xenium")
    )
    spe <- addPolygonsToSPE(spe, polygons, polygonsCol=polygonsCol)
    return(spe)
}



#' .addPolygonsToCD
#' @name .addPolygonsToCD
#' @rdname dot-addPolygonsToCD
#' @description
#' This function enriches a DataFrame (e.g., from colData) with matching polygon
#' geometries.
#'
#' @param cd A DataFrame containing at least `fov` and `cellID` columns.
#' @param polygons An sf object with matching `fov` and `cellID` columns.
#' @param polygonsCol character indicating the name of the polygons column to
#' add into the colData (default is `polygons`).
#' @return A DataFrame identical to `cd`, but row‑subset to cells present in
#' `polygons` and with a new `polygons` list‑column of sf geometries.
#' @keywords internal
.addPolygonsToCD <- function(cd, polygons, polygonsCol="polygons") {
    stopifnot(inherits(cd, "DataFrame"), inherits(polygons, "sf"))

    if("fov" %in% colnames(polygons)) ## case of CosMx
    {
        # Build consistent cell IDs
        cd$cell_id <- paste0("f", cd$fov, "_c", cd$cellID)
        polygons$cell_id <- paste0("f", polygons$fov, "_c", polygons$cellID)
        rownames(cd) <- cd$cell_id
        rownames(polygons) <- polygons$cell_id
    } else { ## valid for both xenium and merfish
        rownames(cd) <- cd$cell_id
        rownames(polygons) <- polygons$cell_id
    }

    # Identify matching cells
    common_ids <- intersect(rownames(cd), rownames(polygons))
    if (length(common_ids) == 0) {
        stop("No matching cell IDs between 'cd' and 'polygons'.")
    }
    if (length(common_ids) < nrow(cd)) {
        warning(sprintf(
            "%d/%d cells have polygon data; subsetting to matches.",
            length(common_ids), nrow(cd)
        ))
    }

    # Subset and attach
    cd <- cd[common_ids, , drop = FALSE]
    polygons <- polygons[common_ids, , drop = FALSE]
    cd[[polygonsCol]] <- polygons

    return(cd)
}


#' addPolygonsToSPE
#' @name addPolygonsToSPE
#' @rdname addPolygonsToSPE
#' @description This function adds polygon data to a `SpatialExperiment` object.
#'
#' @param spe A `SpatialExperiment` object to which polygons will be added.
#' @param polygons An `sf` object containing the polygon data.
#' @param polygonsCol character indicating the name of the polygons column to
#' add into the colData (default is `polygons`).
#' @return The `SpatialExperiment` object with polygons added to the `colData`.
#'
#' @importFrom dplyr left_join
#' @export
#'
#' @examples
#' example(readCosmxSPE)
#' polygons <- readPolygonsCosmx(metadata(spe)$polygons)
#' spe <- addPolygonsToSPE(spe, polygons)
#' spe$polygons
addPolygonsToSPE <- function(spe, polygons, polygonsCol="polygons") {
    stopifnot(inherits(spe, "SpatialExperiment"), inherits(polygons, "sf"))

    # Extract and enrich colData via DataFrame
    cd <- S4Vectors::DataFrame(colData(spe))
    cd <- .addPolygonsToCD(cd, polygons, polygonsCol)

    # Subset and reorder the SpatialExperiment
    spe <- spe[, which(colnames(spe) %in% cd$cell_id), drop = FALSE]

    # Replace colData with enriched DataFrame
    colData(spe) <- cd

    return(spe)
}


#' .createPolygons
#' @name .createPolygons
#' @rdname dot-createPolygons
#' @description This internal function creates polygons from a data.frame or
#' similar object.
#'
#' @param spat_obj A data frame or similar object containing spatial data.
#' @param x A character vector specifying the x-coordinates.
#' @param y A character vector specifying the y-coordinates.
#' @param polygon_id A character string specifying the polygon ID.
#'
#' @return An `sf` object containing the created polygons.
#' @keywords internal
#' @importFrom sfheaders sf_polygon
#' @importFrom sf st_as_sf
.createPolygons <- function(spat_obj, x=NULL, y=NULL, polygon_id=NULL,
                            geometry="Geometry")
{
    if(all(!is.null(x), !is.null(y)))
    {
        polygons <- sfheaders::sf_polygon(obj=spat_obj, x=x, y=y,
                                    polygon_id=polygon_id, keep=TRUE)
    } else {
        polygons <- sf::st_as_sf(as.data.frame(spat_obj))
        ## check structure of polygons
        # ll <- lapply(polygons$Geometry, lengths)
        # sums <- lapply(ll, sum)
        # idx <- which(sums==1)
        # polygons$is_multi <- FALSE
        # polygons$is_multi[!idx] <- TRUE
        # polygons$multi_n <- 1
        # polygons$multi_n[!idx] <- sums[!idx]
        # polygons[idx, ] <- sf::st_cast(polygons[idx, ], "POLYGON")
    }
    # polygons <- polygons[order(polygons$cell_id),]
    return(polygons)
}

#' readPolygonsCosmx
#' @name readPolygonsCosmx
#' @rdname readPolygonsCosmx
#' @description This function reads polygon data specific to CosMx technology.
#'
#' @param polygonsFile A character string specifying the file path to the
#' polygon data.
#' @param type A character string specifying the file type ("csv" or "parquet").
#' @param x A character string specifying the x-coordinate column.
#' @param y A character string specifying the y-coordinate column.
#' @param xloc A character string specifying the local x-coordinate column.
#' @param yloc A character string specifying the local y-coordinate column.
#' @param keepMultiPol A logical value indicating whether to keep multipolygons.
#' @param verbose A logical value indicating whether to print additional
#' information.
#'
#' @return An `sf` object containing the CosMx polygon data.
#' @export
#'
#' @examples
#' example(readCosmxSPE)
#' polygons <- readPolygonsCosmx(metadata(spe)$polygons)
#' polygons
readPolygonsCosmx <- function(polygonsFile, type=c("csv", "parquet"),
                            x="x_global_px",
                            y="y_global_px",
                            xloc="x_local_px",
                            yloc="y_local_px",
                            keepMultiPol=TRUE,
                            verbose=FALSE)
{
    type <- match.arg(type)
    polygons <- readPolygons(polygonsFile, type=type, x=x, y=y, xloc=xloc,
                    yloc=yloc, keepMultiPol=keepMultiPol, verbose=verbose)
    polygons <- sf::st_cast(polygons, "GEOMETRY")
    mandatory <- c("cell_id", "global", "is_multi", "multi_n")
    cnames <- colnames(polygons)[!colnames(polygons) %in% mandatory]
    polygons <- polygons[,c(mandatory, cnames)]
    polygons <- .setActiveGeometry(polygons, "local")
    polygons <- sf::st_cast(polygons, "GEOMETRY")
    polygons <- .setActiveGeometry(polygons, "global")
    return(polygons)
}

#' readPolygonsXenium
#' @name readPolygonsXenium
#' @rdname readPolygonsXenium
#' @description This function reads polygon data specific to Xenium technology.
#'
#' @param polygonsFile A character string specifying the file path to the
#' polygon data.
#' @param type A character string specifying the file type ("parquet" or "csv").
#' Default is parquet.
#' @param x A character string specifying the x-coordinate column.
#' @param y A character string specifying the y-coordinate column.
#' @param keepMultiPol A logical value indicating whether to keep multipolygons.
#' @param verbose A logical value indicating whether to print additional
#' information.
#'
#' @return An `sf` object containing the Xenium polygon data.
#' @export
#'
#' @examples
#' example(readXeniumSPE)
#' polygons <- readPolygonsXenium(metadata(spe)$polygons, type="parquet")
#' polygons
readPolygonsXenium <- function(polygonsFile, type=c("parquet", "csv"),
                    x="vertex_x", y="vertex_y", keepMultiPol=TRUE,
                    verbose=FALSE)
{
    type <- match.arg(type)
    polygons <- readPolygons(polygonsFile=polygonsFile, type=type, x=x, y=y,
        keepMultiPol=keepMultiPol, verbose=verbose)
    mandatory <- c("cell_id", "global", "is_multi", "multi_n")
    cnames <- colnames(polygons)[!colnames(polygons) %in% mandatory]
    polygons <- polygons[,c(mandatory, cnames)]
    return(polygons)
}

#' readPolygonsMerfish
#' @name readPolygonsMerfish
#' @rdname readPolygonsMerfish
#' @description This function reads polygon data specific to MERFISH technology.
#'
#' @param polygons A character string specifying the folder containing the
#' polygon data files in case of HDF5, or a path to a parquet file (see `type`).
#' @param type A character string specifying the file type("HDF5" or "parquet").
#' Default is parquet.
#' @param hdf5pattern A character string specifying the pattern to match HDF5
#' files.
#' @param keepMultiPol A logical value indicating whether to keep multipolygons.
#' @param zLev An integer specifying the Z level to filter the data. Default is
#' `3L`.
#' @param zcolumn A character string specifying the column name for the Z index.
#' @param geometry A character string specifying the geometry column name.
#' @param verbose A logical value indicating whether to print additional
#' information.
#'
#' @return An `sf` object containing the MERFISH polygon data.
#' @export
#' @importFrom sf st_sf st_cast st_sfc
#' @importFrom arrow read_parquet
#'
#' @examples
#' example(readMerfishSPE)
#' polygons <- readPolygonsMerfish(metadata(spe)$polygons, type="parquet")
#' polygons
readPolygonsMerfish <- function(polygons, type=c("parquet", "HDF5"),
                                keepMultiPol=TRUE, hdf5pattern="hdf5",
                                zLev=3L, zcolumn="ZIndex",
                                geometry="Geometry",
                                verbose=FALSE)
{
    type <- match.arg(type)
    if (type=="HDF5")
    {
        polfiles <- list.files(polygons, pattern=hdf5pattern,
                                full.names=TRUE)
        dfsfl <- lapply(seq_along(polfiles), function(i)
        {
            poll <- readh5polygons(polFile=polfiles[i])
            df <- data.frame(cell_id=paste0("f", i-1, "_c", poll$ids),
                cell_ID=poll$ids, fov=i-1,
                geometry=sf::st_sfc(poll$g)) ## geometry can be a simple column
            dfsf <- sf::st_sf(df)
        })
        polygons <- do.call(rbind, dfsfl)
        ## check if polygons geometry are polygons
        polygons <- .checkPolygonsValidity(polygons, keepMultiPol=keepMultiPol)
        return(polygons)
    } else { ## case parquet
        # polfile <- list.files(polygonsFolder, pattern=type,
                            # full.names=TRUE)
        polfile <- polygons
        polygons <- arrow::read_parquet(polfile, as_data_frame=TRUE)
        polygons <- polygons[polygons[[zcolumn]]==zLev,]
        polygons$cell_id <- polygons$EntityID
        polygons <- .createPolygons(polygons)
        polygons <- .renameGeometry(polygons, geometry, "global")
        polygons <- .checkPolygonsValidity(polygons, keepMultiPol=keepMultiPol,
                                        verbose=verbose)

    }
    mandatory <- c("cell_id", "global", "is_multi", "multi_n")
    cnames <- colnames(polygons)[!colnames(polygons) %in% mandatory]
    polygons <- polygons[,c(mandatory, cnames)]
    return(polygons)
}
#' computeCenterFromPolygons
#' @name computeCenterFromPolygons
#' @rdname computeCenterFromPolygons
#' @description This function computes the center coordinates on x and y axis
#' from polygon data and adds it to the `colData`. It is necessary only for
#' Merfish.
#'
#' @param polygons An `sf` object containing polygon data.
#' @param coldata A `DataFrame` containing the `colData` to which center
#' coordinates information will be added.
#'
#' @return A `DataFrame` with the added center information.
#' @export
#'
#' @examples
#' example(readPolygonsMerfish)
#' coldata <- computeCenterFromPolygons(polygons, colData(spe))
#' colData(spe) <- coldata
computeCenterFromPolygons <- function(polygons, coldata)
{
    # this needs to be changed to be independent from coldata
    cd <- coldata
    # cd$Area <- NA
    centroid <- sf::st_centroid(polygons)
    ## get active name of geometry instead of global
    center_x <- lapply(centroid$geometry, function(x)
    {
        x[1]
    })
    ## get active name of geometry instead of global
    center_y <- lapply(centroid$geometry, function(x)
    {
        x[2]
    })
    cd$center_x <- unlist(center_x)
    cd$center_y <- unlist(center_y)
    return(cd)
}


#' computeAreaFromPolygons
#' @name computeAreaFromPolygons
#' @rdname computeAreaFromPolygons
#' @description This function computes the area from polygon data.
#'
#' @param polygons An `sf` object containing polygon data.
#'
#' @return A `numeric` vector with the area information.
#' @export
#'
#' @examples
#' example(readPolygonsMerfish)
#' area <- computeAreaFromPolygons(polygons)
#' area
computeAreaFromPolygons <- function(polygons)
{
    stopifnot(is(polygons, "sf"))
    area <- sf::st_area(polygons)
    Area_um <- unlist(area)
    return(Area_um)
}


#' computeAspectRatioFromPolygons
#' @name computeAspectRatioFromPolygons
#' @rdname computeAspectRatioFromPolygons
#' @description This function computes the aspect ratio from polygon data.
#'
#' @param polygons An `sf` object containing polygon data.
#'
#' @return A `numeric` vector with the aspect ratio information.
#' @export
#'
#' @examples
#' example(readPolygonsMerfish)
#' ar <- computeAspectRatioFromPolygons(polygons)
#' ar
computeAspectRatioFromPolygons <- function(polygons)
{
    stopifnot(all(is(polygons, "sf"), ("is_multi" %in% colnames(polygons))))
    aspRatL <- numeric(nrow(polygons))
    if(any(polygons$is_multi)) {
        aspRatL[which(polygons$is_multi)] <- NA
        warning("Found ", sum(polygons$is_multi),
            " multi-poligons: returning NA aspect ratio for them.")
    }
    aspRatL[!polygons$is_multi] <- lapply(polygons$global[!polygons$is_multi],
        function(x) {
            (max(x[[1]][, 2]) - min(x[[1]][, 2]))/(max(x[[1]][, 1]) -
                                                    min(x[[1]][, 1]))
    })
    names(aspRatL) <- polygons$cell_id
    ar <- unlist(aspRatL)
    return(ar)
}

#' readh5polygons
#' @name readh5polygons
#' @rdname readh5polygons
#' @description This function reads polygon data from an HDF5 file.
#'
#' @param polFile A character string specifying the file path to the HDF5
#' polygon data.
#'
#' @return A list containing the polygon geometries and their associated cell_id
#' @author Lambda Moses
#' @importFrom rhdf5 h5dump
#' @importFrom sf st_polygon
#' @keywords internal
readh5polygons <- function(polFile)
{
    l <- rhdf5::h5dump(polFile)[[1]]
    cell_ids <- names(l)
    geometries <- lapply(l, function(m) {
        sf::st_polygon(list(t(m[["zIndex_0"]]$p_0$coordinates[,,1])))
    })
    return(list(g=geometries, ids=cell_ids))
}

## ----setup, include=FALSE-----------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    fig.width = 6,
    fig.height = 4
)
suppressPackageStartupMessages({
    library(SpaceTrooper)
    library(SpatialExperiment)
})

## ----cosmx-path---------------------------------------------------------------
cospath <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"), package="SpaceTrooper")
cospath

## ----cosmx-route-a------------------------------------------------------------
spe_cos <- readCosmxSPE(
    dirName=cospath,
    sampleName="DBKero_Tiny",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    countMatFPattern="exprMat_file.csv",
    metadataFPattern="metadata_file.csv",
    polygonsFPattern="polygons.csv",
    fovPosFPattern="fov_positions_file.csv",
    fovdims=c(xdim=4256, ydim=4256)
)
spe_cos

## ----cosmx-route-a-inspect----------------------------------------------------
assayNames(spe_cos)
dim(spe_cos)
head(colnames(spatialCoords(spe_cos)))
metadata(spe_cos)$technology
metadata(spe_cos)$polygons

## ----cosmx-protein, eval=FALSE------------------------------------------------
# protfolder <- system.file("extdata", "S01_prot", package = "SpaceTrooper")
# spe_cos_prot <- readCosmxProteinSPE(
#     dirName=protfolder,
#     sampleName="cosmx_prots",
#     coordNames=c("CenterX_global_px", "CenterY_global_px"),
#     countMatFPattern="exprMat_file.csv",
#     metadataFPattern="metadata_file.csv",
#     polygonsFPattern="polygons.csv",
#     fovPosFPattern="fov_positions_file.csv",
#     fovdims=c(xdim=4256, ydim=4256)
# )
# metadata(spe_cos_prot)$technology

## ----cosmx-route-b------------------------------------------------------------
spe_cos_raw <- SpatialExperimentIO::readCosmxSXE(
    dirName=cospath,
    returnType="SPE",
    countMatPattern="exprMat_file.csv",
    metaDataPattern="metadata_file.csv",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    addFovPos=TRUE,
    fovPosPattern="fov_positions_file.csv",
    altExps=NULL,
    addParquetPaths=FALSE
)
spe_cos_std<-updateCosmxSPE(
    spe=spe_cos_raw,
    dirName=cospath,
    sampleName="DBKero_Tiny",
    polygonsFPattern="polygons.csv",
    fovdims=c(xdim=4256, ydim=4256)
)
identical(spe_cos_std, spe_cos)

## ----xenium-path--------------------------------------------------------------
xepath <- system.file("extdata", "Xenium_small", package = "SpaceTrooper")
xepath

## ----xenium-route-a-----------------------------------------------------------
spe_xen_a <- readXeniumSPE(
    dirName=xepath,
    type="HDF5",
    coordNames=c("x_centroid", "y_centroid"),
    boundariesType="parquet",
    computeMissingMetrics=TRUE,
    keepPolygons=TRUE,
    countsFilePattern="cell_feature_matrix",
    metadataFPattern="cells.csv.gz",
    polygonsFPattern="cell_boundaries",
    polygonsCol="polygons",
    txPattern="transcripts",
    addFOVs=FALSE
)
spe_xen_a

## ----xenium-route-a-inspect---------------------------------------------------
assayNames(spe_xen_a)
dim(spe_xen_a)
colnames(spatialCoords(spe_xen_a))
metadata(spe_xen_a)$polygons
metadata(spe_xen_a)$technology

## ----xenium-route-b-----------------------------------------------------------
spe_xen_b <- SpatialExperimentIO::readXeniumSXE(
    dirName=xepath,
    countMatPattern="cell_feature_matrix.h5",
    metaDataPattern="cells.csv.gz",
    coordNames=c("x_centroid", "y_centroid"),
    returnType="SPE",
    addExperimentXenium=FALSE,
    altExps=NULL,
    addParquetPaths=FALSE
)
spe_xen_b <- updateXeniumSPE(
    spe=spe_xen_b,
    dirName=xepath,
    boundariesType="parquet",
    computeMissingMetrics=TRUE,
    keepPolygons=TRUE,
    polygonsFPattern="cell_boundaries",
    polygonsCol="polygons",
    txPattern="transcripts",
    addFOVs=FALSE
)
spe_xen_b

## ----xenium-route-b-validate--------------------------------------------------
identical(metadata(spe_xen_b)$technology, "10X_Xenium")
identical(spe_xen_a, spe_xen_b)

## ----session-info-------------------------------------------------------------
sessionInfo()

## ----setup, include=FALSE-----------------------------------------------------
# Set chunk options: suppress echo, messages, and warnings in code output
knitr::opts_chunk$set(echo=TRUE, message=FALSE, warning=FALSE)

## ----load-cosmx---------------------------------------------------------------
library(SpaceTrooper)
library(ggplot2)

cosmxFolder <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"), 
                        package="SpaceTrooper")
spe <- readCosmxSPE(cosmxFolder, sampleName="DBKero_CosMx")

spe

## ----load-xenmer--------------------------------------------------------------
xeniumFolder <- system.file( "extdata", "Xenium_small", package="SpaceTrooper")
xen <- readXeniumSPE(xeniumFolder, computeMissingMetrics=TRUE,
                        keepPolygons=TRUE)

merfishFolder <- system.file("extdata", "Merfish_Tiny", package="SpaceTrooper")
mer <- readMerfishSPE(merfishFolder, boundariesType="parquet",
                        computeMissingMetrics=TRUE, keepPolygons=TRUE)

## ----plot-fovs----------------------------------------------------------------
# to check misalignment
plotCellsFovs(spe, pointSize=3, pointAlpha=0.7, numberSize=6)

## ----fov-correction-----------------------------------------------------------
# code line for shift correction
# metadata(spe)$fov_positions$y_global_px <- metadata(spe)$fov_positions$y_global_px - 4256

## ----load-poly,  message=TRUE-------------------------------------------------
# polygon loading
spe <- readAndAddPolygonsToSPE(spe, boundariesType="csv")

## ----cosmx-analysis-qc,  message=TRUE-----------------------------------------
spe <- spatialPerCellQC(spe, rmZeros=TRUE,
        negProbList=c("NegPrb", "Negative", "SystemControl"))

colnames(colData(spe))

## ----cosmx-analysis-score,  message=TRUE--------------------------------------
set.seed(1713)

spe <- computeQScore(spe)

summary(spe$QScore)

## ----cosmx-analysis-score2, message=TRUE--------------------------------------
spe <- computeQScoreFlags(spe, qsThreshold=0.5)

table(spe$low_QScore)

## ----plot-hist----------------------------------------------------------------
# view quantitative metric distribution
plotMetricHist(spe, metric="QScore")

## ----plot-centroids-labels----------------------------------------------------
labf <- system.file(file.path("extdata", "CosMx_DBKero_Tiny",
                            "labels_tiny.tsv"), package="SpaceTrooper")
labs <- read.table(file=labf, sep="\t", header=TRUE)

spe$labels <- as.factor(labs[match(spe$cell_id, labs$cell_id),]$label)
spe$labels_colors <- as.factor(labs[match(spe$cell_id, labs$cell_id),]$lab_color)

plotCentroids(spe, colourBy="labels", size=3, palette="labels_colors")

## ----plot-polygons-fov-1------------------------------------------------------
plotPolygons(spe, colourBy="log2SignalDensity")
plotPolygons(spe, colourBy="Area_um")
plotPolygons(spe, colourBy="log2Ctrl_total_ratio")
plotPolygons(spe, colourBy="log2AspectRatio")

## ----plot-polygons-fov-2------------------------------------------------------
plotPolygons(spe, colourBy="QScore") + 
    scale_fill_viridis_c(option="plasma")

## ----plot-polygons-fov-3------------------------------------------------------
plotPolygons(spe, colourBy="low_QScore") + 
    scale_fill_manual(values=c("TRUE"="red", "FALSE"="#c0c8cf"))

## ----cosmx-analysis-score3, message=TRUE--------------------------------------
spe <- computeQScoreFlags(spe, qsThreshold=0.75)

table(spe$low_QScore)

## ----plot-polygons-fov-4------------------------------------------------------
plotPolygons(spe, colourBy="low_QScore") + 
    scale_fill_manual(values=c("TRUE"="red", "FALSE"="#c0c8cf"))

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

## ----setup, include=FALSE-----------------------------------------------------
# Set chunk options: suppress echo, messages, and warnings in code output
knitr::opts_chunk$set(echo=TRUE, message=FALSE, warning=FALSE)

## ----load-cosmx---------------------------------------------------------------
library(SpaceTrooper)
library(ggplot2)

cosmxFolder <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"), 
                        package="SpaceTrooper")

spe <- readCosmxSPE(dirName=cosmxFolder, 
                    sampleName="DBKero_CosMx",
                    coordNames=c("CenterX_global_px", "CenterY_global_px"),
                    countMatFPattern="exprMat_file.csv",
                    metadataFPattern="metadata_file.csv",
                    polygonsFPattern="polygons.csv",
                    fovPosFPattern="fov_positions_file.csv",
                    fovdims=c(xdim=4256, ydim=4256)
)

## ----load-cosmx-protein-------------------------------------------------------
protfolder <- system.file( "extdata", "S01_prot", package="SpaceTrooper")

prot <- readCosmxProteinSPE(dirName=protfolder, 
                           sampleName="CosMx_Protein_Tonsil",
                           coordNames=c("CenterX_global_px", "CenterY_global_px"),
                           countMatFPattern="exprMat_file.csv", 
                           metadataFPattern="metadata_file.csv", 
                           polygonsFPattern="polygons.csv", 
                           fovPosFPattern="fov_positions_file.csv", 
                           fovdims=c(xdim=4256, ydim=4256)
)

# code line for shift correction
metadata(prot)$fov_positions$y_global_px <- metadata(prot)$fov_positions$y_global_px - 4256

## ----load-xenium, message=TRUE, warning=TRUE----------------------------------
xeniumFolder <- system.file( "extdata", "Xenium_small", package="SpaceTrooper")

xen <- readXeniumSPE(dirName=xeniumFolder, 
                     sampleName="Xenium_small",
                     type=c("HDF5", "sparse"),
                     coordNames=c("x_centroid", "y_centroid"),
                     boundariesType=c("parquet", "csv"),
                     computeMissingMetrics=TRUE,
                     keepPolygons=FALSE,
                     countsFilePattern="cell_feature_matrix",
                     metadataFPattern="cells.csv.gz",
                     polygonsFPattern="cell_boundaries",
                     polygonsCol="polygons",
                     txPattern="transcripts",
                     addFOVs=FALSE

)

## ----load-merfish, message=TRUE, warning= TRUE--------------------------------
merfishFolder <- system.file("extdata", "Merfish_Tiny", package="SpaceTrooper")

mer <- readMerfishSPE(dirName=merfishFolder, 
                      sampleName="Merfish_Tiny",
                      computeMissingMetrics=TRUE,
                      keepPolygons=FALSE,
                      boundariesType=c("parquet"),
                      countmatFPattern="cell_by_gene.csv",
                      metadataFPattern="cell_metadata.csv",
                      polygonsFPattern="cell_boundaries.parquet",
                      coordNames=c("center_x", "center_y"),
                      polygonsCol="polygons")

## ----load-poly,  message=TRUE-------------------------------------------------
# polygon loading
spe <- readAndAddPolygonsToSPE(spe, 
                               polygonsCol="polygons",
                               keepMultiPol=TRUE,
                               boundariesType=c("csv", "HDF5", "parquet")
)

spe$polygons

## ----load-poly-xen,  message=TRUE---------------------------------------------
# xenium polygon loading
xen <- readAndAddPolygonsToSPE(xen, boundariesType=c("parquet"))

## ----load-poly-mer,  message=TRUE---------------------------------------------
# merfish polygon loading
mer <- readAndAddPolygonsToSPE(mer, boundariesType=c("parquet"))

## ----load-poly-prot,  message=TRUE--------------------------------------------
# protein polygon loading
prot <- readAndAddPolygonsToSPE(prot, boundariesType=c("csv"))

## ----analysis-qc-1------------------------------------------------------------
spe <- spatialPerCellQC(spe, rmZeros=TRUE,
        negProbList=c("NegPrb", "Negative", "SystemControl"))

xen <- spatialPerCellQC(xen, rmZeros=TRUE)

mer <- spatialPerCellQC(mer, rmZeros=TRUE)

prot <- spatialPerCellQC(prot, rmZeros=TRUE)

## ----analysis-qc-2------------------------------------------------------------
# check added columns to colData
colnames(colData(spe))

## ----compute-QS,  message=TRUE------------------------------------------------
# MERSCOPE cell_id conversion
# spe$cell_id <- as.character(spe$cell_id)

set.seed(1713)

spe <- computeQScore(spe, verbose=TRUE)

## ----compute-QS-safe-run,  message=TRUE---------------------------------------
# safe run using tryCatch directly so errors are caught before evaluation
out <- tryCatch(
  list(result=computeQScore(spe), error=NULL),
  error=function(e) list(result=NULL, error=e)
)

if (!is.null(out$error)) {
message("Failed: ", out$error$message)
colData(spe)$QScore <- NA
} else {
spe <- out$result
}

## ----compute-QS-different-formula,  message=TRUE------------------------------

set.seed(1713)

model_formula <- "~ log2Ctrl_total_ratio + log2SignalDensity"

spe <- computeQScore(spe, modelFormula=model_formula, verbose=TRUE)

## ----plot-polygons-dbkero-----------------------------------------------------
plotPolygons(spe, colourBy="QScore") + 
    scale_fill_viridis_c(option="plasma")

## ----apply-QS-model, message=TRUE---------------------------------------------
qs_model <- metadata(spe)$QScore_model
spe <- applyQScoreModel(
    spe,
    qsModel=qs_model,
    scoreName="QScore_transferred"
)

## ----compute-outliers-1,  message=TRUE, warning=TRUE--------------------------
# detecting outliers for cell area in Xenium dataset with both Medcouple
# and MAD
xen <- computeSpatialOutlier(xen, computeBy="Area_um", method="both")

## ----compute-outliers-3,  message=TRUE----------------------------------------
spe <- computeSpatialOutlier(spe, computeBy="Mean.CD68", method="both")

table(spe$Mean.CD68_outlier_mc)

## ----plot-hist-1--------------------------------------------------------------
spe <- computeSpatialOutlier(spe, computeBy="log2SignalDensity", method="both")

plotMetricHist(spe, metric="log2SignalDensity", 
               useFences="log2SignalDensity_outlier_mc")

## ----plot-zoom-fov-1----------------------------------------------------------
# plotting FoVs map and zoom in of selected FoV colored by QS
plotZoomFovsMap(spe, fovs=16, colourBy="QScore", mapPointSize=1, 
                fovNumberSize=7, scaleBars=TRUE)

## ----plot-zoom-fov-2----------------------------------------------------------
plotZoomFovsMap(prot, fovs=344, colourBy="Area_um", mapPointSize=1, 
                fovNumberSize=7, scaleBarMap=TRUE, scaleBarPol=FALSE)

## ----plot-centroids-----------------------------------------------------------
plotCentroids(spe, colourBy="QScore", size=3, scaleBar=FALSE) +
    scale_fill_viridis_c(option="plasma") +
    scale_color_viridis_c(option="plasma")

## ----plot-polygons------------------------------------------------------------
plotPolygons(prot, colourBy="log2SignalDensity", scaleBar=FALSE) + 
    scale_fill_viridis_c(option="mako")

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

---
title: "Loading CosMx and Xenium data with SpaceTrooper"
author: "Dario Righelli"
date: "`r BiocStyle::doc_date()`"
pkg: SpaceTrooper
output:
    BiocStyle::html_document:
        toc: true
        toc_float: true
vignette: >
    %\VignetteIndexEntry{Loading CosMx and Xenium data with SpaceTrooper}
    %\VignetteEncoding{UTF-8}
    %\VignetteEngine{knitr::rmarkdown}
---

```{r setup, include=FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    comment = "#>",
    fig.width = 6,
    fig.height = 4
)
suppressPackageStartupMessages({
    library(SpaceTrooper)
    library(SpatialExperiment)
})
```

# Overview

This vignette shows how to build a `SpatialExperiment` (SPE) from: - **Nanostring CosMx** (RNA/protein) outputs - **10x Genomics Xenium** outputs

For each technology, we demonstrate: - **Route A**: SpaceTrooper’s high-level reader (`readCosmxSPE`, `readCosmxProteinSPE`, `readXeniumSPE`) - **Route B**: read with **SpatialExperimentIO**, then standardize with `updateCosmxSPE` / `updateXeniumSPE`

# CosMx (Nanostring)

We begin with **CosMx**. The package ships with a small CosMx example for demonstration.

```{r cosmx-path}
cospath <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"), package="SpaceTrooper")
cospath
```

## Route A — Direct loading with SpaceTrooper

Use `readCosmxSPE()` to construct an SPE from CosMx outputs; it also normalizes names/metadata and records polygons/FOV info if present.

```{r cosmx-route-a}
spe_cos <- readCosmxSPE(
    dirName=cospath,
    sampleName="DBKero_Tiny",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    countMatFPattern="exprMat_file.csv",
    metadataFPattern="metadata_file.csv",
    polygonsFPattern="polygons.csv",
    fovPosFPattern="fov_positions_file.csv",
    fovdims=c(xdim=4256, ydim=4256)
)
spe_cos
```

Inspect essentials:

```{r cosmx-route-a-inspect}
assayNames(spe_cos)
dim(spe_cos)
head(colnames(spatialCoords(spe_cos)))
metadata(spe_cos)$technology
metadata(spe_cos)$polygons
```

### CosMx protein

If working with **CosMx Protein**, use the convenience wrapper:

```{r cosmx-protein, eval=FALSE}
protfolder <- system.file("extdata", "S01_prot", package = "SpaceTrooper")
spe_cos_prot <- readCosmxProteinSPE(
    dirName=protfolder,
    sampleName="cosmx_prots",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    countMatFPattern="exprMat_file.csv",
    metadataFPattern="metadata_file.csv",
    polygonsFPattern="polygons.csv",
    fovPosFPattern="fov_positions_file.csv",
    fovdims=c(xdim=4256, ydim=4256)
)
metadata(spe_cos_prot)$technology
```

## Route B — Via SpatialExperimentIO, then standardize

If you prefer to read with **SpatialExperimentIO** first, upgrade the object with `updateCosmxSPE()` to harmonize names/metadata and attach polygons.

```{r cosmx-route-b}
spe_cos_raw <- SpatialExperimentIO::readCosmxSXE(
    dirName=cospath,
    returnType="SPE",
    countMatPattern="exprMat_file.csv",
    metaDataPattern="metadata_file.csv",
    coordNames=c("CenterX_global_px", "CenterY_global_px"),
    addFovPos=TRUE,
    fovPosPattern="fov_positions_file.csv",
    altExps=NULL,
    addParquetPaths=FALSE
)
spe_cos_std<-updateCosmxSPE(
    spe=spe_cos_raw,
    dirName=cospath,
    sampleName="DBKero_Tiny",
    polygonsFPattern="polygons.csv",
    fovdims=c(xdim=4256, ydim=4256)
)
identical(spe_cos_std, spe_cos)
```

# Xenium (10x Genomics)

A small Xenium example is also included for demonstration.

```{r xenium-path}
xepath <- system.file("extdata", "Xenium_small", package = "SpaceTrooper")
xepath
```

## Route A — Direct loading with SpaceTrooper

`readXeniumSPE()` builds the SPE from a Xenium Output Bundle (root or `outs/`) and standardizes metadata.

Key options: - `type`: `"HDF5"` or `"sparse"` (feature matrix) - `boundariesType`: `"parquet"` or `"csv"` (cell boundaries) - `computeMissingMetrics`: compute QC metrics (area/aspect ratio) if needed - `keepPolygons`: append polygons to `colData` - `addFOVs`: derive FOV IDs from transcript parquet

```{r xenium-route-a}
spe_xen_a <- readXeniumSPE(
    dirName=xepath,
    type="HDF5",
    coordNames=c("x_centroid", "y_centroid"),
    boundariesType="parquet",
    computeMissingMetrics=TRUE,
    keepPolygons=TRUE,
    countsFilePattern="cell_feature_matrix",
    metadataFPattern="cells.csv.gz",
    polygonsFPattern="cell_boundaries",
    polygonsCol="polygons",
    txPattern="transcripts",
    addFOVs=FALSE
)
spe_xen_a
```

Quick checks:

```{r xenium-route-a-inspect}
assayNames(spe_xen_a)
dim(spe_xen_a)
colnames(spatialCoords(spe_xen_a))
metadata(spe_xen_a)$polygons
metadata(spe_xen_a)$technology
```

## Route B — Via SpatialExperimentIO, then standardize

Read with **SpatialExperimentIO** and then pass through `updateXeniumSPE()` for SpaceTrooper-standardized metadata and optional metrics/FOV extraction.

```{r xenium-route-b}
spe_xen_b <- SpatialExperimentIO::readXeniumSXE(
    dirName=xepath,
    countMatPattern="cell_feature_matrix.h5",
    metaDataPattern="cells.csv.gz",
    coordNames=c("x_centroid", "y_centroid"),
    returnType="SPE",
    addExperimentXenium=FALSE,
    altExps=NULL,
    addParquetPaths=FALSE
)
spe_xen_b <- updateXeniumSPE(
    spe=spe_xen_b,
    dirName=xepath,
    boundariesType="parquet",
    computeMissingMetrics=TRUE,
    keepPolygons=TRUE,
    polygonsFPattern="cell_boundaries",
    polygonsCol="polygons",
    txPattern="transcripts",
    addFOVs=FALSE
)
spe_xen_b
```

Validate:

```{r xenium-route-b-validate}
identical(metadata(spe_xen_b)$technology, "10X_Xenium")
identical(spe_xen_a, spe_xen_b)
```

------------------------------------------------------------------------

# File layout notes

-   **CosMx**: per-cell metadata (e.g. `metadata_file.csv`), expression matrix (`exprMat_file.csv`), optional polygon CSV (`polygons.csv`), and FOV positions. `updateCosmxSPE()` also fixes common field names and records FOV dims in `metadata`.
-   **Xenium**: unzipped bundle contains an `outs/` folder. Feature matrix may be `cell_feature_matrix.h5` (HDF5) or sparse folder; cells metadata `cells.csv.gz`; boundaries as `.parquet` or `.csv.gz`; transcript parquet for FOV attribution. `readXeniumSPE()` auto-detects `outs/` if you pass the bundle root.

# Session info

```{r session-info}
sessionInfo()
```

---
title: "SpaceTrooper utilities overview"
author: 
    - name: "Benedetta Banzi"
    - name: "Dario Righelli"
date: "`r BiocStyle::doc_date()`"
output:
    BiocStyle::html_document:
        toc: true
        toc_float: true
vignette: >
    %\VignetteIndexEntry{SpaceTrooper utilities overview}
    %\VignetteEncoding{UTF-8}
    %\VignetteEngine{knitr::rmarkdown}
editor_options:
    chunk_output_type: console
---

```{r setup, include=FALSE}
# Set chunk options: suppress echo, messages, and warnings in code output
knitr::opts_chunk$set(echo=TRUE, message=FALSE, warning=FALSE)
```

## Introduction

`SpaceTrooper` is an `R/Bioconductor` package for Quality Control (QC)
of imaging-based spatial transcriptomics and proteomics data. It provides
multi-platform data harmonization, cell-level QC, and visualization utilities.
The package leverages [SpatialExperiment](https://www.bioconductor.org/packages/release/bioc/html/SpatialExperiment.html) objects to support data from 
**CosMx**, **Xenium**, and **MERFISH** technologies.

This vignette describes additional SpaceTrooper utilities that extend the
standard workflow and are intended to further improve the user experience.
Please, be aware that this is an appendix vignette and does not present an
exhaustive analytical workflow. Therefore, users should consult [RNA](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/RNA_tutorial_vignette.html) and [Protein](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/Protein_tutorial_vignette.html)
tutorials before proceeding. 

In addition, this vignette provides practical tips for making the most effective
use of the package. It is organized into three main sections, corresponding to
the typical analysis pipeline, with relevant insights for each step:

- Data loading
- Quality Control (QC)
- Visualization

## Data loading

### SpatialExperiment object creation

This section explains how to specify the inputs required by each data-reading 
function. Both `readCosmxSPE` and `readXeniumSPE` are wrapper functions built on 
the [SpatialExperimentIO](https://bioconductor.org/packages/release/bioc/html/SpatialExperimentIO.html) infrastructure. 
Further details on their usage are available in the corresponding function 
documentation, e.g. `help(readCosmxSPE)`.

**COSMX**

```{r load-cosmx}
library(SpaceTrooper)
library(ggplot2)

cosmxFolder <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"), 
                        package="SpaceTrooper")

spe <- readCosmxSPE(dirName=cosmxFolder, 
                    sampleName="DBKero_CosMx",
                    coordNames=c("CenterX_global_px", "CenterY_global_px"),
                    countMatFPattern="exprMat_file.csv",
                    metadataFPattern="metadata_file.csv",
                    polygonsFPattern="polygons.csv",
                    fovPosFPattern="fov_positions_file.csv",
                    fovdims=c(xdim=4256, ydim=4256)
)
```

Here are summarized the default values of the input parameters. The `dirName` 
parameter specifies the path to the directory containing the flat files produced
by the CosMx platform:

- expression matrix, (e.g. `exprMat_file.csv`)
- cell metadata, (e.g. `metadata_file.csv`)
- Fields of View positions, (e.g. `fov_positions_file.csv`)
- cell polygons, (e.g. `polygons.csv`)

These files represent the minimum set of inputs required to run the data-reading
function.

The `sampleName` parameter defines the dataset label used as the title in 
plotting functions. 
The `coordNames` parameter specifies the column names in the cell metadata 
corresponding to centroid coordinates. These are stored in the `spatialCoords` 
slot of the `SpatialExperiment` and can be assessed via `spatialCoords(spe)`. 
They are also the columns used by `plotCentroid` function.
Additional details on working with `SpatialExperiment` objects are available in 
the [SpatialExperiment](https://www.bioconductor.org/packages/release/bioc/html/SpatialExperiment.html) documentation.

The `countMatFPattern`, `metadataFPattern`, `polygonsFPattern` and 
`fovPosFPattern` define file name patterns used to identify the expression 
matrix, cell metadata, Fields of view (FoV) positions, and polygon files, 
respectively. As file naming conventions may change over time with evolving 
technologies, these parameters allow users to adapt to different file names or 
extensions. Currently, CSV files are supported, including compressed `.csv.gz.` 
formats.

**IMPORTANT**: the cell polygon file is the only input that is not imported
directly into the `SpatialExperiment` object by the reading function. However,
because polygons can be added later using the `readAndAddPolygonsToSPE` function
(see the [Load polygons](#load-polygons) subsection), the path to the cell
polygon file is stored in the `metadata` slot and accessed via
`metadata(spe)$polygons`.

The `fovdims` parameter specifies the FoV dimensions as a vector containing the 
x and y sizes. While FoV size and shape have changed over time, e.g. from
rectangular to square, in recent CosMx versions they have stabilized at 4,256 
pixels per side.

**COSMX PROTEIN**

```{r load-cosmx-protein}
protfolder <- system.file( "extdata", "S01_prot", package="SpaceTrooper")

prot <- readCosmxProteinSPE(dirName=protfolder, 
                           sampleName="CosMx_Protein_Tonsil",
                           coordNames=c("CenterX_global_px", "CenterY_global_px"),
                           countMatFPattern="exprMat_file.csv", 
                           metadataFPattern="metadata_file.csv", 
                           polygonsFPattern="polygons.csv", 
                           fovPosFPattern="fov_positions_file.csv", 
                           fovdims=c(xdim=4256, ydim=4256)
)

# code line for shift correction
metadata(prot)$fov_positions$y_global_px <- metadata(prot)$fov_positions$y_global_px - 4256
```

`readCosmxProteinSPE` is a wrapper around `readCosmxSPE`, thus the required 
input files and parameters are identical. Input files may be provided in 
compressed `csv.gz` file format, and the `countMatFPattern`, `metadataFPattern`,
`polygonsFPattern` and `fovPosFPattern` parameters are used to specify the 
corresponding file name patterns.

**XENIUM**

```{r load-xenium, message=TRUE, warning=TRUE}
xeniumFolder <- system.file( "extdata", "Xenium_small", package="SpaceTrooper")

xen <- readXeniumSPE(dirName=xeniumFolder, 
                     sampleName="Xenium_small",
                     type=c("HDF5", "sparse"),
                     coordNames=c("x_centroid", "y_centroid"),
                     boundariesType=c("parquet", "csv"),
                     computeMissingMetrics=TRUE,
                     keepPolygons=FALSE,
                     countsFilePattern="cell_feature_matrix",
                     metadataFPattern="cells.csv.gz",
                     polygonsFPattern="cell_boundaries",
                     polygonsCol="polygons",
                     txPattern="transcripts",
                     addFOVs=FALSE

)
```

`readXeniumSPE` function requires the path to the directory containing the 
Xenium output files. This directory is typically generated automatically by the 
Xenium pipeline and usually includes the pattern `outs` in its name. If the 
directory exists but this pattern is missing from the provided path, a warning 
is issued to allow the path to be corrected.

Xenium outputs are often provided redundantly in multiple formats. The 
minimum set of files required to run the `SpaceTrooper` pipeline includes:

- expression matrix
- cell metadata
- cell polygons

Optionally:

- transcript file

The `dirName`, `sampleName` and `coordNames` parameters behave in the same way 
as in `readCosmxSPE`.

The remaining parameters are described below:

`type` specifies the format of the expression matrix to be loaded. Currently
only the following two formats are supported: MEX for sparse matrices (Cell
Ranger output) and HDF5. By default, the HDF5 file is used. If only the sparse
MEX format is available, you should provide a vector containing only that
format.

`boundariesType` specifies the format of the polygon file and must be either
Parquet or CSV. By default, the function uses Parquet format. If only the CSV
file is available, you should provide a vector containing only that format.

`computeMissingMetrics` is set to TRUE by default and enables the computation of
`Area_um` and `AspectRatio` cell metrics from boundary polygons. 
**IMPORTANT**: it is strongly recommended to keep this parameter set to `TRUE` 
as both metrics are required to run `computeQScore()`, even if they are not used
in the final formula.

`keepPolygons` controls whether polygons are retained in the `SpatialExperiment` 
object and is **ONLY** effective when `computeMissingMetrics=TRUE`. While 
missing metrics can be computed without keeping polygons, polygons cannot be 
retained without computing these metrics. However, polygons may also be loaded 
at a later stage (see the [Load polygons](#load-polygons) subsection).

`countsFilePattern` and `polygonsFPattern` define file name patterns (excluding
the file extension) for the expression matrix and polygon files, correspondingly,
since their formats are already specified via `type` and `boundariesType`.

`metadataFPattern` requires the full file name (including extension) of the cell 
metadata file. Only `.csv.gz` and `.parquet` files are supported.

`polygonsCol` defines the name of the column in `colData` that stores polygon
geometries. 

Optionally, FoV information can be retrieved for each cell from the transcript 
file. The `txPattern` parameter specifies the file name pattern for the 
transcript file (Parquet format only), while `addFOVs` controls whether FoV
information is extracted. This option is not mandatory and is therefore set to
`FALSE` by default.

**OF NOTE**, because `readXeniumSPE` is a wrapper function around
`SpatialExperimentIO::readXeniumSXE`, and multiple probe identifier sets could
be available, Gene Symbols are selected and used as row names.

**MERFISH/MERSCOPE**

```{r load-merfish, message=TRUE, warning= TRUE}
merfishFolder <- system.file("extdata", "Merfish_Tiny", package="SpaceTrooper")

mer <- readMerfishSPE(dirName=merfishFolder, 
                      sampleName="Merfish_Tiny",
                      computeMissingMetrics=TRUE,
                      keepPolygons=FALSE,
                      boundariesType=c("parquet"),
                      countmatFPattern="cell_by_gene.csv",
                      metadataFPattern="cell_metadata.csv",
                      polygonsFPattern="cell_boundaries.parquet",
                      coordNames=c("center_x", "center_y"),
                      polygonsCol="polygons")
```

As with the other data-reading functions, `readMerfishSPE` requires the path to 
the directory containing the MERFISH  output files. The minimum set of inputs 
required to run the `SpaceTrooper` pipeline includes:

- expression matrix
- cell metadata
- cell polygons

The expression matrix and cell metadata are typically provided in CSV format. 

Polygon data can be supplied either as a single Parquet file (recommended when 
available) or as multiple HDF5 files. In MERFISH/MERSCOPE experiments, HDF5 
polygon files are generated per FoV and may easily scale to thousands of files, 
making this option computationally intensive and time-consuming. Fortunately, in
more recent datasets, a Parquet file is usually available.

For this reason, it is recommended to override the default setting and provide a
vector containing only `Parquet` to the `boundariesType` parameter.

**IMPORTANT**: please, pay attention to warnings. `computeMissingMetrics` 
parameter must be set to `TRUE` whenever it is desired to compute Quality Score
(QS). In MERFISH experiments, transcript hybridization signals are captured also 
across the z dimension, and cell size is therefore represented by cell volume in 
the provided metadata. 

In contrast, cell area can also be computed by `SpaceTrooper` from polygon
boundaries obtained from the MERFISH experiment but from a single z plane only
(by running `computeMissingMetricsMerfish()` with `useVolume=FALSE`).
That said, cell volume appears to be a more appropriate measure of cell size for
QS computation with MERFISH data.

In the current version of `SpaceTrooper`, when `computeMissingMetrics=TRUE`, the
`volume` is renamed as `Area_um` for compatibility with the `computeQScore`
function, whereas cell area is not computed.

### Load polygons

For all supported technologies, polygons can be loaded after `SpatialExperiment` 
object creation. The function `readAndAddPolygonsToSPE` performs the following
steps:

- reads the cell polygon file;
- standardizes polygon formats and converts them into an `sf` object;
- validates polygon geometries (e.g. checks that polygons are properly closed 
and contain at least four vertices etc.);
- removes cells from the `SpatialExperiment` object if no valid polygon is 
associated with them;
- stores the resulting `sf` object in a `colData` column whose name can be 
specified via the `polygonsCol` argument;

```{r load-poly,  message=TRUE}
# polygon loading
spe <- readAndAddPolygonsToSPE(spe, 
                               polygonsCol="polygons",
                               keepMultiPol=TRUE,
                               boundariesType=c("csv", "HDF5", "parquet")
)

spe$polygons
```

In some technologies, segmentation may produce multi-polygons (indicated as
`TRUE` in the `is_multi` column of the `sf` object shown here), meaning that
more than one polygon is associated with a single cell. This can cause issues
when computing certain derived metrics, particularly the`AspectRatio`. 

By default, the `keepMultiPol` parameter is set to `TRUE`. Under this setting, 
cells associated with multi-polygons are retained in the dataset; however, their
`AspectRatio` is `NA`.  

```{r load-poly-xen,  message=TRUE}
# xenium polygon loading
xen <- readAndAddPolygonsToSPE(xen, boundariesType=c("parquet"))
```

```{r load-poly-mer,  message=TRUE}
# merfish polygon loading
mer <- readAndAddPolygonsToSPE(mer, boundariesType=c("parquet"))
```

```{r load-poly-prot,  message=TRUE}
# protein polygon loading
prot <- readAndAddPolygonsToSPE(prot, boundariesType=c("csv"))
```

Multiple geometries files can be loaded into a `SpatialExperiment` object using
the `readAndAddPolygonsToSPE` function. The `polygonsCol` parameter allows users
to specify distinct column names under which these geometries are stored in
`colData`.

## Quality Control

### Compute QC metrics

This section describes the cell metadata and quality metrics computed by the 
`spatialPerCellQC` function. When inspecting `colData`, additional metadata and 
metrics provided natively by each technology may also be present; these are 
technology-specific and are generally not shared across all platforms. For
details on these additional fields, please refer to the corresponding technology
reference documentation.

```{r analysis-qc-1}
spe <- spatialPerCellQC(spe, rmZeros=TRUE,
        negProbList=c("NegPrb", "Negative", "SystemControl"))

xen <- spatialPerCellQC(xen, rmZeros=TRUE)

mer <- spatialPerCellQC(mer, rmZeros=TRUE)

prot <- spatialPerCellQC(prot, rmZeros=TRUE)
```

```{r analysis-qc-2}
# check added columns to colData
colnames(colData(spe))
```

This function is a wrapper around `scuttle::addPerCellQC()`, hence a subset of
the reported metrics is derived from that function:

- `sample_id`: character string specifying the dataset sample name. This matches
the value provided to the `sampleName` argument in the reading function;
- `sum`/`total`: numeric value representing total probe counts (RNA) or total 
protein intensity per cell. In this context, `sum` and `total` are equivalent;
- `subsets_controlProbeName_sum`: numeric value giving sum of the counts 
(intensities) obtained from each type of control probe (control protein). For 
each one, a separate column is created by replacing "controlProbeName" with the 
corresponding pattern from `negProbList` , e.g. negative control probes (RNA), 
negative codewords (RNA), negative control antibodies (protein), etc.;
- `subsets_controlProbeName_detected`: numeric value indicating the number of 
uniquely detected control probes (proteins) per each type;
- `subsets_controlProbeName_percent`: numeric value representing the percentage 
of counts (intensities) from each type of control probe (protein) relative to
the total probe counts (protein intensities) per cell;
- `control_sum`: numeric value indicating total control probe counts (protein
intensities) per cell. It is the sum of counts (intensities) of all types of
controls;
- `control_detected`: numeric value representing the number of uniquely detected 
control probes (proteins) per cell;
- `target_sum`: numeric value giving target probe counts (protein intensities)
per cell, obtained as `sum` - `control_sum`;
- `target_detected`: numeric value indicating number of uniquely detected target 
probes (proteins) per cell;
- `ctrl_total_ratio`: numeric value representing the ratio between total control
probe counts (protein intensities) and total probe counts (protein intensities)
per cell;
- `log2Ctrl_total_ratio`: numeric value indicating log2-transformed 
`ctrl_total_ratio`
- `Area_um`: numeric value giving cell area in μm² (for MERFISH technology it 
corresponds to `volume` in μm³);
- `dist_border_x`: numeric value representing the distance in pixels from the 
cell to the closest vertical border of its FoV (only available for CosMx);
- `dist_border_y`: numeric value representing the distance in pixels from the 
cell to the closest horizontal border of its FoV (only available for CosMx);
- `dist_border`: numeric value representing the distance from the cell to the
nearest border of its FoV (only available for CosMx), computed as the minimum of 
`dist_border_x` and `dist_border_y`;
- `log2AspectRatio`: numeric value indicating log2-transformed ratio between 
cell maximum length along the x dimension and cell maximum length along the y 
dimension in pixels;
- `SignalDensity`: numeric value indicating the ratio between total probe counts 
(protein intensities) per cell and cell area in µm² (or volume in µm³ for 
MERFISH technology);
- `log2SignalDensity`: log2-transformed `SignalDensity`.

### Compute Quality Score (QS)

This section provides further details about the core component of the method: 
the Quality Score (QS). As already mentioned in
[RNA](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/RNA_tutorial_vignette.html) and [Protein](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/Protein_tutorial_vignette.html) tutorial vignettes,
a random seed should be set to ensure reproducible results, as the method 
involves underlying stochastic processes.

**IMPORTANT**: when working with recent MERSCOPE datasets, or with any dataset 
that uses numeric cell identifiers, it is strongly recommended to convert cell 
IDs to character strings before calling the `computeQScore` function. This 
prevents errors during the selection of training examples when estimating the 
coefficients of the QS formula.

```{r compute-QS,  message=TRUE}
# MERSCOPE cell_id conversion
# spe$cell_id <- as.character(spe$cell_id)

set.seed(1713)

spe <- computeQScore(spe, verbose=TRUE)
```

When `verbose=TRUE`, additional information regarding the underlying steps is 
reported during the estimation of the QS formula coefficients. For this purpose,
bad and good example cells are selected to train a model that has to assign an
appropriate QS to each cell. 

Bad example cells are identified separately for each metric according to several 
criteria, based on outlier detection through a statistical test. The specific 
test applied depends on the distribution of the metric: for asymmetric 
distributions the Medcouple method is used, whereas for symmetric distributions 
the Median Absolute Deviation (MAD) is applied, as implemented in 
`scuttle::isOutlier()` with default parameters. 

For a detailed description of the selection criteria used for each metric,
please refer to the Methods section of the [paper](https://www.biorxiv.org/content/10.64898/2025.12.24.696336v1). 

When inspecting the printed output, the following information is reported:

- The first section summarizes the number of outliers identified for each
metric. The three reported values correspond to outliers detected in the right
tail, the left tail of the distribution, and the remaining non-outlier cells, 
respectively.
Outlier detection is performed on the full dataset after excluding
cells with zero counts, if these were retained when running the
`spatialPerCellQC` function. 
This step is carried out for all metrics, even those that are not ultimately
included in the final QS formula; however, outliers from excluded metrics are
not used for model training.

- The next section reports the number of selected bad example cells, followed by
the number of good example cells, which matches the former. Because the criteria
for selecting good examples are intentionally permissive, these cells are
sampled at random.

- The final section displays the QS formula used for model training and score
computation, along with the estimated coefficients for each retained term and
all pairwise interactions.

**IMPORTANT**: at present, QS computation requires `log2SignalDensity` as a
mandatory metric in the QS formula. If an insufficient number of outliers is
detected for this metric (fewer than 0.1% of the dataset after excluding
zero-count cells), QS computation cannot proceed using the remaining metrics.
In such cases, the provided code will still add a `QScore` column to `colData`.
This column is populated if the minimum requirement is met and contains `NA`
values otherwise.

```{r compute-QS-safe-run,  message=TRUE}
# safe run using tryCatch directly so errors are caught before evaluation
out <- tryCatch(
  list(result=computeQScore(spe), error=NULL),
  error=function(e) list(result=NULL, error=e)
)

if (!is.null(out$error)) {
message("Failed: ", out$error$message)
colData(spe)$QScore <- NA
} else {
spe <- out$result
}
```

### Compute QS with custom formula

In `computeQScore`, the `modelFormula` parameter allows users to specify which
metrics to include in the QS formula among the four currently supported:
`log2SignalDensity`, `Area_um`, `log2AspectRatio`, and
`log2Ctrl_total_ratio`. The supported border-effect expression is
`I(abs(log2AspectRatio) * as.numeric(dist_border < 50))`; aspect ratio and
border effect are available only for CosMx datasets.

The default is `NULL` and the formula is automatically estimated based on the 
available metrics, their outliers and the dataset technology. The most complete 
formula is generated as follows:

 `~(log2SignalDensity + Area_um + I(abs(log2AspectRatio) * as.numeric(dist_border < 50)) + log2Ctrl_total_ratio)^2`

When user-provided, the formula is fitted as supplied. It must include
`log2SignalDensity`; users may select a subset of the other supported metrics,
use an additive formula, or choose interactions with `:` and `*`. SpaceTrooper
does not add or remove terms. Predictor and transformation
validation is separate from formula structure: arbitrary metrics and
transformations such as `log1p(Area_um)` are rejected. The supported base QC
metrics present in the custom formula are also used to compute outliers, define
good and bad training cells, and construct `QScore_train`; supported metrics
omitted from the formula do not contribute to those labels. Interactions alter
the fitted model but do not create additional training metrics. In the supported
CosMx border expression, `log2AspectRatio` is the training QC metric and
`dist_border` is only an auxiliary variable. With `modelFormula=NULL`, the
established default metric-selection and training-label workflow is unchanged.

```{r compute-QS-different-formula,  message=TRUE}

set.seed(1713)

model_formula <- "~ log2Ctrl_total_ratio + log2SignalDensity"

spe <- computeQScore(spe, modelFormula=model_formula, verbose=TRUE)
```

```{r plot-polygons-dbkero}
plotPolygons(spe, colourBy="QScore") + 
    scale_fill_viridis_c(option="plasma")
```

Compared to the plot displayed in [RNA](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/RNA_tutorial_vignette.html) vignette,
QS values are higher for bigger cells and cells located on the FoV borders, as the 
cell size and border effect contributions are not considered in the QS computation.

### Apply a trained QS model

The fitted model is stored in `metadata(spe)$QScore_model` and can be applied
with `applyQScoreModel()` to an object containing the required QC metrics.
Model transfer across datasets is generally discouraged because QS training is
dataset-specific; when it is necessary, use datasets from the same experiment
and technology, or train a supported formula that excludes incompatible terms.

```{r apply-QS-model, message=TRUE}
qs_model <- metadata(spe)$QScore_model
spe <- applyQScoreModel(
    spe,
    qsModel=qs_model,
    scoreName="QScore_transferred"
)
```

The historical `applyQCScoreModel()` interface remains available with a
deprecation warning and retains `qcModel`, `QC_score`, and
`QCScore_model_applied` naming.

### Single metric flagging 

The `SpaceTrooper` package allows the identification of numerically aberrant
cells through outlier detection for each quantitative metric, using the same
procedures applied internally by the `computeQScore` function. This enables
metric-specific cell flagging in a manner analogous to standard scRNA-seq
quality control workflows.

```{r compute-outliers-1,  message=TRUE, warning=TRUE}
# detecting outliers for cell area in Xenium dataset with both Medcouple
# and MAD
xen <- computeSpatialOutlier(xen, computeBy="Area_um", method="both")
```

Pay attention to warnings: setting the method to `both`, applies both the
Medcouple and MAD approaches to identify outliers. Outlier results are stored in
two separate `colData` columns, with names formed by appending `_outlier_mc`
and `_outlier_sc` to the variable name. If a warning indicates that the variable
is symmetric, only the `_outlier_sc` column should be used for downstream
analyses.

```{r compute-outliers-3,  message=TRUE}
spe <- computeSpatialOutlier(spe, computeBy="Mean.CD68", method="both")

table(spe$Mean.CD68_outlier_mc)
```

The three numbers correspond to outliers in the right tail (`HIGH`), in the 
left tail (`LOW`) of the variable distribution, and remaining non-outlier cells
(`NO`). 

## Visualization

Variable distributions can be visualized using the `plotMetricsHist` function,
as demonstrated in the [RNA](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/RNA_tutorial_vignette.html) and [Protein](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/Protein_tutorial_vignette.html) tutorial vignette. 
If the `computeSpatialOutlier` function has already been executed, the resulting
columns can be provided as input to `useFences` parameter to display lines
corresponding to the upper and lower thresholds, used to define `HIGH` and `LOW`
outliers, respectively.

```{r plot-hist-1}
spe <- computeSpatialOutlier(spe, computeBy="log2SignalDensity", method="both")

plotMetricHist(spe, metric="log2SignalDensity", 
               useFences="log2SignalDensity_outlier_mc")
```

The threshold values are indicated in the plot legend.

The `plotZoomFovsMap` function provides a zoomed-in view of selected FoVs, with
cell polygons colored according to the metric specified in the `colourBy` parameter.
In addition, it includes an overview map of all FoVs, analogous to that produced
by `plotCellsFovs` function.

```{r plot-zoom-fov-1}
# plotting FoVs map and zoom in of selected FoV colored by QS
plotZoomFovsMap(spe, fovs=16, colourBy="QScore", mapPointSize=1, 
                fovNumberSize=7, scaleBars=TRUE)
```

For this plot, `scaleBars` parameter controls whether scale bars are displayed
in both plots simultaneously. To show the scale bar in only one of the two
plots, set either `scaleBarMap=FALSE` or `scaleBarPol=FALSE`.

```{r plot-zoom-fov-2}
plotZoomFovsMap(prot, fovs=344, colourBy="Area_um", mapPointSize=1, 
                fovNumberSize=7, scaleBarMap=TRUE, scaleBarPol=FALSE)
```

For the other plots, the `scaleBar` parameter can be set to `FALSE` to hide the
scale bar.

```{r plot-centroids}
plotCentroids(spe, colourBy="QScore", size=3, scaleBar=FALSE) +
    scale_fill_viridis_c(option="plasma") +
    scale_color_viridis_c(option="plasma")
```

```{r plot-polygons}
plotPolygons(prot, colourBy="log2SignalDensity", scaleBar=FALSE) + 
    scale_fill_viridis_c(option="mako")
```

## Session Information

```{r sessionInfo}
sessionInfo()
```

---
title: "CosMx Protein Assay Data Quality Control with SpaceTrooper"
author: 
    - name: "Benedetta Banzi"
    - name: "Dario Righelli"
date: "`r BiocStyle::doc_date()`"
output:
    BiocStyle::html_document:
        toc: true
        toc_float: true
vignette: >
    %\VignetteIndexEntry{CosMx Protein Assay Data Quality Control with SpaceTrooper}
    %\VignetteEncoding{UTF-8}
    %\VignetteEngine{knitr::rmarkdown}
editor_options:
    chunk_output_type: console
---

```{r setup, include=FALSE}
# Set chunk options: suppress echo, messages, and warnings in code output
knitr::opts_chunk$set(echo=TRUE, message=FALSE, warning=FALSE)
```
## Introduction

`SpaceTrooper` is an `R/Bioconductor` package for Quality Control (QC)
of imaging-based spatial transcriptomics and proteomics data. It provides 
multi-platform data harmonization, cell-level QC, and visualization utilities. 
The package leverages [SpatialExperiment](https://www.bioconductor.org/packages/release/bioc/html/SpatialExperiment.html) objects to support data from 
**CosMx**, **Xenium**, and **MERFISH** technologies.


## Quality Control pipeline

In this section, we show how to run the standard `SpaceTrooper` workflow on  
Spatial Proteomics data obtained with CosMx Protein Assay, to perform QC by 
computing the Quality Score (QS). 

For demonstration purposes, we showcase a small subset of the CosMx human tonsil
dataset analyzed also in the [paper](https://www.biorxiv.org/content/10.64898/2025.12.24.696336v1).  

### Load example data

`SpaceTrooper` provides a reading function to load data into a 
`SpatialExperiment` object. For more details on how to correctly specify the 
inputs, please refer to the [SpaceTrooper utilities](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/SpaceTrooper_utilities.html) vignette.

```{r load-cosmx}
library(SpaceTrooper)
library(ggplot2)

protfolder <- system.file( "extdata", "S01_prot", package="SpaceTrooper")
spe <- readCosmxProteinSPE(protfolder, sampleName="CosMx_Protein_Tonsil")

spe
```

### Field of View (FOVs) visualization

In this section, we show how to plot each cell position onto the map of the 
Fields of View (FOVs), whose coordinates are provided only for CosMx technology.

**IMPORTANT**: we have noticed that according to CosMx version, there could be a 
misalignment of FOVs and cell centroids, that can be easily corrected with a 
single line of code. Therefore, it is crucial to generate this plot to check for 
any spatial shift, as such misalignment directly affects the QS computation.

```{r plot-fovs}
# to check misalignment
plotCellsFovs(spe, pointSize=3, pointAlpha=0.7, numberSize=6)
```

If any misalignment is observed, it can be corrected by adjusting the FoV 
coordinates directly. And this is exactly the case. Indeed, FoVs are shifted 
upward by one FoV height (which, for CosMx technology, corresponds to 4,256 
pixels). You can correct this by subtracting that value from the FoVs y 
coordinates:

```{r fov-correction}
# code line for shift correction
metadata(spe)$fov_positions$y_global_px <- metadata(spe)$fov_positions$y_global_px - 4256
```

Rerun the FoV plot to check whether the shift was corrected.

```{r plot-fovs-2}
# check shift correction
plotCellsFovs(spe, pointSize=3, pointAlpha=0.7, numberSize=6)
```

Cell centroids, shown in dark red, now are all contained by FoV boundaries,
hence FoVs and cell centroids are aligned after shift correction.

The dataset is a subset with just a single FoVs, whose number is displayed at 
the center.

When an experiment has multiple FoVs, you can see the map and the topological
organization of the FoVs, together with their numbers.

### Load polygons

In this section we show how to load polygons after `SpatialExperiment` creation, 
only for visualization purposes. They are stored as an `sf` object within 
`colData`.

This step is not mandatory for CosMx, because the pipeline can be executed even
without them. 

```{r load-poly, message=TRUE}
# polygon loading
spe <- readAndAddPolygonsToSPE(spe, boundariesType="csv")
```

Please pay attention to any warnings. Cell polygons with fewer than four 
vertices cannot be handled by the geometry packages used by `SpaceTrooper`. 
Therefore, the corresponding cells are discarded from the `SpatialExperiment` 
object.

### Add QC metrics

The following sections will work the same way for all types of data.

The `spatialPerCellQC` function computes additional metrics per each cell, that
are saved inside the `SpatialExperiment`and accessible through `colData(spe)`. 
It is mandatory to run this function before computing QS.

The `negProbList` parameter is, by default, a vector containing all control probe 
patterns encountered so far across the supported technologies. Because these 
patterns continue to evolve, some may not yet be included. If you find that your
control probe patterns are missing from the default list, you can define a 
custom vector, as shown below.

By default, the function automatically removes 0 count cells, but this can be 
handled with the `rmZeros` parameter.

```{r cosmx-analysis-qc, message=TRUE}
spe <- spatialPerCellQC(spe, rmZeros=TRUE,
        negProbList=c("Ms IgG1", "Rb IgG"))

colnames(colData(spe))
```

Several metrics are added, both derived from protein assay and cell 
morphology. Some of them are directly used to compute QS:

- **log2SignalDensity**: log2-transformed ratio between total protein 
    intensity per cell and cell area in µm²
- **Area_um**: cell area in µm²;
- **log2Ctrl_total_ratio**: log2-transformed ratio between total negative 
    control protein intensity and total total protein intensity per cell;
- **log2AspectRatio**: log2-transformed ratio between cell maximum length along 
    the x dimension and cell maximum length along the y dimension (pixels). This
    metric is taken in absolute value and combined with **dist_border** to 
    consider only cells within 50 pixels from the nearest FoV border (only for 
    CosMx technology).

For a better detailed explanation of the other metrics, please refer to 
the [SpaceTrooper utilities](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/SpaceTrooper_utilities.html) vignette.

### Compute Quality Score

`computeQScore` function calculates QS per each cell. QS combines several
metrics into a formula: **log2SignalDensity**, corresponding to signal density, 
**Area_um**, i.e. cell size, **log2Ctrl_total_ratio**, namely background signal 
and **log2AspectRatio** combined with **dist_border**, which jointly correspond 
to the border effect (only for CosMx technology). 

`glmnet` package is used to estimate the coefficients of the formula, resulting 
in a robust score that captures low-quality cells. During model training, the
selected cells are used as good or bad examples to learn how each term in the 
formula contributes to cell quality.

**IMPORTANT**: please, pay attention to any warning. If a term has too few bad 
examples (fewer than 0.1% of the total number of cells), it is excluded from the 
formula and therefore not used in the QS computation.

The QS (stored as `QScore` in `coldata`) ranges from 0 to 1, with 0 meaning 
low-quality and 1 high-quality.

We are setting a seed to ensure reproducibility in this tutorial, because there 
are stochastic processes underlying QS computation.

```{r cosmx-analysis-score, message=TRUE}
set.seed(1713)

spe <- computeQScore(spe)

format(summary(spe$QScore), scientific=FALSE, digits=4)
```

In this case, all the terms were used as no warning appeared.

Subsequently, it is possible also to assess which cells have a QS lower than a 
certain threshold (default is 0.5) with the following function. It creates a
new column called `low_QScore` inside of `colData`.

```{r cosmx-analysis-score2, message=TRUE}
spe <- computeQScoreFlags(spe, qsThreshold=0.5)

table(spe$low_QScore)
```

Using this threshold, 274 cells are flagged as low-quality. We do not suggest a 
fixed default threshold, but it is advisable to check the QS distribution before
setting any threshold.

### Data visualization

SpaceTrooper comes with several functions to view cells and metrics.

To view the distribution of whatever quantitative metric, `plotMetricHist` comes 
in handy.

```{r plot-hist}
# view quantitative metric distribution
plotMetricHist(spe, metric="QScore")
```

The QS distribution exhibits a left tail starting around 0.75. 

Cell visualization can be obtained by using either centroids (recommended when 
the dataset has a large number of cells) or polygons. 

`plotCentroids` plots cell centroids that can be colored by a certain metric
contained in the `colData` slot, by using the `colourBy` parameter.

Additionally, if you have a palette column in `colData`, containing colors for
each cell, it can be given to `palette` parameter, so that it automatically 
matches the column passed in `colourBy`. As an example, we are using the 
cell types obtained as described in the [paper](https://www.biorxiv.org/content/10.64898/2025.12.24.696336v1) with their own color palette.

```{r plot-centroids-labels}
labf <- system.file(file.path("extdata", "S01_prot",
                            "labels_tiny.tsv"), package="SpaceTrooper")
labs <- read.table(file=labf, sep="\t", header=TRUE, comment.char="")

spe$labels <- as.factor(labs[match(spe$cell_id, labs$cell_id),]$label)
spe$labels_colors <- as.factor(labs[match(spe$cell_id, labs$cell_id),]$lab_color)

plotCentroids(spe, colourBy="labels", size=3, palette="labels_colors")
```

Cell centroids colored by cell types allow to view the spatial distribution of
cell populations in the sample.

When possible, polygon visualization gives a better overview of the 
cells' spatial organization and morphological characteristics.
Polygons are stored as an `sf` object within `colData`, so they can be
viewed using standard `ggplot2` functions. For CosMx technology, the polygons
must be explicitly loaded as described in [Load polygons](#Load polygons).
SpaceTrooper provides `plotPolygons` function that works just like 
`plotCentroids` but takes polygons instead of centroids.

```{r plot-polygons-fov-1}
plotPolygons(spe, colourBy="log2SignalDensity")
plotPolygons(spe, colourBy="Area_um")
plotPolygons(spe, colourBy="log2Ctrl_total_ratio")
plotPolygons(spe, colourBy="log2AspectRatio")
```

We can see in `yellow` and `darkviolet` that there are few cells with extreme 
values of either `log2SignalDensity`, `Area_um`, `log2Ctrl_total_ratio` and 
`log2AspectRatio`.

Since all plotting functions are based on `ggplot2`, you can easily customize 
the graphical outputs by adding standard ggplot2 components.

```{r plot-polygons-fov-2}
plotPolygons(spe, colourBy="QScore") + 
    scale_fill_viridis_c(option="plasma")
```

We can see that the QS is able to detect all the aspects highlighted by 
`log2SignalDensity`,`Area_um`, `log2Ctrl_total_ratio` or 
`log2AspectRatio`. Cells that showed either lower signal 
density, bigger size, higher background signal or border effect also display 
low QS (darker color).

It's up to the user to choose an appropriate threshold to flag cells according 
to the observed QS distribution. 

```{r plot-polygons-fov-3}
plotPolygons(spe, colourBy="low_QScore") + 
    scale_fill_manual(values=c("TRUE"="red", "FALSE"="#c0c8cf"))
```

You can rerun`computeQScoreFlags` to check how many cells would be flagged 
using another threshold.

```{r cosmx-analysis-score3, message=TRUE}
spe <- computeQScoreFlags(spe, qsThreshold=0.75)

table(spe$low_QScore)
```

Using this threshold, 527 cells are flagged as low-quality.

```{r plot-polygons-fov-4}
plotPolygons(spe, colourBy="low_QScore") + 
    scale_fill_manual(values=c("TRUE"="red", "FALSE"="#c0c8cf"))
```

The threshold is more stringent and more cells are flagged compared to the
previous one. It's up to you to choose a threshold that better suits your
analysis needs.

## Conclusion

In this vignette, we explored the main functionalities of the `SpaceTrooper`
package for imaging-based spatial omics data QC. 

For further insights on SpaceTrooper package usage, please refer to the [SpaceTrooper utilities](https://bioconductor.org/packages/devel/bioc/vignettes/SpaceTrooper/inst/doc/SpaceTrooper_utilities.html) vignette.

Main steps shown are:

- data loading: CosMx Protein Assay
- Quality Control: 
    - per-cell QC metrics
    - Quality Score: a score combining **signal density**, **cell size**, 
    **background noise** and  **border effect**
- visualization: 
    - centroids: with ggplot2
    - polygons: sf + ggplot2

## Session Information

```{r sessionInfo}
sessionInfo()
```

---
title: "Spatial Data Quality Control on Protein data"
author: "Dario Righelli"
date: "`r BiocStyle::doc_date()`"
output:
    BiocStyle::html_document:
        toc: true
        toc_float: true
vignette: >
    %\VignetteIndexEntry{Spatial Data Quality Control on Protein data}
    %\VignetteEncoding{UTF-8}
    %\VignetteEngine{knitr::rmarkdown}
editor_options:
    chunk_output_type: console
---

```{r setup, include=FALSE}
# Set chunk options: suppress echo, messages, and warnings in code output
knitr::opts_chunk$set(echo = TRUE, message = FALSE, warning = FALSE)
```

# Introduction

This vignette introduces the `SpaceTrooper` package for preprocessing and 
quality control of imaging-based spatial omics
from platforms like **CosMx** on Protein assay. 

# Installation

To install `SpaceTrooper`, use the following commands:

```{r install, eval=FALSE}
# Install BiocManager if not already installed, then install SpaceTrooper
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")
BiocManager::install("drighelli/SpaceTrooper")
```

# Data Loading

In this section, we load data from various platforms using the package's
functions. The goal is to provide a uniform `SpatialExperiment` object across
all technologies, allowing for consistent QC analysis.

The functions in `SpaceTrooper` compute missing metrics as needed and allow
for the inclusion of polygons with the `keep_polygons` argument. This stores
polygons in the `colData` of the `SpatialExperiment`.


```{r load-data}
# Load the SpaceTrooper library
library(SpaceTrooper)

# Load Xenium data into a Spatial Experiment object (SPE)
protfolder <- system.file( "extdata", "S01_prot", package="SpaceTrooper")
(spe <- readCosmxProteinSPE(protfolder))
colData(spe)
```

# Data Analysis for CosMx protein

The package offers several functions for spatial data analysis, including
quality control and visualization.

This tutorial focuses on CosMx protein data, which provides Fields of View 
(FoVs) with cell identifiers. Note that FoVs are unique to CosMx.

Additionally, even if not tested, the same approach can be extended on Akoya 
CODEX data, as far as a `SpatialExperiment` object is created.
Polygons can be loaded later if needed.

# Field of Views (FOVs) Visualization

The `plotCellsFovs` function shows a map of the FoVs within an experiment. 
This plot is specific to CosMx data and uses cell centroids.

Please keep in mind, that this specific experiment had unaligned `fov_positions`
and cell centroids positions.
An alignment approach, can be found at the end of the `scripts/datacreation.R`
file.

```{r plot-fovs}
# Plot the cells within their respective Field of Views (FOVs)
plotCellsFovs(spe)
```

Because the dataset is a subset of just one Field of View of the original 
experiment, we are able to see the identifier of the FoV in black and the 
centroids of the cells in purple.

When an experiment has multiple FoVs, you can see the map and the topological
organization of the FoVs, together with their identifiers.

# Quality control 

The `spatialPerCellQC` function, inspired by `scater::addPerCellQC`, computes
additional metrics for each cell in the `SpatialExperiment`. It also allows for
the detection of negative control probes, which is crucial for QC.

By default, it automatically removes 0 counts cells, but this can be handled 
with the `rmZeros` argument.

Here, for transparency, we specified the `negProbList` for CosMx protein assays,
but the algorithm has already a set of negative probes for the mostly used 
probes in multiple technologies.
Notice that despite the same approach can be applied to CODEX data, it is not
provided a list of negative probes for this technology, so the user needs to 
specify them.

```{r cosmx-analysis-qc}
# Perform per-cell quality control checks
spe <- spatialPerCellQC(spe, negProbList=c("Ms IgG1", "Rb IgG"))
names(colData(spe))
```

# Metrics Histograms

You can investigate individual metrics by viewing their histograms. For
outliers, use the `use_fences` argument to display the fences computed by
`computeSpatialOutlier` (see next chunk).

```{r plot-metrics}
# Plot a histogram of counts (sum)
plotMetricHist(spe, metric="sum")
# Plot a histogram of cell areas (Area_um)
plotMetricHist(spe, metric="Area_um")
# Plot a histogram of proportion of counts respect to the cell area in micron 
plotMetricHist(spe, metric="log2CountArea")
# Plot a histogram of proportion of negative probe counts respect to the total
# counts in cells
plotMetricHist(spe, metric="log2Ctrl_total_ratio")
```

These plots show, respectively, the distributions of the total counts (`sum`),
of the Area in micron (`Area_um`), the relationship between the counts and 
the Area of each cell (`log2CountArea`) and the proportion between the 
negative probes counts and the total counts of each cell 
(`log2Ctrl_total_ratio`).

# Spatial Outlier Detection 

Spatial outlier detection is another critical step in QC. While the flag
score addresses some metrics, other outlier detection methods may be needed.

The `computeSpatialOutlier` function allows the computation of the medcouple
statistics on a specified metric (`compute_by` argument). 
The medcouple is specifically designed for asymmetric distributions, indeed the
function stamps a warning message when this requisite is not satisfied.
It can also use `scuttle::isOutlier` for symmetric distributions. 
The `method` argument supports `mc`, `scuttle`, or `both`.

This outlier detection approach can be used to decide if and which cells can be 
discarded on a singular metric.

```{r cosmx-analysis-outlier, warning=FALSE}
# Identify spatial outliers based on cell area (Area_um)
spe <- computeSpatialOutlier(spe, computeBy="Area_um", method="both")

# Identify spatial outliers based on mean DAPI intensity
spe <- computeSpatialOutlier(spe, computeBy="Mean.DAPI", method="both")
names(colData(spe))
```

If we computed outliers with the `computeSpatialOutlier` function, we can also 
visualize which fences have been used to create the filter on the cells.

```{r plot-metrics-2}
# Plot a histogram with fences to identify outliers using the medcouple
plotMetricHist(spe, metric="Area_um", useFences="Area_um_outlier_mc")

# Plot a histogram with fences to identify outliers using scuttle
plotMetricHist(spe, metric="Area_um", useFences="Area_um_outlier_sc")

# Plot a histogram with fences to identify outliers using the medcouple
plotMetricHist(spe, metric="Mean.DAPI", useFences="Mean.DAPI_outlier_mc")

# Plot a histogram with fences to identify outliers using scuttle
plotMetricHist(spe, metric="Mean.DAPI", useFences="Mean.DAPI_outlier_sc")
```

We visualize the fences computed with medcouple and scuttle outlier 
detection approaches, to directly inspect differences and the amount of detected
outlier each method detected.

If we want, we can already use these fences to remove the computed outliers.


# The Quality Score

Next, we use `computeQCScore` to calculate a flag score based on previously
computed metrics. 
The quality score combines **transcript counts** related to 
**cell area**, the **aspect ratio** of each cell, and its 
**distance from the FoV border** (only for CosMx, this last one is not used 
otherwise).

See the `help(computeQCScore)` details section for additional details.

```{r cosmx-analysis-score}
# Calculate quality scores for each cell
spe <- computeQCScore(spe)
names(colData(spe))
```

Logical filters can then be computed using `computeQScoreFlags`, which requires
thresholds for various metrics. Currently, the function considers:

* **Quality Score (`qsThreshold`)**: Cells with scores below this threshold
    (default 0.5) are flagged for exclusion. This value can be used to indicate 
    the quantile for the filtering when setting the `useQSQuantiles` argument 
    to `TRUE`.
    
* **Quality Score Quantiles (`useQSQuantiles`)**: Option to filter based on
    quantiles (default `FALSE`).


```{r cosmx-analysis-score2}
# Compute flags to identify cells for filtering
spe <- computeQCScoreFlags(spe, qsThreshold=0.5)
names(colData(spe))
table(spe$low_qcscore)
```

We detected 61 cells to be removed.

# Additional metrics to filter out cells

While for other metrics such as the total counts and the negative prob ratio, 
the function `computeThresholdFlags` considers:

* **Total Counts (`totalThreshold`)**: Minimum count threshold (default 0).

* **Negative Probe Ratio (`ctrlTotRatioThreshold`)**: Minimum ratio of
    negative probes to total counts (default 0.1).
    
```{r thresholds-outliers}
spe <- computeThresholdFlags(spe, totalThreshold=0, 
                                ctrlTotRatioThreshold=0.1)
table(spe$threshold_flags)
```

In this example, we don't find any cell to be removed.
    

# Adding Polygon and Visualization

To better understand the quality score values we start to load the polygons, 
giving us a better overview of the cells characteristics.

We can load and add polygons to the SPE object using the following functions.
Each technology has its own `readPolygons` function to standardize the
loaded `sf` object and handle different file types.

```{r plot-polygons}
# Read polygon data associated with cells in the SPE
# the polygon file path is stored in the spe metadata
pols <- readPolygonsCosmx(metadata(spe)$polygons)

# Add the polygon data to the SPE object
spe <- addPolygonsToSPE(spe, pols)
```

Once the polygons are stored in an `sf` object within `colData`, they can be
visualized using functions based on the `ggplot2` library.

```{r plot-polygons-fov-1}
# Plot the polygons of the selected cells
plotPolygons(spe, bgColor="white")
```

Showing the cells on a white background for better visualization.

```{r plot-polygons-fov-2}
# Plot polygons colored by cell area
plotPolygons(spe, colourBy="log2AspectRatio")
plotPolygons(spe, colourBy="Area_um")
```

We can see in `yellow` and `darkviolet` that there are few cells with extreme 
values of `log2AspectRatio` and `Area:um` in micron.

```{r plot-polygons-fov-3}
plotPolygons(spe, colourBy="QC_score")
plotPolygons(spe, colourBy="low_qcscore")
```

We can see that the quality score is able to detect both these aspects and 
highlight the cells that are mostly isolated on the FoV border or showing a
weird confomation.

We always recommend to be aware of the cell populations in the under-study
context, before proceeding to remove the detected cells.

# Fov Zoom and Map

The `plotZoomFovsMap` function allows you to visualize a map of the FoVs with
a zoom-in of selected FoVs, colored by the `colour_by` argument.

```{r plot-zoom-fov}
plotZoomFovsMap(spe, fovs=c(201), colourBy="QC_score")
plotZoomFovsMap(spe, fovs=c(201), colourBy="low_qcscore")
```

We see on the left side the map of all the FoVs (only the FoV 16 in this case),
together with the poligons on the right, coloured by the quality score.
Allowing us to have a better view of a specific tissue area in the whole 
experiment.

# Conclusion

In this vignette, we explored the main functionalities of the `SpaceTrooper`
package for spatial data analysis. 
Main steps shown are:
* data and polygons loading for CosMx Protein, CosMx 
* quality control: 
    + outlier detection: medcouple and scuttle MAD
    + flag score: a score combining **transcript counts**, **cell area**, 
    **aspect ratio** and  **distance from the FoV border**
* visualization: 
    + centroids: with ggplot2
    + polygons: sf + ggplot2

# Session Information

```{r sessionInfo}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotUtils.R
\name{.light_theme}
\alias{.light_theme}
\title{.light_theme}
\usage{
.light_theme(fillColor = "white", foreColor = "black")
}
\value{
a ggplot2 theme object
}
\description{
internal function to setup the white background theme for the First Filter
plot
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotUtils.R
\name{.negative_image_theme}
\alias{.negative_image_theme}
\title{.negative_image_theme}
\usage{
.negative_image_theme(fillColor = "black", foreColor = "white")
}
\arguments{
\item{fillColor}{color to fill the element_rect (default is "black")}

\item{foreColor}{color for all the other elements (default is "white")}
}
\value{
a ggplot2 theme object
}
\description{
internal function to setup the theme for the negative background for
negative plots
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotUtils.R
\name{createPaletteFromColData}
\alias{createPaletteFromColData}
\title{createPaletteFromColData}
\usage{
createPaletteFromColData(spe, paletteNames, paletteColors)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial transcriptomics data.}

\item{paletteNames}{A character string specifying the column in
`colData(spe)` to be used for the names in the palette.}

\item{paletteColors}{A character string specifying the column in
`colData(spe)` to be used for the colors in the palette.}
}
\value{
A character vector representing the palette mapping, where each
element is a string in the format `"name=color"`.
}
\description{
Create a Palette from colData in a SpatialExperiment Object

This function generates a palette mapping based on specified columns in
the `colData` of a `SpatialExperiment` object.
}
\details{
The function creates a new palette based on the unique combinations
of values in the specified `paletteNames` and `paletteColors` columns in
`colData(spe)`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotUtils.R
\name{.fov_image_theme}
\alias{.fov_image_theme}
\title{.fov_image_theme}
\usage{
.fov_image_theme(backColor = "black", backBorder = NA, titleCol = "white")
}
\arguments{
\item{backColor}{not used}

\item{backBorder}{color for the borders of the background (default=NA)}

\item{titleCol}{character indicating the color of the title}
}
\value{
a ggplot2 theme object
}
\description{
internal function to setup the theme for the fov background on the whole
image
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readCosMx.R
\name{updateCosmxProteinSPE}
\alias{updateCosmxProteinSPE}
\title{updateCosmxProteinSPE}
\usage{
updateCosmxProteinSPE(
  spe,
  dirName,
  sampleName = "sample01",
  coordNames = c("CenterX_global_px", "CenterY_global_px"),
  countMatFPattern = "exprMat_file.csv",
  metadataFPattern = "metadata_file.csv",
  polygonsFPattern = "polygons.csv",
  fovPosFPattern = "fov_positions_file.csv",
  fovdims = c(xdim = 4256, ydim = 4256)
)
}
\arguments{
\item{spe}{SpatialExperiment object.}

\item{dirName}{Directory containing CosMx Protein data files.}

\item{sampleName}{Character sample ID. Default \code{"sample01"}.}

\item{coordNames}{Character vector of length two indicating coordinate
column names in the per-cell metadata. Default
\code{c("CenterX_global_px","CenterY_global_px")}.}

\item{countMatFPattern}{Character pattern for the counts matrix file.}

\item{metadataFPattern}{Character pattern for the single-cell metadata file.}

\item{polygonsFPattern}{Character pattern for the polygon file(s).}

\item{fovPosFPattern}{Character pattern for the FOV positions file.}

\item{fovdims}{Named numeric vector with FOV size in pixels.}
}
\value{
SpatialExperiment object with updated technology metadata.
}
\description{
Update a SpatialExperiment object corresponding to Nanostring CosMx
Protein data by adding metadata identifying the technology and
optionally passing through file-location parameters.
}
\details{
This function sets \code{metadata(spe)$technology <- "Nanostring_CosMx_Protein"}.
It does not modify other assay or metadata components.
}
\examples{
protfolder <- system.file( "extdata", "S01_prot", package="SpaceTrooper")
spe <- SpatialExperimentIO::readCosmxSXE(dirName=protfolder,
    addParquetPaths=FALSE)
spe <- updateCosmxProteinSPE(spe, protfolder, sampleName="cosmx_prots")
}
\seealso{
readCosmxProteinSPE, readCosmxSPE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{.addPolygonsToCD}
\alias{.addPolygonsToCD}
\title{.addPolygonsToCD}
\usage{
.addPolygonsToCD(cd, polygons, polygonsCol = "polygons")
}
\arguments{
\item{cd}{A DataFrame containing at least `fov` and `cellID` columns.}

\item{polygons}{An sf object with matching `fov` and `cellID` columns.}

\item{polygonsCol}{character indicating the name of the polygons column to
add into the colData (default is `polygons`).}
}
\value{
A DataFrame identical to `cd`, but row‑subset to cells present in
`polygons` and with a new `polygons` list‑column of sf geometries.
}
\description{
This function enriches a DataFrame (e.g., from colData) with matching polygon
geometries.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeTrainDF}
\alias{computeTrainDF}
\title{computeTrainDF}
\usage{
computeTrainDF(colData, formulaVars, tech, verbose = FALSE)
}
\arguments{
\item{colData}{A per-cell metadata table. Typically
`as.data.frame(colData(spe))`. Must include at least:
`cell_id`, raw metric columns named as in `formulaVars` (e.g.
`log2SignalDensity`, `Area_um`, `log2Ctrl_total_ratio`,
optionally `log2AspectRatio` and `dist_border`), and the corresponding outlier-label
columns referenced by `formulaVars`.}

\item{formulaVars}{A named character vector mapping variable name to
its outlier label column name, e.g.
`c(log2SignalDensity="log2SignalDensity_outlier_train", ...)`.}

\item{tech}{Character string with the acquisition technology. Used to
enable CosMx-specific inclusion of border effect. Expected
values include `"Nanostring_CosMx"` or `"Nanostring_CosMx_Protein"`.}

\item{verbose}{`logical(1)` (default \code{FALSE})
If \code{TRUE}, prints the number of “bad” and “good” cells selected.}
}
\value{
A \code{data.frame} with one row per cell, including:
  \code{QScore_train} and its compatibility alias
  \code{qcscore_train} (0/1) indicating “bad” vs “good”,
  relevant \code{colData} columns used for modeling.
  Deduplicates and down-samples “good” cells to match the number of “bad” cells.
}
\description{
Builds a balanced training data frame from a SpatialExperiment.

\code{computeTrainDF} takes a \code{SpatialExperiment} object
and assembles a balanced training set of “good” vs “bad” cells for
subsequent model fitting.
}
\details{
The function builds a training set using the variables specified
in the `metadata` of the `SpatialExperiment` object.
}
\examples{
example(spatialPerCellQC)
spe <- computeOutliersQScore(spe)
spe <- checkOutliers(spe)
df_train <- computeTrainDF(colData(spe), metadata(spe)$formula_variables,
    metadata(spe)$technology)
table(df_train$QScore_train)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{getFencesOutlier}
\alias{getFencesOutlier}
\title{getFencesOutlier}
\usage{
getFencesOutlier(
  spe,
  fencesOf,
  highLow = c("both", "lower", "higher"),
  decimalRound = NULL
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object containing spatial transcriptomics
data.}

\item{fencesOf}{A character string specifying the name of the column in
`colData(spe)` from which to extract the fence values. This column should
contain an `outlier.filter` object (see `computeSpatialOutlier```).}

\item{highLow}{character indicating which fence to get if "higher", "lower"
or "both" (default is "both").}

\item{decimalRound}{An optional integer specifying the number of decimal
places to which the fence values should be rounded. If `NULL`, no rounding is
 applied. Default is `NULL`.}
}
\value{
A numeric vector containing the lower and upper threshold values
extracted from the specified column.
}
\description{
Retrieve Threshold (Fence) Values from a SpatialExperiment Object

This function extracts the threshold values, also known as fences,
from a specified column in the `colData` of a `SpatialExperiment` object.
}
\examples{
example(computeSpatialOutlier)
getFencesOutlier(spe, fencesOf="log2CountArea_outlier_mc")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeQScore}
\alias{computeQScore}
\title{computeQScore}
\usage{
computeQScore(spe, bestLambda = NULL, modelFormula = NULL, verbose = FALSE)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial omics data.}

\item{bestLambda}{the best lambda typically computed using `computeLambda`.}

\item{modelFormula}{A one-sided formula or character string. If `NULL`, the
default formula is generated from available supported metrics and their
outliers. If supplied, its supported terms and interactions are fitted
without rewriting. See Details.}

\item{verbose}{logical for having a verbose output. Default is FALSE.}
}
\value{
The `SpatialExperiment` object with added Quality Score in `colData`.
}
\description{
Computes Quality Score (QS) and automatically define weights of the QS model
terms through glm training. This function computes QS with a formula
defined on the metrics "log2SignalDensity", "Area_um",
"log2AspectRatio", "log2Ctrl_total_ratio", as computed from
the `spatialPerCellQC` function.
It automatically computes the number of outliers for each available
metric, which are needed for the glm training.
See Details for further information.
}
\details{
For all the technologies, the QS formula can contain
only the following terms at most:

Quality Score ~ signal density - size - border effect - background signal 

Where signal density is the total counts/protein intensities-to-size ratio, 
border effect is represented by the aspect ratio (i.e. width/height of the cell) 
and is considered only for cells with a distance from the FOV border less than 
50 px and only for CosMx datasets. Background signal is represented by the ratio 
of control probe counts to total counts (protein intensities) and cell size is 
the area for CosMx and Xenium, while it is the volume for Merfish.
For each couple of variables interaction terms are computed.

Note that the function is responsible for automatically including/excluding
metrics in the formula based on their availability in the `colData` of the
`SpatialExperiment` object.

For the default formula, inclusion of metrics also depends on the number of
available outliers. If the number of outliers for a metric is less than
0.1\% of the dataset, that metric is excluded from the default QS formula.

- Model fitting: ridge (L2) logistic regression is fitted (via `glmnet`) on
the balanced training set. The function uses `trainModel()` for fitting
and `computeLambda()` to select lambda unless `bestLambda` is supplied.

- Lambda details: because of the randomness in the training set selection,
results may vary. The easiest way to ensure reproducibility is to let lambda 
be computed internally, just set a seed with `set.seed()` before running 
`computeQScore`. Otherwise, a fixed value of lambda previously computed with 
`computeLambda` preceeded by `computeTrainDF` and `getModelFormula` can be set.

- Model formula details: the most complete default formula is:
`~(log2SignalDensity + Area_um + I(abs(log2AspectRatio) * as.numeric(dist_border < 50)) + log2Ctrl_total_ratio)^2`.
A user-supplied formula is fitted as supplied: additive structures, selected
interactions, `:` and `*` are preserved. The formula must include
`log2SignalDensity` and may use any subset of the other supported predictors:
`Area_um`,
`log2AspectRatio`, and `log2Ctrl_total_ratio`. Arbitrary predictors and
transformations are rejected. The supported CosMx-only border expression is
`I(abs(log2AspectRatio) * as.numeric(dist_border < 50))`.

When a custom `modelFormula` is supplied, the supported base QC metrics in
that formula are also used to compute outliers, define good and bad training
cells, and construct `QScore_train`. Interactions do not create additional
training metrics. In the supported border expression, `dist_border` is an
auxiliary variable and `log2AspectRatio` is the corresponding training QC
metric. Supported metrics omitted from the custom formula do not contribute
to training-label construction. When `modelFormula=NULL`, the established
default metric-selection and training-label workflow is unchanged.

The computed model output is stored in `metadata(spe)$QScore_model`.
For inspection and model coefficient transfer between datasets
see also \code{\link{applyQScoreModel}}. However, transferring models between
datasets is not recommended. As described in the paper, QS model training is
dataset-specific and does not generalize well across datasets. Moreover,
this step is computationally efficient. Please, refer to the paper for  
details on scalability and coefficient transfer.
}
\examples{
example(spatialPerCellQC)
set.seed(1998)
spe <- computeQScore(spe)
summary(spe$QScore)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{computeAreaFromPolygons}
\alias{computeAreaFromPolygons}
\title{computeAreaFromPolygons}
\usage{
computeAreaFromPolygons(polygons)
}
\arguments{
\item{polygons}{An `sf` object containing polygon data.}
}
\value{
A `numeric` vector with the area information.
}
\description{
This function computes the area from polygon data.
}
\examples{
example(readPolygonsMerfish)
area <- computeAreaFromPolygons(polygons)
area
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readXenium.R
\name{dot-addFovFromTx}
\alias{dot-addFovFromTx}
\alias{.addFovFromTx}
\title{.addFovFromTx}
\usage{
.addFovFromTx(txFile, colData)
}
\arguments{
\item{txFile}{`character(1)` path to a Xenium Output tx file.}

\item{colData}{A `DataFrame` containing the `colData` for the Xenium dataset.}
}
\value{
A `DataFrame` containing the updated `colData` with FOV information.
}
\description{
Add FOV information from transcript file to cell metadata.

This function retrieves FOV information from transcript file and appends
the data to the resulting `colData`.
}
\details{
The function reads the transcript file then groups it by cell_id
and merges the FOV information to the cell metadata in `colData`. Only
parquet file is supported for this operation
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{readPolygonsXenium}
\alias{readPolygonsXenium}
\title{readPolygonsXenium}
\usage{
readPolygonsXenium(
  polygonsFile,
  type = c("parquet", "csv"),
  x = "vertex_x",
  y = "vertex_y",
  keepMultiPol = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{polygonsFile}{A character string specifying the file path to the
polygon data.}

\item{type}{A character string specifying the file type ("parquet" or "csv").
Default is parquet.}

\item{x}{A character string specifying the x-coordinate column.}

\item{y}{A character string specifying the y-coordinate column.}

\item{keepMultiPol}{A logical value indicating whether to keep multipolygons.}

\item{verbose}{A logical value indicating whether to print additional
information.}
}
\value{
An `sf` object containing the Xenium polygon data.
}
\description{
This function reads polygon data specific to Xenium technology.
}
\examples{
example(readXeniumSPE)
polygons <- readPolygonsXenium(metadata(spe)$polygons, type="parquet")
polygons
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialQCPlots.R
\name{plotCentroids}
\alias{plotCentroids}
\title{plotCentroids}
\usage{
plotCentroids(
  spe,
  colourBy = NULL,
  colourLog = FALSE,
  sampleId = unique(spe$sample_id),
  isNegativeProbe = FALSE,
  palette = NULL,
  pointCol = "darkmagenta",
  size = 0.05,
  alpha = 0.8,
  aspectRatio = 1
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object containing spatial
transcriptomics data.}

\item{colourBy}{An optional character string specifying the column in
`colData(spe)` to use for coloring the points. If `NULL`, all points will be
colored the same.}

\item{colourLog}{Logical to log-transform the data to enhance visualization
(Default is FALSE).}

\item{sampleId}{A character string specifying the sample identifier to be
used as the plot title. (Default is the unique sample ID from `spe`)}

\item{isNegativeProbe}{A logical value indicating whether to apply a custom
color gradient for negative probe data. (Default is `FALSE`)}

\item{palette}{A vector of colors to be used as a custom palette. For
categorical data, this should be a vector of colors with the same length as
the number of levels in `colourBy`. For continuous data, this should be a
vector of colors used to create a gradient.}

\item{pointCol}{A character string specifying the color of the points when
`colourBy` is `NULL`. (Default is `"darkmagenta"`)}

\item{size}{A numeric value specifying the size of the points. (Default is
`0.05`)}

\item{alpha}{A numeric value specifying the transparency level of the points.
(Default is `0.2`)}

\item{aspectRatio}{A numeric value specifying the aspect ratio of the plot.
(Default is `1`)}
}
\value{
A `ggplot` object representing the spatial coordinates plot of
polygon centroids.
}
\description{
Plot Spatial Coordinates for a SpatialExperiment Object
This function generates a ggplot of spatial coordinates from a
`SpatialExperiment` object, optionally coloring the points by a specified
column in `colData`.
}
\examples{
example(readCosmxSPE)
g <- plotCentroids(spe, colourBy="Mean.DAPI")
print(g)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{addPolygonsToSPE}
\alias{addPolygonsToSPE}
\title{addPolygonsToSPE}
\usage{
addPolygonsToSPE(spe, polygons, polygonsCol = "polygons")
}
\arguments{
\item{spe}{A `SpatialExperiment` object to which polygons will be added.}

\item{polygons}{An `sf` object containing the polygon data.}

\item{polygonsCol}{character indicating the name of the polygons column to
add into the colData (default is `polygons`).}
}
\value{
The `SpatialExperiment` object with polygons added to the `colData`.
}
\description{
This function adds polygon data to a `SpatialExperiment` object.
}
\examples{
example(readCosmxSPE)
polygons <- readPolygonsCosmx(metadata(spe)$polygons)
spe <- addPolygonsToSPE(spe, polygons)
spe$polygons
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{dot-computeCosmxProteinTrainSet}
\alias{dot-computeCosmxProteinTrainSet}
\alias{.computeCosmxProteinTrainSet}
\title{.computeCosmxProteinTrainSet}
\usage{
.computeCosmxProteinTrainSet(spe)
}
\arguments{
\item{spe}{\code{SpatialExperiment}}
}
\value{
A list with elements \code{bad} and \code{good}, each a data.frame
with \code{qscore_train} and (for “good”) an \code{is_a_bad_boy} flag.
}
\description{
Internal: Build Training Set for CosMx-Protein
Splits a SpatialExperiment into “bad” vs “good” cells based on
outliers in aspect ratio near tissue border or low count area.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{.computeBorderDistanceCosMx}
\alias{.computeBorderDistanceCosMx}
\title{.computeBorderDistanceCosMx}
\usage{
.computeBorderDistanceCosMx(
  spe,
  xwindim = metadata(spe)$fov_dim[["xdim"]],
  ywindim = metadata(spe)$fov_dim[["ydim"]]
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with CosMx data.}

\item{xwindim}{Width of FOV in x (default from `metadata(spe)$fov_dim`).}

\item{ywindim}{Height of FOV in y (default from `metadata(spe)$fov_dim`).}
}
\value{
A `SpatialExperiment` object with `dist_border` columns in
`colData`.
}
\description{
Calculates the minimum distance of each cell to the field‐of‐view border
and adds it to `colData`.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readXenium.R
\name{readXeniumSPE}
\alias{readXeniumSPE}
\title{Load data from a 10x Genomics Xenium experiment}
\usage{
readXeniumSPE(
  dirName,
  sampleName = "sample01",
  type = c("HDF5", "sparse"),
  coordNames = c("x_centroid", "y_centroid"),
  boundariesType = c("parquet", "csv"),
  computeMissingMetrics = TRUE,
  keepPolygons = FALSE,
  countsFilePattern = "cell_feature_matrix",
  metadataFPattern = "cells.csv.gz",
  polygonsFPattern = "cell_boundaries",
  polygonsCol = "polygons",
  txPattern = "transcripts",
  addFOVs = FALSE
)
}
\arguments{
\item{dirName}{`character(1)`
Path to a Xenium Output Bundle directory.}

\item{sampleName}{`character(1)`
Sample identifier to assign to `sample_id`. Default: `"sample01"`.}

\item{type}{`character(1)`
One of `"HDF5"` or `"sparse"`; method to read the feature matrix.}

\item{coordNames}{`character(2)`
Names of X/Y spatial coordinate columns. Default:
`c("x_centroid", "y_centroid")`.}

\item{boundariesType}{`character(1)`
One of `"parquet"` or `"csv"`; format of the polygon file.}

\item{computeMissingMetrics}{`logical(1)`
If `TRUE`, compute area and aspect‐ratio from boundary polygons.}

\item{keepPolygons}{`logical(1)`
If `TRUE`, append raw polygon geometries to `colData`.}

\item{countsFilePattern}{`character(1)`
Pattern to locate the feature matrix file. Default:
`"cell_feature_matrix"`.}

\item{metadataFPattern}{`character(1)`
Pattern to locate the cell metadata file. Default: `"cells"`.}

\item{polygonsFPattern}{`character(1)`
Pattern to locate the cell boundaries file. Default:
`"cell_boundaries"`.}

\item{polygonsCol}{`character(1)`
Name of the polygons column to add to `colData`. Default:
`"polygons"`.}

\item{txPattern}{`character(1)`
Pattern (base filename, without extension) to locate the transcript file
(usually a `.parquet` file) from which to extract Field-Of-View (FOV)
information for each cell. Default: `"transcripts"`.}

\item{addFOVs}{`logical(1)`
If `TRUE`, extract Field-Of-View (FOV) information from the transcript file
(as located by `txPattern`) and append it to cell metadata (`colData`).
Default: `FALSE`.}
}
\value{
A [`SpatialExperiment`] object with assays, `colData`, spatial
  coordinates, and `metadata$polygons` & `metadata$technology`.
}
\description{
Creates a [`SpatialExperiment`] from an unzipped Xenium Output Bundle
directory containing spatial gene expression data.
}
\details{
readXeniumSPE


Expects the unzipped bundle to contain an `outs/` folder with:
- `cell_feature_matrix.h5` or `cell_feature_matrix/`
- `cells.csv.gz`
}
\examples{
xepath <- system.file(
  "extdata", "Xenium_small", package = "SpaceTrooper"
)
(spe <- readXeniumSPE(
  dirName = xepath,
  keepPolygons = TRUE
))
}
\author{
Dario Righelli, Benedetta Banzi
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialQCPlots.R
\name{plotMetricHist}
\alias{plotMetricHist}
\title{plotMetricHist}
\usage{
plotMetricHist(
  spe,
  metric,
  fillColor = "#c0c8cf",
  useFences = NULL,
  fencesColors = c(lower = "purple4", higher = "tomato"),
  bins = 30,
  binWidth = NULL
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object containing spatial transcriptomics
data.}

\item{metric}{A character string specifying the name of the metric
(column in `colData(spe)`) to plot.}

\item{fillColor}{A character string specifying the fill color of the
histogram bars. (Default is `"#69b3a2"`)}

\item{useFences}{A character string specifying the name of the column in
`colData(spe)` that contains the fence thresholds (typically from an outlier
filter). If `NULL`, no fences will be plotted. (Default is `NULL`)}

\item{fencesColors}{A named character vector specifying the colors to use
for the lower and higher fences. The names should be `"lower"` and `"higher"`
. (Default is `c("lower"="purple4", "higher"="tomato")`)}

\item{bins}{An integer specifying the number of bins to use in the histogram.
(Default is `30`)}

\item{binWidth}{A numeric value specifying the width of the bins. If `NULL`,
the bin width will be automatically determined based on the `bins` parameter.
(Default is `NULL`)}
}
\value{
A `ggplot` object representing the histogram of the specified metric.
}
\description{
Plot a Histogram for a Given Metric in a SpatialExperiment
Object

This function generates a histogram for a specified metric in a
`SpatialExperiment` object.
}
\examples{
example(readCosmxSPE)
g <- plotMetricHist(spe, metric="Mean.DAPI")
print(g)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{computeCenterFromPolygons}
\alias{computeCenterFromPolygons}
\title{computeCenterFromPolygons}
\usage{
computeCenterFromPolygons(polygons, coldata)
}
\arguments{
\item{polygons}{An `sf` object containing polygon data.}

\item{coldata}{A `DataFrame` containing the `colData` to which center
coordinates information will be added.}
}
\value{
A `DataFrame` with the added center information.
}
\description{
This function computes the center coordinates on x and y axis
from polygon data and adds it to the `colData`. It is necessary only for
Merfish.
}
\examples{
example(readPolygonsMerfish)
coldata <- computeCenterFromPolygons(polygons, colData(spe))
colData(spe) <- coldata
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialQCPlots.R
\name{qcFlagPlots}
\alias{qcFlagPlots}
\title{qcFlagPlots}
\usage{
qcFlagPlots(
  spe,
  fov = unique(spe$fov),
  theme = c("light", "dark"),
  custom = FALSE
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object containing spatial transcriptomics
data.}

\item{fov}{An integer or numeric vector specifying the FOVs to be plotted
Must match values in the `fov` column of `colData(spe)`.}

\item{theme}{A character string among "light" or "dark".}

\item{custom}{A boolean value. If TRUE, custom polygons derived metrics will
be used.}
}
\value{
A panel with multiple plots showing flagged cells for different
variables.
}
\description{
Plots the flagged cells identified with first filter, based on control count
on total count ratio, area in um and DAPI signal.

This function generates a plot that shows selected (FOVs)
within a `SpatialExperiment` object, with cells flagged in different colors
over a light or dark layout chosen by the user.
}
\examples{

example(readAndAddPolygonsToSPE)
spe <- spatialPerCellQC(spe)
spe <- computeThresholdFlags(spe)
p <- qcFlagPlots(spe, fov=16, theme="dark")
print(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.setActiveGeometry}
\alias{.setActiveGeometry}
\title{.setActiveGeometry}
\usage{
.setActiveGeometry(sf, name)
}
\arguments{
\item{sf}{an sf object}

\item{name}{character for the geometry to activate}
}
\value{
an sf object
}
\description{
.setActiveGeometry
}
\examples{
example(readPolygonsCosmx)
.setActiveGeometry(polygons, "local")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{readPolygonsMerfish}
\alias{readPolygonsMerfish}
\title{readPolygonsMerfish}
\usage{
readPolygonsMerfish(
  polygons,
  type = c("parquet", "HDF5"),
  keepMultiPol = TRUE,
  hdf5pattern = "hdf5",
  zLev = 3L,
  zcolumn = "ZIndex",
  geometry = "Geometry",
  verbose = FALSE
)
}
\arguments{
\item{polygons}{A character string specifying the folder containing the
polygon data files in case of HDF5, or a path to a parquet file (see `type`).}

\item{type}{A character string specifying the file type("HDF5" or "parquet").
Default is parquet.}

\item{keepMultiPol}{A logical value indicating whether to keep multipolygons.}

\item{hdf5pattern}{A character string specifying the pattern to match HDF5
files.}

\item{zLev}{An integer specifying the Z level to filter the data. Default is
`3L`.}

\item{zcolumn}{A character string specifying the column name for the Z index.}

\item{geometry}{A character string specifying the geometry column name.}

\item{verbose}{A logical value indicating whether to print additional
information.}
}
\value{
An `sf` object containing the MERFISH polygon data.
}
\description{
This function reads polygon data specific to MERFISH technology.
}
\examples{
example(readMerfishSPE)
polygons <- readPolygonsMerfish(metadata(spe)$polygons, type="parquet")
polygons
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialQCPlots.R
\name{plotCellsFovs}
\alias{plotCellsFovs}
\title{plotCellsFovs}
\usage{
plotCellsFovs(
  spe,
  sampleId = unique(spe$sample_id),
  pointCol = "firebrick",
  numbersCol = "black",
  alphaNumbers = 0.8,
  fovDim = metadata(spe)$fov_dim,
  size = 0.05,
  alpha = 0.8,
  scaleBar = TRUE,
  micronConvFact = 0.12,
  pointSize = NULL,
  pointAlpha = NULL,
  numberSize = 1,
  numbersAlpha = NULL
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with `fov` in `colData`.}

\item{sampleId}{Character string identifying which sample to plot.
Default: `unique(spe$sample_id)`.}

\item{pointCol}{Color for the cell centroids. Default: `"firebrick"`.}

\item{numbersCol}{Color for the FoV labels. Default: `"black"`.}

\item{alphaNumbers}{Deprecated alias for `numbersAlpha`.}

\item{fovDim}{numeric with two named dimensions xdim, ydim. (Default is
metadata(spe)$fov_dim)}

\item{size}{Deprecated alias for `pointSize`.}

\item{alpha}{Deprecated alias for `pointAlpha`.}

\item{scaleBar}{A logical value indicating whether to add a scale bar to the
plot. (Default is `TRUE`)}

\item{micronConvFact}{Numeric conversion factor from pixels to microns.
Default is `0.12`.}

\item{pointSize}{Numeric point size for the cell centroids. If supplied,
takes precedence over `size`.}

\item{pointAlpha}{Numeric transparency for the cell centroids. If supplied,
takes precedence over `alpha`.}

\item{numberSize}{Numeric size for the FoV labels. Default: `1`.}

\item{numbersAlpha}{Numeric transparency for FoV labels. If supplied, takes
precedence over `alphaNumbers`.}
}
\value{
A `ggplot` object showing cell centroids and FoV boundaries.
}
\description{
Plot cell centroids in FoVs, creating a map of the whole experiment, where
cells are plotted as points and FoV boundaries and numbers are overlaid.
}
\details{
The function expects spe (a SpatialExperiment) to include field-of-view
metadata in metadata(spe):
- `metadata(spe)$fov_positions`: a matrix or data.frame
(or list with named elements) containing at minimum `x_global_px`, `y_global_px`,
and `fov`. Values `x_global_px`/`y_global_px` are in pixels and represent
the origin (bottom-left) of each FoV.
- `metadata(spe)$fov_dim` (or the `fovDim` argument): a named numeric with
`xdim` and `ydim` giving FoV width/height in pixels.

For each FoV the rectangle is drawn as:
- `xmin` = `x_global_px`,
- `xmax` = `x_global_px` + `xdim`,
- `ymin` = `y_global_px`,
- `ymax` = `y_global_px` + `ydim`.

`sampleId`: if `NULL` no plot title is added; if a vector of length > 1 is
provided the first element is used. Prefer passing a single sample identifier
string.
`micronConvFact` is the conversion factor from pixels to micrometers (µm/pixel).
It is forwarded to `.plotScaleBar()` and controls how the scale bar is labeled
(useful for technologies where coordinates are in pixels but the scale should
display µm).
The function uses the coordinates returned by `spatialCoords(spe)` and the
names from `spatialCoordsNames(spe)`. It performs basic input checks but
assumes the described metadata structures are present; if they are missing
or malformed the function will raise an informative error.
}
\examples{
example(readCosmxSPE)
g <- plotCellsFovs(spe)
print(g)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeQCScore}
\alias{computeQCScore}
\title{computeQCScore}
\usage{
computeQCScore(spe, bestLambda = NULL, verbose = FALSE)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial transcriptomics data.}

\item{bestLambda}{the best lambda typically computed using `computeLambda`.}

\item{verbose}{logical for having a verbose output. Default is FALSE.}
}
\value{
The `SpatialExperiment` object with added QC score in `colData`.
}
\description{
Compute QC score and automatically define weights for QC score
through glm training. This function computes QC score with a formula
that is defined based on the metrics specified in metric_list and on the
number of available outliers for each metric.
}
\details{
For CosMx datasets, also CosMx Protein, the QC Score formula is
defined as follows:

QC score ~ count density - aspect ratio - control-total ratio

count density is total counts-to-area ratio, aspect ratio represents
FOV border effect typical of CosMx datasets and control-total ratio is
the aspecific signal. For each couple of variables interaction terms are
computed.

For Xenium and Merscope datasets, QC score cannot depend on aspect ratio
as no FOV border effect was captured through this metric.

Inclusion of metrics in the formula depends also on the number of available
outliers. If the number of outliers for each metric is < 0.1% out of the
entire dataset, the metric will be excluded from the QC score formula.

To automatically define the formula coefficient weights, model training
is performed through ridge regression.
}
\examples{
example(spatialPerCellQC)
set.seed(1998)
spe <- computeQCScore(spe)
summary(spe$QC_score)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeThresholdFlags}
\alias{computeThresholdFlags}
\title{computeThresholdFlags}
\usage{
computeThresholdFlags(spe, totalThreshold = 0, ctrlTotRatioThreshold = 0.1)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial transcriptomics data.}

\item{totalThreshold}{A numeric value for the threshold of total counts to
identify cells with low counts. Default is `0`.}

\item{ctrlTotRatioThreshold}{A numeric value for the threshold of
control-to-total ratio to flag cells over a certain threshold. Default is
`0.1`.}
}
\value{
The `SpatialExperiment` object with added filter flags in `colData`.
}
\description{
Compute Flagged cells using fixed thresholds for SpatialExperiment.

This function calculates flagged cells only for total counts and control on
total probe counts ratio using fixed thresholds for a `SpatialExperiment`
object.
}
\details{
The function flags cells basing on zero counts and control-to-total
ratio to identify junk cells.
It also combines these flags into a single filter flag.
}
\examples{
example(readCosmxSPE)
spe <- spatialPerCellQC(spe)
spe <- computeThresholdFlags(spe)
table(spe$threshold_flags)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readXenium.R
\name{computeMissingMetricsXenium}
\alias{computeMissingMetricsXenium}
\title{computeMissingMetricsXenium}
\usage{
computeMissingMetricsXenium(
  polFile,
  colData,
  keepPolygons = FALSE,
  polygonsCol = "polygons"
)
}
\arguments{
\item{polFile}{A character string specifying the file path to the polygon
data.}

\item{colData}{A `DataFrame` containing the `colData` for the Xenium dataset.}

\item{keepPolygons}{A logical value indicating whether to keep the polygon
data in the resulting `colData`. Default is `FALSE`.}

\item{polygonsCol}{character indicating the name of the polygons column to
add into the colData (default is `polygons`).}
}
\value{
A `DataFrame` containing the updated `colData` with computed metrics.
If `keepPolygons` is `TRUE`, the polygon data is also included.
}
\description{
Compute Missing Metrics for Xenium Data

This function computes missing metrics, such as the aspect ratio, from
polygon data in a Xenium dataset and optionally appends the polygon data to
the resulting `colData`.
}
\details{
The function reads the polygon data from the specified file,
computes the aspect ratio for each polygon, and merges these metrics with
the provided `colData`. Optionally, the polygon data can be kept in the
returned `colData`.
}
\examples{
example(readXeniumSPE)
colData(spe) <- computeMissingMetricsXenium(metadata(spe)$polygons,
    colData(spe), keepPolygons=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialQCPlots.R
\name{plotPolygons}
\alias{plotPolygons}
\title{plotPolygons}
\usage{
plotPolygons(
  spe,
  colourBy = "darkgrey",
  colourLog = FALSE,
  polyColumn = "polygons.global",
  sampleId = unique(spe$sample_id),
  bgColor = "white",
  fillAlpha = 1,
  palette = NULL,
  borderCol = NA,
  borderAlpha = 1,
  borderLineWidth = 0.1,
  drawBorders = TRUE,
  scaleBar = TRUE,
  micronConvFact = 0.12
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with polygon data as an `sf` object.}

\item{colourBy}{A column in `colData(spe)` for coloring the polygons or a
string color in colors(). (Default is "darkgrey")}

\item{colourLog}{Logical to log-transform the data to enhance visualization
(Default is FALSE).}

\item{polyColumn}{character for the name of the column where the polygons sf
are stored (default is "polygons.global")}

\item{sampleId}{Sample ID for plot title. Default is the unique sample ID.}

\item{bgColor}{character indicating color for the background
(default is "white")}

\item{fillAlpha}{Transparency level for polygon fill. Default is `1`.}

\item{palette}{Colors to use if `colourBy` is a factor. Default is `NULL`.}

\item{borderCol}{Color of polygon borders. Default is `"black"`.}

\item{borderAlpha}{Transparency level for borders. Default is `1`.}

\item{borderLineWidth}{Width of polygon borders. Default is `0.1`.}

\item{drawBorders}{Logical; whether to draw borders. Default is `TRUE`.}

\item{scaleBar}{A logical value indicating whether to add a scale bar to the
plot. (Default is `TRUE`)}

\item{micronConvFact}{Numeric conversion factor from pixels to microns.
DEFAULT is `0.12`.}
}
\value{
A `ggplot` object representing the polygon plot of the spatial data.
}
\description{
Plot polygons from a `SpatialExperiment` object using ggplot2.
}
\details{
Renders polygon geometries stored in `colData` of a `SpatialExperiment`
object using `ggplot2::geom_sf()` and provides options for fill, borders
and an optional scale bar.

Input expectations:
- `spe` must be a `SpatialExperiment` and `polyColumn` must name a column
  in `colData(spe)` that contains polygon geometries (`sfc`/`sf` or a list
  of geometry objects) suitable for `geom_sf()`.
- `colourBy` supports two modes: (a) the name of a `colData` column to
  map fill to data values, or (b) a literal colour name (e.g. `"darkgrey"`)
  used as a constant fill.

Colour / palette behaviour:
- If `colourBy` refers to a `colData` column and `colourLog = TRUE`, the
  function computes `log1p()` of that column in the local plotting data
  frame (it does not mutate `spe`).
- `palette` may be a vector of colour values (used for discrete scales).
  Numeric `colourBy` values use a continuous Viridis-like scale by default.

Scale bar and units:
- If `scaleBar = TRUE` the function calls `.plotScaleBar()` to add a length
  scale. `micronConvFact` is the conversion factor (µm per pixel) used by
  `.plotScaleBar()` to label the bar in micrometres when appropriate.

Robustness notes:
- If `colourBy` is neither a `colData` column nor a valid colour name the
  function issues a warning and falls back to the default fill.
}
\examples{
example(readAndAddPolygonsToSPE)
plotPolygons(spe, colourBy="Mean.DAPI")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{spatialPerCellQC}
\alias{spatialPerCellQC}
\title{spatialPerCellQC}
\usage{
spatialPerCellQC(
  spe,
  micronConvFact = 0.12,
  rmZeros = TRUE,
  negProbList = c("NegPrb", "Negative", "SystemControl", "Ms IgG1", "Rb IgG", "BLANK_",
    "NegControlProbe", "NegControlCodeword", "UnassignedCodeword", "Blank"),
  use_altexps = NULL
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object containing spatial omics data.}

\item{micronConvFact}{Numeric factor to convert pixels to microns. Default
is 0.12.}

\item{rmZeros}{logical for removing zero total counts (protein intensities) 
cells (default is TRUE).}

\item{negProbList}{Character vector of patterns to identify negative probes.
Defaults include:
  Nanostring CosMx: `"NegPrb"`, `"Negative"`, `"SystemControl"`
  Xenium: `"NegControlProbe"`, `"NegControlCodeword"`, `"UnassignedCodeword"`
  MERFISH: `"Blank"`}

\item{use_altexps}{logical for `use_altexps` in `scuttle` package.
If TRUE uses the altexps for computing some metrics on it.
Useful for interoperability with `SpatialExperimentIO`.
(See \link[scuttle]{addPerCellQC} for additional details).}
}
\value{
A `SpatialExperiment` object with added QC metrics in `colData`.
}
\description{
Computes quality‐control metrics for each cell and adds them to `colData`.
}
\details{
The function computes and appends per‑cell QC metrics to `colData(spe)` and
may subset the returned SpatialExperiment.

Key behaviours and expectations:
- Feature detection: negative‑probe patterns supplied in `negProbList` are
passed to `scater::addPerCellQC()` (via `use_altexps` when requested). 
`addPerCellQC()` must be able to find matching control feature names 
(`rownames` of spe) and will create `subsets_*_sum` and `subsets_*_detected` 
columns used below for each of control feature types.

The function appends `sum`, `detected` and `total` to the SPE colData by
calling internally `addPerCellQC()`; these are used to compute `control_sum`, 
`control_detected`, `target_sum` and `target_detected.`

- Control metrics: `control_sum` and `control_detected` are computed by
summing matching `subsets_*` columns; `target_*` metrics are computed as
sum(detected) - control_sum(_detected).

- Ratios and logs: `ctrl_total_ratio` (`control_sum`/`sum`) and its stabilized
log2-transformed `log2Ctrl_total_ratio` are added.

- Coordinate and area handling:
- For CosMx technologies (Nanostring_CosMx and Nanostring_CosMx_Protein)
spatial coordinates are converted from pixels to microns using
`micronConvFact` and appended to `colData` (column names have px -> um).

- For CosMx, `Area_um` is derived from an existing Area column scaled by
micronConvFact^2 and `.computeBorderDistanceCosMx()` is invoked to
compute `dist_border`.

- For Nanostring_CosMx_Protein, a legacy `Area.um2` column (if present)
is renamed to `Area_um` to standardize naming.

- For Xenium (10X_Xenium), if `Area_um` is missing the function will
attempt to use `cell_area` as a fallback and issue a warning.

- Aspect ratio: if `AspectRatio` exists it is logged (`log2AspectRatio`); if
missing a warning is emitted.

- Signal density: `SignalDensity` is computed as `sum` / `Area_um` for most
technologies; for Nanostring_CosMx_Protein it coincides with `total`. A
log2-transformed `log2SignalDensity` is also added.

- Zero‑count (protein intensity) removal: when `rmZeros = TRUE` cells with 
`sum == 0` are removed from the returned SpatialExperiment (message printed).

- Side effects: the function modifies `colData(spe)` (adds multiple new
columns). It may add spatial coordinates into `colData` if missing, and may
subset the SPE to remove zero‑count cells. It issues warnings when expected
inputs (e.g. area, aspect ratio, polygon‑derived fields) are missing or
when fallbacks are used.

Use this information to ensure the input SpatialExperiment contains the
necessary assays and fields (feature names, `sum`/`detected`/`total`,
`Area`/`cell_area`, log2AspectRatio, dist_border when available) so 
metrics are computed and assigned correctly.
}
\examples{
example(readCosmxSPE)
spe <- spatialPerCellQC(spe)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeSpatialOutlier}
\alias{computeSpatialOutlier}
\title{computeSpatialOutlier}
\usage{
computeSpatialOutlier(
  spe,
  computeBy = NULL,
  method = c("mc", "scuttle", "both"),
  mcDoScale = FALSE,
  scuttleType = c("both", "lower", "higher")
)
}
\arguments{
\item{spe}{a SpatialExperiment object with quantitative cell metrics in the `colData`.}

\item{computeBy}{character indicating a `colData` column name to compute the outliers on.}

\item{method}{one of `mc`, `scuttle`, `both`.
Use `mc` for medcouple, `scuttle` for MAD as computed in `scuttle`, `both` 
for computing both of them.}

\item{mcDoScale}{logical indicating if the metric values should be scaled 
to compute medcouple (default is FALSE, as suggested by the original medcouple 
authors). See \link[robustbase]{mc} for further readings.}

\item{scuttleType}{One of `"both"`, `"lower"`, `"higher"` for scuttle method.}
}
\value{
a SpatialExperiment object with additional column(s) (named as
the column name indicated in `computeBy` followed by the outlier_sc/mc).
Resulting columns are `outlier.filter` logical class objects, allowing to store 
the thresholds as attributes of the columns.
Use attr(,"thresholds") to retrieve them.
}
\description{
Computes outliers for quantitative cell metadata based on a statistical test:
medcouple (`mc` method argument) or the Median Absolute Deviation (MAD, `scuttle` 
method argument), when the metric distribution is respectively asymmetric or symmetric.
}
\details{
The medcouple method is a measure for the skeweness of univariate
distribution as described in Hubert M. et al. (2008).
In particular, the computed medcouple value must be in a range between -0.6
and 0.6 to compute adjusted boxplots and perform the outlier detection.
For MAD method we just wrap the isOutlier function in the scuttle package. 
Please see McCarthy DJ et al (2017) for further details.
}
\examples{
example(spatialPerCellQC)
spe <- computeSpatialOutlier(spe, computeBy="log2SignalDensity", method="both")
table(spe$log2SignalDensity_outlier_mc)
table(spe$log2SignalDensity_outlier_sc)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotUtils.R
\name{.centroid_image_theme}
\alias{.centroid_image_theme}
\title{.centroid_image_theme}
\usage{
.centroid_image_theme(backBorder = NA)
}
\arguments{
\item{backBorder}{color for the borders of the background (default=NA)}
}
\value{
a ggplot2 theme object
}
\description{
internal function to setup the theme for the centroid plot background
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.getActiveGeometryName}
\alias{.getActiveGeometryName}
\title{.getActiveGeometryName}
\usage{
.getActiveGeometryName(sf)
}
\arguments{
\item{sf}{an sf object}
}
\value{
character with the name of the active geometry
}
\description{
.getActiveGeometryName
}
\examples{
example(readPolygonsCosmx)
.getActiveGeometryName(polygons)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{SpaceTrooper-deprecated}
\alias{SpaceTrooper-deprecated}
\alias{computeQCScore}
\alias{computeQCScoreFlags}
\alias{computeOutliersQCScore}
\alias{applyQCScoreModel}
\title{Deprecated functions in SpaceTrooper}
\usage{
computeQCScore(spe, bestLambda = NULL, modelFormula = NULL, verbose = FALSE)

computeQCScoreFlags(spe, qsThreshold = 0.5, useQSQuantiles = FALSE)

computeOutliersQCScore(
  spe,
  metricList = c("log2SignalDensity", "Area_um", "log2AspectRatio",
    "log2Ctrl_total_ratio")
)

applyQCScoreModel(spe, qcModel, scoreName = "QC_score")
}
\arguments{
\item{spe}{A `SpatialExperiment` object with QC metrics already computed.}

\item{bestLambda}{Passed to \code{\link{computeQScore}}.}

\item{modelFormula}{Passed to \code{\link{computeQScore}}.}

\item{verbose}{Passed to \code{\link{computeQScore}}.}

\item{qsThreshold}{Passed to \code{\link{computeQScoreFlags}}.}

\item{useQSQuantiles}{Passed to \code{\link{computeQScoreFlags}}.}

\item{metricList}{Passed to \code{\link{computeOutliersQScore}}.}

\item{qcModel}{A historical QC score model object, usually stored in
`metadata(spe)$QCScore_model`.}

\item{scoreName}{Name of the legacy output column in `colData`.}
}
\value{
A `SpatialExperiment` object; see \code{\link{computeQScore}}.

A `SpatialExperiment` object; see \code{\link{computeQScoreFlags}}.

A `SpatialExperiment` object; see
  \code{\link{computeOutliersQScore}}.

A `SpatialExperiment` object with the applied score in `colData`
  and the model in `metadata(spe)$QCScore_model_applied`.
}
\description{
These functions are retained for compatibility with published SpaceTrooper
APIs. They issue a deprecation warning and direct users to the canonical
Quality Score functions.
}
\details{
\itemize{
  \item `computeQCScore()` replaces canonical `QScore` and
    `metadata(spe)$QScore_model` with historical `QC_score` and
    `metadata(spe)$QCScore_model`.
  \item `computeQCScoreFlags()` consumes `QC_score` and creates
    `low_qcscore` and, when applicable, `low_threshold_qcscore`.
  \item `computeOutliersQCScore()` is the historical name for
    `computeOutliersQScore()`.
  \item `applyQCScoreModel()` retains the `qcModel` argument, the default
    `QC_score` output, and `metadata(spe)$QCScore_model_applied`.
}

Use `computeQScore()`, `computeQScoreFlags()`,
`computeOutliersQScore()`, and `applyQScoreModel()` for canonical names.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeQScoreFlags}
\alias{computeQScoreFlags}
\title{computeQScoreFlags}
\usage{
computeQScoreFlags(spe, qsThreshold = 0.5, useQSQuantiles = FALSE)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial omics data.}

\item{qsThreshold}{Numeric threshold or quantile for QS. Default is `0.5`.}

\item{useQSQuantiles}{Logical; if `TRUE`, treat `qsThreshold` as a percentile.}
}
\value{
The `SpatialExperiment` object with added low-quality flagged cells as
TRUE/FALSE in `colData`.
}
\description{
Computes flagged cells based on a manually chosen threshold on Quality Score (QS).

This function computes flagged cells based on a manually chosen threshold on
QS stored in `SpatialExperiment` object.
}
\examples{
example(computeQScore)
spe <- computeQScoreFlags(spe)
table(spe$low_QScore)
# if fixed filters are defined we have an additional column
spe <- computeThresholdFlags(spe)
spe <- computeQScoreFlags(spe)
table(spe$low_threshold_QScore)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeOutliersQCScore}
\alias{computeOutliersQCScore}
\title{computeOutliersQCScore}
\usage{
computeOutliersQCScore(
  spe,
  metricList = c("log2CountArea", "Area_um", "log2AspectRatio", "log2Ctrl_total_ratio")
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial omics data.}

\item{metricList}{A character vector specifying the metrics to include in
the QC score formula. Default is `c("log2CountArea", "Area_um",
"log2AspectRatio", "log2Ctrl_total_ratio")`.}
}
\value{
The `SpatialExperiment` object with added outlier variables in
`colData` and the temporary QCScore metric variables that in the
`metadata`.
}
\description{
Compute outlier cells for each metric that can be used in QC score formula
for SpatialExperiment.

This function calculates outlier cells for each variable specified
in `metricList` for a `SpatialExperiment`. Log2CountArea must be present in
the `colData` of the `SpatialExperiment` object as a minimum requirement.
The user can choose which metrics to include among the following: Area_um,
log2Ctrl_total_ratio, log2AspectRatio. For Xenium and Merfish datasets,
log2AspectRatio is automatically removed from the formula.
}
\details{
The function computes outliers for each specified metric after
automatically choosing the appropriate method according to the skewness of
the distribution.
Internally the function:
\enumerate{
   \item Calls \code{.checkSkw()} to choose the proper outlier detection
    method according to the variable skewness,
   \item Calls \code{computeSpatialOutlier()} on each included metric to get
   fences,
   \item Labels cells as “LOW”/“HIGH” outliers or “NO”
}
}
\examples{
example(readCosmxSPE)
spe <- spatialPerCellQC(spe)
spe <- computeOutliersQCScore(spe)
table(spe$log2CountArea_outlier_train)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeQCScoreFlags}
\alias{computeQCScoreFlags}
\title{computeQCScoreFlags}
\usage{
computeQCScoreFlags(spe, qsThreshold = 0.5, useQSQuantiles = FALSE)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial transcriptomics data.}

\item{qsThreshold}{Numeric threshold or quantile for quality score. Default
`0.5`.}

\item{useQSQuantiles}{Logical; if `TRUE`, treat `qsThreshold` as a
percentile.}
}
\value{
The `SpatialExperiment` object with added filter flags in `colData`.
}
\description{
Compute flagged cells based on a manually chosen threshold on quality score

This function Compute flagged cells based on a manually chosen threshold on
quality score stored in `SpatialExperiment` object.
}
\examples{
example(computeQCScore)
spe <- computeQCScoreFlags(spe)
table(spe$low_qcscore)
# if fixed filters are defined we have an additional column
spe <- computeThresholdFlags(spe)
spe <- computeQCScoreFlags(spe)
table(spe$low_threshold_qcscore)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readMerfish.R
\name{computeMissingMetricsMerfish}
\alias{computeMissingMetricsMerfish}
\title{computeMissingMetricsMerfish}
\usage{
computeMissingMetricsMerfish(
  polFile,
  coldata,
  boundariesType = c("parquet", "HDF5"),
  keepPolygons = FALSE,
  polygonsCol = "polygons",
  useVolume = TRUE
)
}
\arguments{
\item{polFile}{character for the path to the polygon file. Tipically a
parquet file or a folder of HDF5 files in the `metadata(spe)$polygons`.}

\item{coldata}{`DataFrame` or `data.frame`
Cell metadata with at least a `cell_id` column.}

\item{boundariesType}{`character(1)`
One of `"HDF5"` or `"parquet"`—passed on to `readPolygonsMerfish()`.}

\item{keepPolygons}{`logical(1)`
If `TRUE`, cbinds the raw polygon `sf` columns onto `coldata`.}

\item{polygonsCol}{character indicating the name of the polygons column to
add into the colData (default is `polygons`).}

\item{useVolume}{`logical(1)` it assigns the area from the "volume" column.
If the column is not present it computes the area from the polygons.
Default: `TRUE`.}
}
\value{
A `DataFrame` (or `data.frame`) with:
  - all columns of `coldata`
  - `Area_um`: area/volume of each cell’s polygon
  - `AspectRatio`: width/height aspect ratio
  - (optionally) the polygon geometries
}
\description{
`computeMissingMetricsMerfish()` takes cell metadata and boundary
polygons, calculates per‐cell area and aspect‐ratio, and optionally
appends the raw polygon geometries.
}
\details{
`computeMissingMetricsMerfish()` reads polygon geometries via
`readPolygonsMerfish(polFile, type=boundariesType)` where `polFile`
is a Parquet file (parquet) or a folder of HDF5 files (HDF5). It expects
`coldata` to contain at least `cell_id`.

Behavior:
- If `useVolume=TRUE` and `coldata` contains a `volume` column, the
  returned `Area_um` is set to `volume`. Otherwise `Area_um` is computed
  from the polygons using `computeAreaFromPolygons()`.
- `AspectRatio` is computed from polygons using `computeAspectRatioFromPolygons()`.
- If `keepPolygons=TRUE`, geometries are attached to the returned
  `DataFrame` under the name given by `polygonsCol`.

Important: polygons must align (in order or by matching logic) with the
rows of `coldata`; otherwise area/aspect values may be assigned incorrectly.
}
\examples{
example(readMerfishSPE)
cd <- computeMissingMetricsMerfish(metadata(spe)$polygons, colData(spe),
    boundariesType="parquet")
colData(spe) <- cd
cd
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readXenium.R
\name{updateXeniumSPE}
\alias{updateXeniumSPE}
\title{updateXeniumSPE}
\usage{
updateXeniumSPE(
  spe,
  dirName,
  sampleName = "sample01",
  boundariesType = c("parquet", "csv"),
  computeMissingMetrics = TRUE,
  keepPolygons = FALSE,
  polygonsFPattern = "cell_boundaries",
  polygonsCol = "polygons",
  txPattern = "transcripts",
  addFOVs = FALSE
)
}
\arguments{
\item{spe}{SpatialExperiment object to update.}

\item{dirName}{Directory containing Xenium outputs.}

\item{sampleName}{Sample identifier to store (default \code{"sample01"}).}

\item{boundariesType}{One of \code{c("parquet","csv")} indicating the source
format of cell boundaries.}

\item{computeMissingMetrics}{Logical; if \code{TRUE}, compute metrics that are
not already present from transcripts/polygons.}

\item{keepPolygons}{Logical; if \code{TRUE}, keep polygons in the resulting
object (e.g., in metadata or colData, depending on implementation).}

\item{polygonsFPattern}{Character pattern used to locate polygon files when
boundaries are provided as CSVs (default \code{"polygons.csv"}).}

\item{polygonsCol}{Name of the geometry/column storing polygons when reading
from parquet (default \code{"polygons"}).}

\item{txPattern}{Pattern (file/glob) for transcript-level files
(default \code{"transcripts"}).}

\item{addFOVs}{Logical; if \code{TRUE}, derive and attach FOV identifiers
from transcript resources.}
}
\value{
Updated SpatialExperiment object.
}
\description{
Update a SpatialExperiment created from 10x Genomics Xenium outputs by
wiring polygons/boundaries, computing optional QC metrics, and standardizing
metadata and column names. This is a thin wrapper that delegates to the
internal helper `.setupXeniumSPE()`.
}
\details{
This function performs input checks and then calls `.setupXeniumSPE()`,
which does the heavy lifting (I/O, renaming, metadata updates, metrics).
}
\examples{
xepath <- system.file("extdata", "Xenium_small", package="SpaceTrooper")
(spe <- SpatialExperimentIO::readXeniumSXE(dirName=xepath))
spe <- updateXeniumSPE(spe, dirName=xepath, boundariesType="parquet",
    computeMissingMetrics=TRUE, keepPolygons=TRUE)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialQCPlots.R
\name{plotZoomFovsMap}
\alias{plotZoomFovsMap}
\title{plotZoomFovsMap}
\usage{
plotZoomFovsMap(
  spe,
  fovs = NULL,
  title = NULL,
  mapPointCol = "darkmagenta",
  mapNumbersCol = "black",
  mapAlphaNumbers = 0.8,
  csize = 0.05,
  calpha = 0.8,
  scaleBars = NULL,
  scaleBarMap = TRUE,
  scaleBarPol = TRUE,
  ...,
  mapPointSize = NULL,
  mapPointAlpha = NULL,
  fovNumbersCol = NULL,
  fovNumberSize = 1,
  fovNumbersAlpha = NULL
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object containing spatial transcriptomics
data.}

\item{fovs}{A character vector specifying the FOVs to be zoomed in and
plotted. Must match values in the `fov` column of `colData(spe)`.}

\item{title}{An optional character string specifying the title of the final
plot. If `NULL`, no title is added. Default is `NULL`.}

\item{mapPointCol}{A character string specifying the color of the points
in the map. Default is `"darkmagenta"`.}

\item{mapNumbersCol}{Deprecated alias for `fovNumbersCol`.}

\item{mapAlphaNumbers}{Deprecated alias for `fovNumbersAlpha`.}

\item{csize}{Deprecated alias for `mapPointSize`.}

\item{calpha}{Deprecated alias for `mapPointAlpha`.}

\item{scaleBars}{Logical or NULL. Default is `NULL`.
Master switch controlling the presence of scale bars in both panels.
If \code{TRUE}, scale bars are shown in both the map and polygon panels.
If \code{FALSE}, scale bars are hidden in both panels.
If \code{NULL}, individual settings defined by \code{scaleBarMap} and
\code{scaleBarPol} are used.}

\item{scaleBarMap, scaleBarPol}{Logical. Default is \code{TRUE}.
Control the presence of the scale bar in the map (cells/FOV overview)
and polygon (segmentation) panels, respectively.
These parameters are only used when \code{scaleBars} is \code{NULL};
otherwise they are overridden by \code{scaleBars}.}

\item{...}{Additional arguments passed to `plotPolygons`.}

\item{mapPointSize}{Numeric size for points in the map. If supplied, takes
precedence over `csize`.}

\item{mapPointAlpha}{Numeric transparency for points in the map. If supplied,
takes precedence over `calpha`.}

\item{fovNumbersCol}{Color for FoV labels. If supplied, takes precedence over
`mapNumbersCol`.}

\item{fovNumberSize}{Numeric size for the FoV labels. Default: `1`.}

\item{fovNumbersAlpha}{Transparency for FoV labels. If supplied, takes
precedence over `mapAlphaNumbers`.}
}
\value{
A combined plot showing a map of all FOVs with zoomed-in views of
the specified FOVs and their associated polygons.
}
\description{
Plot Zoomed-in FOVs with Map and Polygons

This function generates a plot that shows a map of all fields of view (FOVs)
within a `SpatialExperiment` object, alongside a zoomed-in view of the
specified FOVs with an overlay of polygons and optional coloring.
}
\details{
The function first filters the `SpatialExperiment` object to the
specified FOVs, generates a plot of the cells for the entire map, then
creates a detailed polygon plot of the selected FOVs, and finally combines
these into a single side-by-side visualization. If `title` is not `NULL`, it
adds a title to the combined plot.
}
\examples{
example(readAndAddPolygonsToSPE)
plotZoomFovsMap(spe, fovs=16, title="FOV 16")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readMerfish.R
\name{readMerfishSPE}
\alias{readMerfishSPE}
\title{readMerfishSPE}
\usage{
readMerfishSPE(
  dirName,
  sampleName = "sample01",
  computeMissingMetrics = TRUE,
  keepPolygons = FALSE,
  boundariesType = c("HDF5", "parquet"),
  countmatFPattern = "cell_by_gene.csv",
  metadataFPattern = "cell_metadata.csv",
  polygonsFPattern = "cell_boundaries.parquet",
  coordNames = c("center_x", "center_y"),
  polygonsCol = "polygons"
)
}
\arguments{
\item{dirName}{`character(1)`
Path to a folder containing MERFISH output files.}

\item{sampleName}{`character(1)`
Identifier to assign to the `sample_id` field in the returned object.
Default: `"sample01"`.}

\item{computeMissingMetrics}{`logical(1)`
If `TRUE`, compute area and aspect‐ratio metrics from the cell boundary
polygons. Default: `TRUE`.}

\item{keepPolygons}{`logical(1)`
If `TRUE`, attach raw polygon geometries as extra columns in `colData`.
Default: `FALSE`.}

\item{boundariesType}{`character(1)`
One of `"HDF5"` or `"parquet"`. If `"HDF5"`, uses a folder of HDF5
polygon files; if `"parquet"`, reads a single Parquet file of boundaries.}

\item{countmatFPattern}{`character(1)`
Regex passed to `list.files()` to find the count matrix CSV. Default:
`"cell_by_gene.csv"`.}

\item{metadataFPattern}{`character(1)`
Pattern to find the cell metadata CSV. Default: `"cell_metadata.csv"`.}

\item{polygonsFPattern}{`character(1)`
Pattern to find the cell boundaries file. Default:
`"cell_boundaries.parquet"`.}

\item{coordNames}{`character(2)`
Names of the columns in `colData` that store X/Y spatial coordinates.
Default: `c("center_x", "center_y")`.}

\item{polygonsCol}{character indicating the name of the polygons column to
add into the colData (default is `polygons`).}
}
\value{
A `SpatialExperiment` object with:
  - `assays$counts`: gene × cell count matrix
  - `colData`: per‐cell metadata (including computed metrics)
  - spatial coordinates named by `coordNames`
  - `metadata$polygons`: path to the boundaries file
  - `metadata$technology`: `"Vizgen_MERFISH"`.
}
\description{
`readMerfishSPE()` imports MERFISH/Merscore outputs (counts, metadata,
and optionally cell boundary polygons) from a directory and builds a
SpatialExperiment object.
}
\examples{
path <- system.file("extdata", "Merfish_Tiny",
                    package = "SpaceTrooper")
spe <- readMerfishSPE(
  dirName = path,
  sampleName = "Patient2",
  keepPolygons = TRUE,
  boundariesType = "parquet"
)
spe
}
\author{
Dario Righelli, Benedetta Banzi
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{readAndAddPolygonsToSPE}
\alias{readAndAddPolygonsToSPE}
\title{readAndAddPolygonsToSPE}
\usage{
readAndAddPolygonsToSPE(
  spe,
  polygonsCol = "polygons",
  keepMultiPol = TRUE,
  boundariesType = c("csv", "HDF5", "parquet")
)
}
\arguments{
\item{spe}{A \code{SpatialExperiment} object. The object should contain
metadata with the field \code{"technology"}, specifying the technology used
(e.g., "Nanostring_CosMx", "Vizgen_MERFISH", or "10X_Xenium").}

\item{polygonsCol}{character indicating the name of the polygons column to
add into the colData (default is `polygons`).}

\item{keepMultiPol}{Logical. If \code{TRUE}, multi-polygon features will be
kept when reading the boundary data. Defaults to \code{TRUE}.}

\item{boundariesType}{Character. Specifies the type of boundary file format
to read. Options are \code{"HDF5"} or \code{"parquet"}. Defaults to
\code{"HDF5"}.}
}
\value{
A \code{SpatialExperiment} object with the added polygon data.
}
\description{
Read and Add Polygons to a SpatialExperiment Object

This function reads polygon boundary data based on the technology associated
with the provided SpatialExperiment (SPE) object and adds the polygons to the
SPE.
}
\examples{
example(readCosmxSPE)
spe <- readAndAddPolygonsToSPE(spe)
colData(spe)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{.createPolygons}
\alias{.createPolygons}
\title{.createPolygons}
\usage{
.createPolygons(
  spat_obj,
  x = NULL,
  y = NULL,
  polygon_id = NULL,
  geometry = "Geometry"
)
}
\arguments{
\item{spat_obj}{A data frame or similar object containing spatial data.}

\item{x}{A character vector specifying the x-coordinates.}

\item{y}{A character vector specifying the y-coordinates.}

\item{polygon_id}{A character string specifying the polygon ID.}
}
\value{
An `sf` object containing the created polygons.
}
\description{
This internal function creates polygons from a data.frame or
similar object.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{readPolygonsCosmx}
\alias{readPolygonsCosmx}
\title{readPolygonsCosmx}
\usage{
readPolygonsCosmx(
  polygonsFile,
  type = c("csv", "parquet"),
  x = "x_global_px",
  y = "y_global_px",
  xloc = "x_local_px",
  yloc = "y_local_px",
  keepMultiPol = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{polygonsFile}{A character string specifying the file path to the
polygon data.}

\item{type}{A character string specifying the file type ("csv" or "parquet").}

\item{x}{A character string specifying the x-coordinate column.}

\item{y}{A character string specifying the y-coordinate column.}

\item{xloc}{A character string specifying the local x-coordinate column.}

\item{yloc}{A character string specifying the local y-coordinate column.}

\item{keepMultiPol}{A logical value indicating whether to keep multipolygons.}

\item{verbose}{A logical value indicating whether to print additional
information.}
}
\value{
An `sf` object containing the CosMx polygon data.
}
\description{
This function reads polygon data specific to CosMx technology.
}
\examples{
example(readCosmxSPE)
polygons <- readPolygonsCosmx(metadata(spe)$polygons)
polygons
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeOutliersQScore}
\alias{computeOutliersQScore}
\title{computeOutliersQScore}
\usage{
computeOutliersQScore(
  spe,
  metricList = c("log2SignalDensity", "Area_um", "log2AspectRatio",
    "log2Ctrl_total_ratio")
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial omics data.}

\item{metricList}{A character vector specifying the metrics to include in
the Quality Score formula. Default is `c("log2SignalDensity", "Area_um",
"log2AspectRatio", "log2Ctrl_total_ratio")`.}
}
\value{
The `SpatialExperiment` object with added outlier variables in
`colData` and the temporary metrics considered for Quality Score computation
stored in the object `metadata`.
}
\description{
Computes outlier cells for each metric that can be used in Quality Score formula
for SpatialExperiment.

This function calculates outlier cells for each variable specified
in `metricList` for a `SpatialExperiment`. log2SignalDensity must be present in
the `colData` of the `SpatialExperiment` object as a minimum requirement.
The user can choose which metrics to include among the following: Area_um,
log2Ctrl_total_ratio, log2AspectRatio. For Xenium and Merfish datasets,
even if outliers for log2AspectRatio are detected, neither the metric will be 
included in the Quality Score formula nor nor the outliers will be used for
training.
}
\details{
The function computes outliers for each specified metric after
automatically choosing the appropriate method according to the skewness of
the distribution.
Internally the function:
\enumerate{
   \item Calls \code{.checkSkw()} to choose the proper outlier detection
    method according to the variable skewness;
   \item Calls \code{computeSpatialOutlier()} on each included metric to get
   fences;
   \item Labels cells as “LOW”/“HIGH” outliers or “NO”.
}
}
\examples{
example(readCosmxSPE)
spe <- spatialPerCellQC(spe)
spe <- computeOutliersQScore(spe)
table(spe$log2SignalDensity_outlier_train)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{readh5polygons}
\alias{readh5polygons}
\title{readh5polygons}
\usage{
readh5polygons(polFile)
}
\arguments{
\item{polFile}{A character string specifying the file path to the HDF5
polygon data.}
}
\value{
A list containing the polygon geometries and their associated cell_id
}
\description{
This function reads polygon data from an HDF5 file.
}
\author{
Lambda Moses
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{computeLambda}
\alias{computeLambda}
\title{computeLambda}
\usage{
computeLambda(trainDF, modelFormula)
}
\arguments{
\item{trainDF}{`data.frame`
A data frame for QS model training that must include:
  Predictor columns: All columns referenced in the formula returned by `getModelFormula()`.
  `QScore_train` or historical `qcscore_train`, containing the binary
  (0/1) response to be modeled.}

\item{modelFormula}{A one-sided formula or character string describing the
supported predictor terms.}
}
\value{
`numeric`
  The value of \eqn{\lambda} (i.e., `lambda.min`) from
  `cv.glmnet` that minimizes the cross-validation error.
}
\description{
Computes optimal ridge regularization parameter \eqn{\lambda} via
k-fold cross-validation.

\code{computeLambda} performs ridge (L2) logistic regression with
k-fold cross-validation to identify the optimal regularization parameter
\eqn{\lambda} for Quality Score (QS) model training.
}
\details{
Complete cases for the formula variables and response are retained. The
function constructs a model matrix and passes it to an internal
matrix-based implementation that runs ridge logistic-regression
cross-validation using `cv.glmnet` with `alpha = 0`.
}
\examples{
example(computeTrainDF)
modform <- getModelFormula(names(metadata(spe)$formula_variables))
best_lambda <- computeLambda(df_train, modform)
print(best_lambda)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{dot-computeXenMerTrainSet}
\alias{dot-computeXenMerTrainSet}
\alias{.computeXenMerTrainSet}
\title{.computeXenMerTrainSet}
\usage{
.computeXenMerTrainSet(spe)
}
\arguments{
\item{spe}{\code{SpatialExperiment}}
}
\value{
A list with elements \code{bad} and \code{good}, each a data.frame
with \code{qscore_train} and (for “good”) an \code{is_a_bad_boy} flag.
}
\description{
Internal: builds Training Set for Xenium & MERFISH.
Splits a SpatialExperiment into “bad” vs “good” cells based on
pre-computed outlier labels on log2SignalDensity.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{.checkPolygonsValidity}
\alias{.checkPolygonsValidity}
\title{.checkPolygonsValidity}
\usage{
.checkPolygonsValidity(
  sf,
  geometry = NULL,
  keepMultiPol = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{sf}{An `sf` class object containing the spatial data.}

\item{geometry}{character for the geometry to check validity, if `NULL`
it checks the active geometry (default is `NULL`)}

\item{keepMultiPol}{logical for keeping/removing moltipolygons, if any
(default is `TRUE`, so keeping the multipolygons)}

\item{verbose}{logical to print verbose output (default is `FALSE`)}
}
\value{
An `sf` object with valid geometries, possibly with multipolygons
removed.
}
\description{
checks validity on a geometry of `sf` object.
It removes multipolygons when `keepMultiPol` is `FALSE`
}
\details{
In case geometry is NULL validity is checked on the active geometry,
otherwise it is checked on the passed geometry without changing the active
geometry of the sf object.
In case of not valid polygons, these are removed.
If keeMultiPol is FALSE, possible detected multipolygons are removed.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{checkOutliers}
\alias{checkOutliers}
\title{checkOutliers}
\usage{
checkOutliers(spe, verbose = FALSE)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with spatial omics data.}

\item{verbose}{Logical. If `TRUE`, prints how many outliers were found for
each metric.}
}
\value{
The `SpatialExperiment` object with added QCScore metric variables
 in the `metadata`.
}
\description{
Checks if computed outliers meet the minimum numerical requirement, being
at least 0.1% of total cells for each metric to be used in QC score formula.
If the requirement is not met, the variable is removed from the formula.
}
\details{
The function checks if computed outliers for each metric meet
the minimum number to get the metric included in the QC score formula.
If verbose is TRUE, it also prints how many outliers were found for each
metric.
}
\examples{
example(computeOutliersQCScore)
spe <- checkOutliers(spe, verbose = TRUE)
metadata(spe)$formula_variables
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{getModelFormula}
\alias{getModelFormula}
\title{getModelFormula}
\usage{
getModelFormula(formulaVars, verbose = FALSE, ...)
}
\arguments{
\item{formulaVars}{A named character vector mapping metric names to outlier
columns, as historically stored in
`metadata(spe)$formula_variables`. An unnamed character vector of metric
names is also accepted.}

\item{verbose}{Logical. If `TRUE`, prints the generated formula.}

\item{...}{May contain the named replacement `metricList`. If both
`formulaVars` and `metricList` are supplied, `metricList` takes precedence
with a warning.}
}
\value{
`character`
  A one‐sided formula as a string (e.g. "~ log2SignalDensity + ...").
}
\description{
Returns the right‐hand side of a model formula string based on a vector of
metric names.
}
\examples{
example(checkOutliers)
getModelFormula(metadata(spe)$formula_variables)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{dot-checkSkw}
\alias{dot-checkSkw}
\alias{.checkSkw}
\title{.checkSkw}
\usage{
.checkSkw(
  cd,
  metricList = c("log2CountArea", "Area_um", "log2AspectRatio", "log2Ctrl_total_ratio")
)
}
\arguments{
\item{cd}{colData of `SpatialExperiment` object.}

\item{metricList}{A character vector specifying the metrics to include in
the QC score formula. Defaults are "log2CountArea", "Area_um",
"log2AspectRatio", "log2Ctrl_total_ratio".}
}
\value{
A vector containing the list of chosen outlier detection method for each
metric.
}
\description{
Check skewness of metrics to choose outlier detection method.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{.renameGeometry}
\alias{.renameGeometry}
\title{.renameGeometry}
\usage{
.renameGeometry(sf, from, to, activate = FALSE)
}
\arguments{
\item{sf}{an sf object with the `from` geometry}

\item{from}{character indicating the name of the geometry to change}

\item{to}{character indicating the new name of the geometry}

\item{activate}{logical indicating if the renamed geometry has to be
activated}
}
\value{
an sf object
}
\description{
renames the `from` to `to` geometry of the `sf` object.
If `activate` is `TRUE` it set as the active geometry the new geometry name.
Default behaviour is to check if the renamed geometry is already active and
leave it as active with the new name.
}
\examples{
example(readPolygonsCosmx)
.renameGeometry(polygons, "global", "global1")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spatialQCPlots.R
\name{plotQScoreTerms}
\alias{plotQScoreTerms}
\title{plotQScoreTerms}
\usage{
plotQScoreTerms(
  spe,
  sampleId = unique(spe$sample_id),
  size = 0.05,
  alpha = 0.8,
  aspectRatio = 1,
  custom = FALSE
)
}
\arguments{
\item{spe}{A `SpatialExperiment` object with `quality_score` and term
columns in `colData`.}

\item{sampleId}{Character string for plot title. Must match values in the
`fov` column of `colData(spe)`. Default: `unique(spe$sample_id)`.}

\item{size}{Numeric point size for the scatter plots. Default: `0.05`.}

\item{alpha}{Numeric transparency for the scatter plots. Default: `0.2`.}

\item{aspectRatio}{Numeric aspect ratio of the plots. Default: `1`.}

\item{custom}{Logical; if `TRUE`, use custom polygon‐derived metrics.}
}
\value{
A combined plot (via `cowplot::plot_grid`) showing spatial maps
  of each QS term.
}
\description{
Plots the individual terms that combine into the quality score formula,
allowing assessment of each term’s impact on the final score.
}
\examples{
example(readAndAddPolygonsToSPE)
example(spatialPerCellQC)
p <- plotQScoreTerms(spe)
print(p)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{trainModel}
\alias{trainModel}
\title{trainModel}
\usage{
trainModel(modelMatrix, trainDF)
}
\arguments{
\item{modelMatrix}{a matrix describing the model variables, tipically created
with `getModelFormula` and `model.matrix` functions.}

\item{trainDF}{`data.frame`
A data frame containing at least the response column
`qscore_train`, coded as 0/1.}
}
\value{
A \code{\link[glmnet]{glmnet}} model object fitted with
\code{family="binomial"}, \code{alpha=0} (ridge), and a sequence of
\eqn{\lambda} values.
}
\description{
Fit a Ridge Logistic Regression Model

\code{trainModel} fits an L2-regularized (ridge) logistic regression
using \pkg{glmnet}, given a design matrix and a training data frame.
}
\examples{
example(computeTrainDF)
model_formula <- getModelFormula(metadata(spe)$formula_variables)
model_matrix <- model.matrix(as.formula(model_formula), data=df_train)
fit <- trainModel(model_matrix, df_train)
coef(fit, s = 0.01)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plotUtils.R
\docType{data}
\name{firstFlagPalette}
\alias{firstFlagPalette}
\title{firstFlagPalette}
\format{
An object of class \code{character} of length 6.
}
\usage{
firstFlagPalette
}
\value{
a palette for firstFlagPlot
}
\description{
neon color palette for firstFlagPlot
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/QC.R
\name{applyQScoreModel}
\alias{applyQScoreModel}
\title{applyQScoreModel}
\usage{
applyQScoreModel(spe, qsModel, scoreName = "QScore")
}
\arguments{
\item{spe}{A `SpatialExperiment` object with QC metrics already computed.}

\item{qsModel}{A Quality Score model object, usually stored in
`metadata(spe)$QScore_model`.}

\item{scoreName}{Name of the output column in `colData`.}
}
\value{
A `SpatialExperiment` object with added Quality Score in `colData`.
}
\description{
Applies a previously trained Quality Score model to a new SpatialExperiment
object.
See details for important considerations when applying a model to a
different dataset.
}
\details{
The authors do not recommend applying a Quality Score (QS) model trained on one
dataset to a different dataset, but this could be useful in some cases,
for example to transfer a model trained on a dataset sourced from a tissue 
slide sample and then applied to other slides from the same experiment.
When transferring the model from a reference dataset of the same technology
of the query dataset, there is no need to set a common formula. However,
when transferring from a different technology, reference QS computation 
should be performed providing a formula that excludes untransferable or unwanted 
variables (e.g. computing QS on a reference CosMx dataset omitting the border 
effect term to transfer the QS model to a Xenium query dataset. This is done by
providing a custom formula in computeQScore, see link computeQScore). The function will check 
for the presence of required model variables in the new dataset and will handle 
missing or extra variables accordingly.
}
\examples{

example(readCosmxSPE)

## Compute per-cell quality control metrics
spe <- spatialPerCellQC(spe)

## Split the object only for example purposes
set.seed(1998)
idx <- sample(seq_len(ncol(spe)), floor(ncol(spe) / 2))
spe_train <- spe[, idx]
spe_test <- spe[, -idx]

## Train the Quality Control (QC) score model on one dataset
spe_train <- computeQScore(spe_train)
qs_model <- metadata(spe_train)$QScore_model

## Apply the trained model to another dataset
spe_test <- applyQScoreModel(
    spe=spe_test,
    qsModel=qs_model,
    scoreName="QScore_transferred"
)

summary(spe_test$QScore_transferred)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readCosMx.R
\name{.checkFovPositionVersion}
\alias{.checkFovPositionVersion}
\title{.checkFovPositionVersion}
\usage{
.checkFovPositionVersion(spe)
}
\arguments{
\item{spe}{A `SpatialExperiment` containing FOV position information
in the metadata to be standardized.}
}
\value{
A `SpatialExperiment` with updated and standardized column names
for the metadata `fov_position` `data.frame`.
}
\description{
Check and Standardize FOV Position Column Names

This internal utility function standardizes column names of a data frame
containing Field of View (FOV) positional information.
It modifies column names to ensure compatibility with expected naming
conventions, including support for older formats.

Specifically, it:
- Renames any column containing "FOV" to "fov"
- Converts columns with coordinates matching "X", "Y", or "Z" to lowercase
- Replaces suffix "_px" with "_global_px" for coordinate pixel columns
- If the input contains `x_mm` and `y_mm` columns, the function computes
corresponding `x_global_px` and `y_global_px` values by converting from
millimeters to pixels using a fixed resolution factor (0.12028 mm/pixel).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readCosMx.R
\name{readCosmxSPE}
\alias{readCosmxSPE}
\alias{readCosmxProteinSPE}
\title{readCosmxSPE}
\usage{
readCosmxSPE(
  dirName,
  sampleName = "sample01",
  coordNames = c("CenterX_global_px", "CenterY_global_px"),
  countMatFPattern = "exprMat_file.csv",
  metadataFPattern = "metadata_file.csv",
  polygonsFPattern = "polygons.csv",
  fovPosFPattern = "fov_positions_file.csv",
  fovdims = c(xdim = 4256, ydim = 4256)
)
}
\arguments{
\item{dirName}{A character string specifying the directory containing the
CosMx data files.}

\item{sampleName}{A character string specifying the sample name. Default is
`"sample01"`.}

\item{coordNames}{A character vector specifying the names of the spatial
coordinate columns in the data. Default is `c("CenterX_global_px",
"CenterY_global_px")`.}

\item{countMatFPattern}{A character string specifying the pattern to match
the count matrix file. Default is `"exprMat_file.csv"`.}

\item{metadataFPattern}{A character string specifying the pattern to match
the metadata file. Default is `"metadata_file.csv"`.}

\item{polygonsFPattern}{A character string specifying the pattern to match
the polygons file. Default is `"polygons.csv"`.}

\item{fovPosFPattern}{A character string specifying the pattern to match the
FOV positions file. Default is `"fov_positions_file.csv"`.}

\item{fovdims}{A named numeric vector specifying the dimensions of the FOV
in pixels. Default is `c(xdim=4256, ydim=4256)`.}
}
\value{
A `SpatialExperiment` object containing the read CosMx data,
including count matrices, metadata, and optionally polygons.
}
\description{
Read and Construct a SpatialExperiment Object from CosMx Data

This function reads in data from Nanostring CosMx files and constructs a
`SpatialExperiment` object, optionally including polygon data.
}
\details{
The function reads in the specified files for count matrices,
metadata, and FOV positions, and constructs a `SpatialExperiment` object.
Optionally, polygon data can be read and added to the object.

readCosmxProteinSPE is a wrapper of readCosmxSPE, it only changes the
technology metadata in Nanostring_CosMx_Protein.
}
\examples{
cospath <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"),
   package="SpaceTrooper")
spe <- readCosmxSPE(cospath, sampleName="DBKero_Tiny")
}
\author{
Dario Righelli, Benedetta Banzi
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readCosMx.R
\name{updateCosmxSPE}
\alias{updateCosmxSPE}
\title{updateCosmxSPE}
\usage{
updateCosmxSPE(
  spe,
  dirName,
  sampleName = "sample01",
  polygonsFPattern = "polygons.csv",
  fovdims = c(xdim = 4256, ydim = 4256)
)
}
\arguments{
\item{spe}{SpatialExperiment object.}

\item{dirName}{Directory containing CosMx output files (e.g., polygon CSVs).}

\item{sampleName}{Character scalar, sample identifier stored in
\code{colData(spe)$sample_id}. Default \code{"sample01"}.}

\item{polygonsFPattern}{Character, pattern used by \code{list.files()} to
locate polygon files. Default \code{"polygons.csv"}.}

\item{fovdims}{Named numeric vector with entries \code{xdim} and \code{ydim}
representing the FOV dimensions in pixels.}
}
\value{
A SpatialExperiment object with updated metadata and column names.
}
\description{
Update a SpatialExperiment object derived from CosMx data by adding
polygons, FOV dimensions, standardized column names, and metadata.
}
\details{
The function standardizes CosMx SPE structure by:
- creating unique cell names of the form \code{f<fov>_c<cell_ID>};
- ensuring consistent cell identifiers and sample metadata;
- recording FOV dimensions, polygon paths, and technology type in
  \code{metadata(spe)}.
}
\examples{
cospath <- system.file(file.path("extdata", "CosMx_DBKero_Tiny"),
    package="SpaceTrooper")
spe <- SpatialExperimentIO::readCosmxSXE(dirName=cospath,
    addParquetPaths=FALSE)
spe <- updateCosmxSPE(spe, dirName=cospath, sampleName="DBKero_Tiny")
}
\seealso{
readCosmxSPE, readCosmxProteinSPE
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/polygons.R
\name{readPolygons}
\alias{readPolygons}
\title{readPolygons}
\usage{
readPolygons(
  polygonsFile,
  type = c("csv", "parquet", "h5"),
  x = c("x_global_px", "vertex_x"),
  y = c("y_global_px", "vertex_y"),
  xloc = "x_local_px",
  yloc = "y_local_px",
  keepMultiPol = TRUE,
  verbose = FALSE
)
}
\arguments{
\item{polygonsFile}{A character string specifying the path to the polygon
file.}

\item{type}{A character string specifying the file type. Supported types are
`"csv"`, `"parquet"`, and `"h5"`. Default is `"csv"`.}

\item{x}{A character vector specifying the column names for the x-coordinates
in the polygon data. Default is `c("x_global_px", "vertex_x")`.}

\item{y}{A character vector specifying the column names for the y-coordinates
in the polygon data. Default is `c("y_global_px", "vertex_y")`.}

\item{xloc}{A character string specifying the column name for the local
x-coordinates. Default is `"x_local_px"`.}

\item{yloc}{A character string specifying the column name for the local
y-coordinates. Default is `"y_local_px"`.}

\item{keepMultiPol}{A logical value indicating whether to keep multipolygons
during validation. Default is `TRUE`.}

\item{verbose}{A logical value indicating whether to print additional
information during processing. Default is `FALSE`.}
}
\value{
An `sf` object with the loaded and validated polygons.
}
\description{
Read and Validate Polygons from a File

This function reads polygon data from a specified file, validates the
polygons, and returns them as an `sf` object. It supports multiple file
formats and can handle both global and local coordinates.
}
\details{
The function reads polygon data from the specified file and formats.
It validates the polygons and handles both global and local coordinates if
provided. If the file type is `"h5"`, the function currently does not handle
the data, as this part of the code is not implemented.
}
\examples{
example(readCosmxSPE)
polygons <- readPolygons(metadata(spe)$polygons)
polygons
}
